#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include <exec/execbase.h>
#include <exec/exec_protos.h>
#include <exec/tasks.h>
#include "serial.h"
#include <exec/io.h>
#include <exec/semaphores.h>
#include <dos/dos.h>


/* core.c */

void	MonitorTask(void);
void    DebugWrite(char *);
int     DebugRead(char *);
int     Call(ULONG address);

struct ExecBase *SysBase = NULL;

void DebugWriteUlong(char *p, ULONG u)
{
    char a[16];
    DebugWrite(p);
    sprintf(a,"0x%08x\n\r",u);
    DebugWrite(a);
}

void DebugWriteUword(char *p, ULONG u)
{
    char a[16];
    DebugWrite(p);
    sprintf(a,"0x%04x\n\r",u&0xffff);
    DebugWrite(a);
}

static const char welcome[]="\n\r"\
"       ____   ___   ____   ____  ____   ___   _  _ \n\r"\
"      | |__) / / \\ | |__) | |_  | |  \\ / / \\ | \\/ |\n\r"\
"      |_|__) \\_\\_/ |_| \\  |_|__ |_|__/ \\_\\_/ |_||_|\n\r"\
"BOAR's Odd Realtime Executive & Design Operation's Manager\n\r"\
"    ==================================================\n\r"\
"    Pre-emptive Multitasking Executive Kernel for BOAR\n\r"\
"                (C)1992-95 Pasi Ojala\n\r"\
"    ==================================================\n\r"\
"    Licenced to:   TTKK Signal Processing Laboratory\n\r"\
"                   Nokia Telecommunications\n\r";

#define INLINESIZE 256

static const struct SegList fakeSegment[] =
{{0, 0},
 {(struct SegList *)0x4e714ef9, (ULONG)MonitorTask}};

void MonitorTask(void)
{
    int i;
    APTR ser;
    char *a, out[80];
    struct Process *myself = (struct Process *)FindTask(NULL);
    struct SegList *segl;
    struct Library *DOSBase; /* Global DOSBase is used for calls */

    if(!(DOSBase = OpenLibrary(DOSL_NAME)))
	return;

    if((segl = ROMLoadSeg("aux-handler.b68k")))
    {
	CreateProc("aux-launcher", 5, segl, 8192, NULL);
	Delay(1);
    }
    if((segl = ROMLoadSeg("rom.b68k")))
    {
	CreateProc("rom-launcher", 5, segl, 2048, NULL);
	Delay(1);
    }
    /* "Monitor 1" opens unit 0, "Monitor 2" opens unit 1 */

    a = FindTask(NULL)->tc_Node.ln_Name;
    while(*a && !isdigit(*a))
	a++;

    i = atol(a); /* 1 or 2 */

    if(i==1)
    {
	ser = Open("SER:serial.device/0/con", MODE_OLDFILE);
    }
    else
    {
	ser = Open("SER:serial.device/1/con", MODE_OLDFILE);
    }
    if( !ser )
    {
	CloseLibrary(DOSBase);
	return;
    }
    Write(ser, (char *)welcome,sizeof(welcome));

    if(myself->pr_Task.tc_Node.ln_Type == NT_PROCESS)
    {
	myself->pr_CIS = ser;
	myself->pr_COS = ser;
	myself->pr_CES = ser;
    }

    while(1)
    {
	SetMode(ser, MODE_RAW);

	if((segl = LoadSeg("ROM:l/shell-seg")))
	{
	    Write(ser, "Press a key to use the resident shell\n", 38);
	    if(WaitForChar(ser, 200))
	    {
		UnLoadSeg(segl);
		segl =  NULL;
	    }
	}

	if(segl || (segl = ROMLoadSeg("shell-seg")))
	{
	    struct Remember *mem;
	    ULONG sigalloc = myself->pr_Task.tc_SigAlloc;
	    struct SegList *oldsegl = (struct SegList *)myself->pr_SegList;
	
	    sprintf(out, "Starting shell\n");
	    Write(ser, out, strlen(out));
	    myself->pr_SegList = (APTR)segl;
	    sprintf(out, "Returned: %d\n\r", RunCommand(segl, 4000, "", 0));
	    Write(ser, out, strlen(out));
	    if(myself->pr_SegList)
	    {
		UnLoadSeg((struct SegList *)myself->pr_SegList);
	    }
	    myself->pr_SegList = (APTR)oldsegl;
	    
	    while((mem =
		   (struct Remember *)RemHead(&FindTask(NULL)->tc_MemList)))
	    {
		sprintf(out, "Freeing @ 0x%08x, %ld bytes\n\r",
			mem+1, mem->rm_Size - sizeof(struct Remember));
		Write(ser, out, strlen(out));
		FreeMem(mem, mem->rm_Size);
	    }
	    if(myself->pr_Task.tc_SigAlloc != sigalloc)
	    {
		sprintf(out, "Signals not freed by the program! (before: 0x%08x, after: 0x%08x)\n\rRestoring old allocation state\n\r", sigalloc, myself->pr_Task.tc_SigAlloc);
		myself->pr_Task.tc_SigAlloc = sigalloc;
		Write(ser, out, strlen(out));
	    }
	}
	else
	{
	    Write(ser, "Could not find ROM:l/shell-seg nor a resident version of shell-seg.\n", 68);
	    break;
	}
    }
    Close(ser);
    CloseLibrary(DOSBase);
}

void _init(void)
{
    extern char __nil_handler[];
    extern char __ram_handler[];
    extern char __rom_handler[];
    extern char __aux_handler[];
    extern char __rom_shell[];
    extern char __rom_rx[];
    extern char __rom_sx[];
    ULONG  Signals;
    struct Library *DOSBase; /* global DOSBase is used for calls */
    struct Program *prg;
    struct SignalSemaphore *ss;

    SysBase = *(struct ExecBase **)4;

    FindTask(NULL)->tc_Node.ln_Name = "init";

    InitDosLibrary(); /* InitDosLibrary() initializes DOSBase */
    DOSBase = OpenLibrary(DOSL_NAME);

    AddSerialDevice();

    if((prg = (struct Program *)
	AllocMem(sizeof(struct Program) + strlen("nil.b68k"), MEMF_CLEAR)))
    {
	prg->pm_Node.ln_Name = &prg->pm_Name[0];
	strcpy(prg->pm_Node.ln_Name, "nil.b68k");
	prg->pm_Exec = (struct Executable *)&__nil_handler[0];
	prg->pm_Size = 0; /* To prevent the EPROM to be made free memory */
	AddProgram(prg);
    }

    if((prg = (struct Program *)
	AllocMem(sizeof(struct Program) + strlen("ram.b68k"), MEMF_CLEAR)))
    {
	prg->pm_Node.ln_Name = &prg->pm_Name[0];
	strcpy(prg->pm_Node.ln_Name, "ram.b68k");
	prg->pm_Exec = (struct Executable *)&__ram_handler[0];
	prg->pm_Size = 0; /* To prevent the EPROM to be made free memory */
	AddProgram(prg);
    }

    if((prg = (struct Program *)
	AllocMem(sizeof(struct Program) + strlen("rom.b68k"), MEMF_CLEAR)))
    {
	prg->pm_Node.ln_Name = &prg->pm_Name[0];
	strcpy(prg->pm_Node.ln_Name, "rom.b68k");
	prg->pm_Exec = (struct Executable *)&__rom_handler[0];
	prg->pm_Size = 0; /* To prevent the EPROM to be made free memory */
	AddProgram(prg);
    }

    if((prg = (struct Program *)
	AllocMem(sizeof(struct Program) + strlen("rx.b68k"), MEMF_CLEAR)))
    {
	prg->pm_Node.ln_Name = &prg->pm_Name[0];
	strcpy(prg->pm_Node.ln_Name, "rx.b68k");
	prg->pm_Exec = (struct Executable *)&__rom_rx[0];
	prg->pm_Size = 0; /* To prevent the EPROM to be made free memory */
	AddProgram(prg);
    }

    if((prg = (struct Program *)
	AllocMem(sizeof(struct Program) + strlen("sx.b68k"), MEMF_CLEAR)))
    {
	prg->pm_Node.ln_Name = &prg->pm_Name[0];
	strcpy(prg->pm_Node.ln_Name, "sx.b68k");
	prg->pm_Exec = (struct Executable *)&__rom_sx[0];
	prg->pm_Size = 0; /* To prevent the EPROM to be made free memory */
	AddProgram(prg);
    }

    if((prg = (struct Program *)
	AllocMem(sizeof(struct Program) + strlen("aux-handler.b68k"), MEMF_CLEAR)))
    {
	prg->pm_Node.ln_Name = &prg->pm_Name[0];
	strcpy(prg->pm_Node.ln_Name, "aux-handler.b68k");
	prg->pm_Exec = (struct Executable *)&__aux_handler[0];
	prg->pm_Size = 0; /* To prevent the EPROM to be made free memory */
	AddProgram(prg);
    }

#if 1
    if((prg = (struct Program *)
	AllocMem(sizeof(struct Program) + strlen("shell-seg"), MEMF_CLEAR)))
    {
	prg->pm_Node.ln_Name = &prg->pm_Name[0];
	strcpy(prg->pm_Node.ln_Name, "shell-seg");
	prg->pm_Exec = (struct Executable *)&__rom_shell[0];
	prg->pm_Size = 0; /* To prevent the EPROM to be made free memory */
	AddProgram(prg);
    }
#endif

#if 0 /* DOES NOT WORK!! ?? Too little stack ? */
    if((segl = ROMLoadSeg("nil.b68k")))
    {
	CreateProc("nil-launcher", 5, segl, 8192, NULL);
    }
    if((segl = ROMLoadSeg("rom.b68k")))
    {
	CreateProc("rom-launcher", 5, segl, 8192, NULL);
    }
    if((segl = ROMLoadSeg("aux-handler.b68k")))
    {
	CreateProc("aux-launcher", 5, segl, 8192, NULL);
	Delay(2);
    }
#else
#endif
#define PORT_NAME "emulation-rendezvous"
    if(!(ss = FindSemaphore(PORT_NAME)))
    {
	ss = (struct SignalSemaphore *)
	    AllocMem(sizeof(struct SignalSemaphore)+strlen(PORT_NAME)+1, 0);
	if(ss)
	{
	    char *name = (char *)ss + sizeof(struct SignalSemaphore);

	    strcpy(name, PORT_NAME);
	    AddSemaphore(ss, name, 0);
	}
    }

    CreateProc("Monitor 1", 0, &fakeSegment, 2048, NULL);
    CreateProc("Monitor 2", 0, &fakeSegment, 2048, NULL);

    CloseLibrary(DOSBase);
    SetTaskPri(FindTask(NULL), -5);
    while(1)
    {
	Signals = Wait(SIGF_PORT | SIGF_ABORT);
	if(Signals & SIGF_ABORT)
	{
	    /* Child exited */
	}
	if(Signals & SIGF_PORT)
	{
	    /* Message arrived */
	}
    }
}

