#include "execbase.i"
#include "tasks.i"
#include "io.i"
.text

|---------------------------------------------------------------------------
| AddPort( port )
|---------------------------------------------------------------------------
AddPort:
	jsr	A5@(LIB_Forbid)
	lea	A5@(PortList),A0	| List to add to
	jsr	A5@(LIB_Enqueue)
	jsr	A5@(LIB_Permit)
	rts

|---------------------------------------------------------------------------
| RemPort( port )
|---------------------------------------------------------------------------
RemPort:
	jsr	A5@(LIB_Forbid)
	jsr	A5@(LIB_Remove)
	jsr	A5@(LIB_Permit)
	rts

|---------------------------------------------------------------------------
| CreatePort( name ) (NULL for private port)
|---------------------------------------------------------------------------
CreatePort:
	moveml	A2-A3,SP@-		| save registers
	moveal	A1,A3
	movel	#mp_SIZE,D0
	jsr	A5@(LIB_AllocMem)
	tstl	D0
	beq	CreateEnd
	moveal	D0,A2
	jsr	A5@(LIB_AllocSignal)
	tstl	D0
	bmi	CreateFailed

	moveb	D0,A2@(mp_SigBit)
	moveb	#PA_SIGNAL,A2@(mp_Flags)
	lea	A2@(mp_MsgList),A0
	jsr	A5@(LIB_NewList)

	moveb	#NT_PORT,A2@(mp_Node+ln_Type)
	moveb	#0,A2@(mp_Node+ln_Pri)
	movel	A5@(ThisTask),A2@(mp_SigTask)

	movel	A2,D0		| return (struct MsgPort *)D0
	movel	A3,D1		| Name ?
	movel	D1,A2@(mp_Node+ln_Name)
	tstl	D1
	beq	CreateEnd	| Ok, it is not a public port then..

	jsr	A5@(LIB_Forbid)
	lea	A5@(PortList),A0 | List to add to
	moveal	A2,A1		| Node to add
	jsr	A5@(LIB_Enqueue)
	jsr	A5@(LIB_Permit)
	movel	A2,D0
		
	moveml	SP@+,A2-A3	| restore registers
	rts
CreateFailed:
	movel	#mp_SIZE,D0
	movel	A2,A0
	jsr	A5@(LIB_FreeMem)
	moveql	#0,D0
CreateEnd:
	moveml	SP@+,A2-A3	| restore registers
	rts

|---------------------------------------------------------------------------
| DeletePort( port )
|---------------------------------------------------------------------------
DeletePort:
	moveml	A2,SP@-		| save registers
	moveal	A1,A2		| struct MsgPort *
	tstl	A2@(mp_Node+ln_Name)
	beq	DeletePortPrivate
	| Public port - Remove from Exec PortList
	jsr	A5@(LIB_Forbid)
	moveal	A2,A1
	jsr	A5@(LIB_Remove)
	jsr	A5@(LIB_Permit)
DeletePortPrivate:
	moveb	A2@(mp_Flags),D0
	cmpb	#PA_SIGNAL,D0
	bne	DeletePortNoSignal
	| Signal allocated - FreeSignal
	clrl	D0
	moveb	A2@(mp_SigBit),D0
	jsr	A5@(LIB_FreeSignal)
DeletePortNoSignal:
	movel	#mp_SIZE,D0
	movel	A2,A0
	jsr	A5@(LIB_FreeMem)
	moveml	SP@+,A2		| restore registers
	rts

|---------------------------------------------------------------------------
| FindPort( portname )
|---------------------------------------------------------------------------
FindPort:
	jsr	A5@(LIB_Forbid)
	lea	A5@(PortList),A0
	jsr	A5@(LIB_FindName)
	jmp	A5@(LIB_Permit)		| Does not mess D0

| ReleasePort: ??
| FindPort increases usecount, ReleasePort decreases it ?

|---------------------------------------------------------------------------
| WaitPort( MsgPort )
|---------------------------------------------------------------------------
WaitPort:
	movel	A0@(mp_MsgList+lh_Head),A1
	tstl	A1@(ln_Next)
	beq	L10
	movel	A1,D0
	rts
L10:
	movel	A0,SP@-
	moveb	A0@(mp_SigBit),D1
	clrl	D0
	bset	D1,D0
	jsr	A5@(LIB_Wait)
	movel	SP@+,A0
	bra	WaitPort

|---------------------------------------------------------------------------
| PutMsg( MsgPort, message )
|---------------------------------------------------------------------------
PutMsg:
	moveml	A2-A3,SP@-
	moveal	A0,A2
	moveal	A1,A3
	jsr	A5@(LIB_Disable)
	moveb	#NT_MESSAGE,A1@(mn_Node+ln_Type)
	lea	A2@(mp_MsgList),A0
	movel	A3,A1
	jsr	A5@(LIB_AddTail)
	moveb	A2@(mp_Flags),D0
	andb	#PF_ACTION,D0
	beq	PutMsgSignal
	andb	#PA_IGNORE,D0
	bne	PutMsgEnd
	| SoftInt..
	movel	A2@(mp_SigTask),D0
	moveml	SP@+,A2-A3
	jsr	A5@(LIB_Enable)
	jsr	A5@(LIB_Cause)
	rts
PutMsgSignal:
	moveb	A2@(mp_SigBit),D1
	clrl	D0
	bset	D1,D0
	movel	A2@(mp_SigTask),A0
	movel	A0,D1			| tstl	A0
	beq	PutMsgEnd
	jsr	A5@(LIB_Signal)
PutMsgEnd:
	moveml	SP@+,A2-A3
	jmp	A5@(LIB_Enable)

|---------------------------------------------------------------------------
| GetMsg( MsgPort )
|---------------------------------------------------------------------------
GetMsg:
	jsr	A5@(LIB_Disable)	| Disable() does not mess regs
	lea	A0@(mp_MsgList),A0
	jsr	A5@(LIB_RemHead)	| removed node in D0
	jsr	A5@(LIB_Enable)		| Enable() does not mess regs
	rts

|---------------------------------------------------------------------------
| ReplyMsg( Message )
|---------------------------------------------------------------------------
ReplyMsg:
	movel	A2,SP@-
	movel	A1,A2
	jsr	A5@(LIB_Disable)
	movel	A2@(mn_ReplyPort),A0
	movel	A0,D0			| tstl	A0
	beq	ReplyMsgEnd
	movel	A2,A1
	jsr	A5@(LIB_PutMsg)
ReplyMsgEnd:
	moveb	#NT_REPLYMSG,A2@(mn_Node+ln_Type)
	movel	SP@+,A2
	jmp	A5@(LIB_Enable)

|---------------------------------------------------------------------------
| SendTimerReq( message )
|---------------------------------------------------------------------------
SendTimerReq:
	jsr	A5@(LIB_Disable)	| Disable does not mess regs
	lea	A5@(TimerRequests),A0	| TimerRequests
	jsr	A5@(LIB_AddTail)
	jmp	A5@(LIB_Enable)


|----------------------------------------------------------------
| AddDevice( device )
|----------------------------------------------------------------

AddDevice:
	movel	A2,SP@-
	movel	A1,A2
	jsr	A5@(LIB_Forbid)
	lea	A5@(DeviceList),A0

	movel	A2@(lib_Node+ln_Name),A1
	jsr	A5@(LIB_FindName)	| See if there is already one!
	tstl	D0
	bnes	AddDeviceEnd
	lea	A5@(DeviceList),A0
	movel	A2,A1
	jsr	A5@(LIB_AddTail)
AddDeviceEnd:
	movel	SP@+,A2
	jmp	A5@(LIB_Permit)		| Does not mess D0

|----------------------------------------------------------------
| RemDevice( device )
|----------------------------------------------------------------

RemDevice:
	jsr	A5@(LIB_Forbid)
	tstw	A1@(lib_OpenCnt)
	bne	LNotRemDev
	movel	A1,D0
	jsr	A5@(LIB_Remove)		| Does not mess D0
	jmp	A5@(LIB_Permit)		| Does not mess D0
LNotRemDev:
	jsr	A5@(LIB_Permit)
	moveql	#0,D0
	rts

|----------------------------------------------------------------
| OpenDevice( devicename .. )
|----------------------------------------------------------------

OpenDevice:
	moveml	D2-D3/A2,SP@-
	movel	A0,A2
	movel	D0,D2
	movel	D1,D3
	jsr	A5@(LIB_Forbid)
	lea	A5@(DeviceList),A0
	jsr	A5@(LIB_FindName)
	tstl	D0
	bne	OpenDeviceOK
	moveql	#50,D0
	bra	OpenDeviceEnd
OpenDeviceOK:
	movel	D0,A5
	movel	A5,A2@(io_Device)
	movel	D3,SP@-		| Flags      \
	movel	A2,SP@-		| StdIOReq *  > paramters for C-programs
	movel	D2,SP@-		| Unit       /
	jsr	A5@(LIB_Open)	| Call device's Open()
	addl	#12,SP
	movel	AbsExecBase,A5
OpenDeviceEnd:
	jsr	A5@(LIB_Permit)	| Does not mess D0
	moveml	SP@+,D2-D3/A2
	rts

CloseDevice:
	movel	A1,A5
	movel	A5,SP@-		| Parameter for C-programs
	movel	A5@(io_Device),A5
	jsr	A5@(LIB_Close)	| Call devices Close
	addql	#4,SP
	movel	AbsExecBase,A5	| Restore Execbase
	rts

DoIO:
	moveml	A2,SP@-
	movel	A1,A2
	jsr	A5@(LIB_SendIO)
	movel	A2,A1
	jsr	A5@(LIB_WaitIO)
	moveml	SP@+,A2
	rts

SendIO:
	movel	A1@(io_Device),A5
	movel	A1,SP@-
	jsr	A5@(DEV_BeginIO)
	addql	#4,SP
	movel	AbsExecBase,A5
	rts

CheckIO:
	moveql	#0,D0
	cmpb	#NT_REPLYMSG,A1@(io_Message+mn_Node+ln_Type)
	bne	CheckIONotOK
	moveql	#-1,D0
CheckIONotOK:
	rts

WaitIO:
	moveml	A2,SP@-
	movel	A1,A2
LAgain:
	movel	A2,A1
	jsr	A5@(LIB_CheckIO)
	tstl	D0
	bne	WaitIOEnd
	movel	A2@(io_Message+mn_ReplyPort),A1
	moveb	A1@(mp_SigBit),D1
	clrl	D0
	bset	D1,D0
	movel	AbsExecBase,A5
	jsr	A5@(LIB_Wait)
	bra	LAgain
WaitIOEnd:
	jsr	A5@(LIB_Disable)
	movel	A2,A1
	jsr	A5@(LIB_Remove)	| Remove the message from the replyport
	jsr	A5@(LIB_Enable)
	moveml	SP@+,A2
	rts

AbortIO:
	cmpb	#NT_REPLYMSG,A1@(io_Message+mn_Node+ln_Type)
	beq	AbortIOEnd
	movel	A1@(io_Device),A5
	movel	A1,SP@-
	jsr	A5@(DEV_AbortIO)
	addql	#4,SP
	movel	AbsExecBase,A5
AbortIOEnd:
	rts

