#ifndef SERIAL_H
#define SERIAL_H

#include <exec/types.h>

/* 68302 internal serial port definitions */

struct BD
{
    volatile UWORD Status;
    volatile UWORD Length;
    volatile ULONG Address;
};

struct UART
{
    volatile UBYTE RFCR;         /* Receive Function Code Register */
    volatile UBYTE TFCR;         /* Transmit Function Code Register */
    volatile UWORD MRBLR;
    volatile UWORD RxInternalState;
    volatile UBYTE Reserved0;
    volatile UBYTE RBDnum;
    volatile ULONG RxInternalDataPointer;
    volatile UWORD RxInternalByteCount;
    volatile UWORD RxTemp;
    volatile UWORD TxInternalState;
    volatile UBYTE Reserved1;
    volatile UBYTE TBDnum;
    volatile ULONG TxInternalDataPointer;
    volatile UWORD TxInternalByteCount;
    volatile UWORD TxTemp;
/* Start of Uart-specific data */
    volatile UWORD MAX_IDL;      /* Maximum IDLE Characters (Receive) */
    volatile UWORD IDLC;         /* Temporary Receive IDLE Counter */
    volatile UWORD BRKCR;        /* Break Count Register (Transmit) */
    volatile UWORD PAREC;        /* Receive Parity Error Counter */
    volatile UWORD FRMEC;        /* Receive Framing Error Counter */
    volatile UWORD NOSEC;        /* Receive Noise Counter */
    volatile UWORD BRKEC;        /* Receive Break Condition Counter */
    volatile UWORD UADDR1;       /* UART Address Character 1 */
    volatile UWORD UADDR2;       /* UART Address Character 2 */
    volatile UWORD RCCR;         /* Receive Control Character Register */
    volatile UWORD CHARACTER[8]; /* Control Characters */
};

struct SCC
{
    struct BD RxBD[8];
    struct BD TxBD[8];
    struct UART Param;
};

struct SCCConf
{
    volatile UWORD Reserved0;
    volatile UWORD SCON;       /* SCC Configuration Register */
    volatile UWORD SCM;        /* SCC Mode Register */
    volatile UWORD DSR;        /* SCC Data Synchronization Register */
    volatile UBYTE SCCE;       /* SCC Event Register */
    volatile UBYTE Reserved1;
    volatile UBYTE SCCM;       /* SCC Mask Register */
    volatile UBYTE Reserved2;
    volatile UBYTE SCCS;       /* SCC Status Register */
    volatile UBYTE Reserved3;
    volatile UWORD Reserved4;
};

#define SIMODE *((volatile UWORD *)0x6508b2)
#define SIMASK *((volatile UWORD *)0x6508b4)
#define CR *((volatile UBYTE *)0x650860)

/* hardware flags - SER_BRK is currently returned */
#define SER_RX    1
#define SER_TX    2
#define SER_BSY   4
#define SER_CCR   8
#define SER_BRK  16
#define SER_IDL  32
#define SER_CD   64
#define SER_CTS 128
/* software flags */
#define SER_CTRL_C 256
#define SER_CTRL_D 512

#define SDCMD_QUERY     20
#define SDCMD_BREAK     21


#define SERF_EXCLUSIVE  (1<<0)	/* Exclusive - only one open allowed */
#define SERF_SHORTRET   (1<<1)  /* 1 character satisfies read requests */
#define SERF_COOKED     (1<<2)	/* Line-buffered mode, return until \n|\r */

/* 
   In ShortRet mode requests for more bytes than have been received
   will return the number of bytes currently in buffer. If no characters
   have been received, the read request will be blocked until at least
   one character arrives.

   In Cooked mode no read requests are returned until a newline
   character (\n or \r, whichever is the first one encountered)
   is received.

   If both SHORTRET and COOKED are present, request will not block
   if there is at least one character ready, but also, only characters
   upto the newline character (whichever is first encountered) are
   returned.

   Echo-mode is not yet supported.
*/

void    AddSerialDevice(void);

#endif
