   /************************************************************/
  /*	SER:serial.device/0/con/speed<speed>/                 */
 /*		          /raw/                              */
/************************************************************/

#include <exec/tasks.h>
#include <exec/io.h>
#include <exec/exec_protos.h>
#include <dos/dos.h>
#include <dos/dos_protos.h>

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

LONG handler(void);
LONG real_handler(void);

const UWORD kludgejump = 0x4ef9;
const ULONG kludgeaddr = (ULONG)handler;

static int launched = 0;
struct DOSLibrary *DOSBase = NULL;
struct MsgPort *myport = NULL;
struct DosList *dol = NULL;

LONG handler()
{
    struct Process *myself = (struct Process *)FindTask(NULL);

    /* Detaching (only once!) */
    if(!launched)
    {
	if((DOSBase = (struct DOSLibrary *)OpenLibrary(DOSL_NAME)))
	{
	    struct SegList *segl = (struct SegList *)myself->pr_SegList;

	    launched = 1;
	    myself->pr_SegList = NULL;
	    if(CreateProc("SER:", 40, segl, 2048, NULL ))
	    {
		Delay(2);
		return RETURN_OK;
	    }
	    CloseLibrary((struct Library *)DOSBase);
	    return RETURN_FAIL;
	}
	return RETURN_FAIL;
    }
    /* Ok, we are now running freely.. */

    /* Have we run previously ? */
    if(dol)
    {
	dol->dol_Task = myport;

	real_handler();

	Forbid();
	myself->pr_SegList = NULL;
	dol->dol_Task = NULL;
	Permit();
    }
    else
    {
	/* No, first invocation presumably */

	/* NOTE: take good care of Permit()'s.. ! */
	Forbid();
	if(FindName(&DOSBase->dl_Handlers, (UBYTE *)"SER:"))
	{
	    Permit();
	    return RETURN_FAIL;
	}
	/* forbid 1 */
	if(!(myport = CreatePort(NULL)))
	{
	    Permit();
	    return RETURN_FAIL;
	}
	
	/* forbid 1 */
	/* First create a devicenode */
	if((dol = (struct DosList *)
	    AllocMem(sizeof(struct DosList), MEMF_CLEAR)))
	{
	    dol->dol_Task = myport;
	    dol->dol_Node.ln_Name = "SER:";
	    dol->dol_Node.ln_Type = DLT_HANDLER;
	    
	    dol->dol_misc.dol_handler.dol_Handler = NULL;
	    dol->dol_misc.dol_handler.dol_Stacksize = 2048;
	    dol->dol_misc.dol_handler.dol_Priority = 40;
	    dol->dol_misc.dol_handler.dol_Startup = 0;
	    /* Prevent us to be unloaded and ready to be awakened by DOS */
	    dol->dol_misc.dol_handler.dol_SegList =
		(struct SegList *)myself->pr_SegList;
	    myself->pr_SegList = NULL;
	    
	    /* add to dos list ! */
	    AddTail(&DOSBase->dl_Handlers, (struct Node *)dol);
	    Permit();
	    
	    real_handler();

	    Forbid();
#if 1
	    dol->dol_Task = NULL;
	}
	else
	{
	    DeletePort(myport);
	}

#else
	    Remove((struct Node *)dol);
	    FreeMem((char *)dol, sizeof(struct DosList));
	}
	CloseLibrary((struct Library *)DOSBase);
#endif
	Permit();
    }
    return 0;
}


char vers[] = "\0$VER: AUX-Handler 1.33\n";

#define AUX_ECHO	1	/* echo characters */
#define AUX_CRLF	2	/* append LF to CR and vice versa (output) */

#define AUX_CMD_READ	32	/* serial CMD_READ active */
#define AUX_CMD_WRITE	64	/* serial CMD_WRITE active */

#define AUXBUFSIZE	8192	/* typeahead/edit buffer size */
#define MAXLINESIZE	AUXBUFSIZE-2	/* save two for \n\0  */
#define OUTBUFSIZE	2048	/* Temporary buffer for synchronous writes */
#define TEMPBUFSIZE	1024	/* Temporary buffer for synchronous reads and writes */
#define DEVICENAMELEN	32
#define MAXCNX		16

#define MYPORT_SIG	(1L<<myport->mp_SigBit)	/* signal bit for requests */
#define WRITE_SIG	(1L<<writereply->mp_SigBit)

#define SERIALNAME      "serial.device"

#define SDCMD_QUERY     20
#define SDCMD_BREAK     21

/* Global, per-connection information */
struct Connection
{
    char	devicename[DEVICENAMELEN];
    long	deviceunit;
    long	devicespeed;

    int		aux_open;	/* Open count */

    struct	StdIOReq	*ReadSER;
    struct	StdIOReq	*WriteSER;
    struct	StdIOReq	*TempSER;
    struct	TimerReq	*Timer;
    struct	MsgPort		*Timer_Port;

    struct	DosPacket	*rdpkt, *waitpkt, *writepkt, *writepkt2;

    struct	Task	*owner;	 /* the first opener (for Reopen) */
    struct	Task	*reader; /* the opener that gets signals */
    UBYTE	last;			/* last byte sent */
    UBYTE	in_c;			/* async read buffer */
    UBYTE	Mode;
    UBYTE	reserved;
    int		aux_stat;	/* aux flags */
    int		aux_avail;	/* available lines in buffer (cooked mode) */
    int		in_len;		/* available bytes in buffer (raw mode) */
    int		ser_len;	/* bytes in serial device buffer (approx.) */
    UBYTE	auxbuf[AUXBUFSIZE+1];	/* type ahead buffer */
    UBYTE	outbuf[OUTBUFSIZE+1];	/* async write buffer */
};

/* Prototypes */

void returnpkt(struct DosPacket*, ULONG, ULONG);
struct DosPacket *taskwait(struct Process *myproc);

int ret_readpacket(struct Connection *, struct Process*);

void close_ser(struct Connection *);
int  open_stuff(struct Connection *);
void puts_ser(struct Connection *, UBYTE *buf, int len);
int  write_ser(struct Connection *, UBYTE *, int);
int  buf_ser(struct Connection *, UBYTE *frombuf, int size);
int  query_ser(struct Connection *);
void get_ser(struct Connection *, UBYTE *destbuf, int size);
void set_read(struct Connection *);

void chk_file(char *, struct Connection *);

LONG handler(void);

struct MsgPort	*tempreply  = NULL;
struct MsgPort	*writereply = NULL;
UBYTE  tempbuf[TEMPBUFSIZE+1];

LONG real_handler()	/* this is kinda main() */
{
    struct Process	*myproc; /* my process */
    struct DosPacket	*mypkt;
    struct FileHandle	*fh = NULL;
    BOOL  run;	/* handler main loop flag */
    UBYTE *ptr;	/* ptr for name translation */
    char  *s;	/* ptr to name for open */
    int	i, tmp;
    char name[200];	/* filename passed to */
    struct Connection *cnx[MAXCNX], *cc;
    int	cnxs, cn;
    long signals, sigmask;
    
    /* Initializing the handler */
    
    myproc = (struct Process *)FindTask(NULL);
    
    /* Shared replyports */
    /* tempreply is used with DoIO() (CMD_WRITE and SDCMD_QUERY) */
    /* This way we have enough signals to serve quite many channels.. */
    
    tempreply  = (struct MsgPort *)CreatePort(0);
    writereply = (struct MsgPort *)CreatePort(0);
    
    for(cn=0;cn<MAXCNX;cn++)
	cnx[cn] = NULL;
    cnxs = 0;
    run = TRUE;

    /* done initial stuff, now for some work */
    while(run)
    {
	/* Create signal mask for active channels */
	sigmask = 0;
	for(cn=0;cn<MAXCNX;cn++)
	{
	    if(cnx[cn])
		sigmask |= (1L<<cnx[cn]->ReadSER->io_Message.mn_ReplyPort->mp_SigBit);
	}
	    
	signals = Wait(MYPORT_SIG | SIGF_DELAY | WRITE_SIG | sigmask);
	    
	/* WaitForChar() timeout */
	if((signals & SIGF_DELAY))
	{
	    for(cn=0;cn<MAXCNX;cn++)
	    {
		if(cnx[cn] &&
		   cnx[cn]->waitpkt &&
		   (cnx[cn]->Timer->tr_Node.ln_Type==NT_REPLYMSG))
		{
		    cnx[cn]->ser_len = query_ser(cnx[cn]);
		    /* TODO: handle chars in serial buffer */
		    if(cnx[cn]->in_len)
		    {
			returnpkt(cnx[cn]->waitpkt,
				  DOS_TRUE,
				  cnx[cn]->in_len);
		    }
		    else
		    {
			returnpkt(cnx[cn]->waitpkt,
				  DOS_FALSE,
				  cnx[cn]->waitpkt->dp_Result2);
		    }
		    cnx[cn]->waitpkt = NULL;
		}
	    }
	}
	
	/* Read() returned */
	if((signals & sigmask))
	{
	    for(cn=0;cn<MAXCNX;cn++)
	    {
		if(cnx[cn] && (cnx[cn]->aux_stat & AUX_CMD_READ) &&
		   CheckIO(cnx[cn]->ReadSER))
		{
		    /* fine, got a char */
		    WaitIO(cnx[cn]->ReadSER);
		    cnx[cn]->aux_stat &= ~AUX_CMD_READ;
		    
		    /* Buffer the character just read */
		    if(buf_ser(cnx[cn], &cnx[cn]->in_c, 1))
		    {
			/* Anything else in the serial buffer ? */
			cnx[cn]->ser_len = tmp = query_ser(cnx[cn]);
			if(tmp && AUXBUFSIZE-cnx[cn]->in_len-2 > 10)
			{
			    if(tmp > AUXBUFSIZE-cnx[cn]->in_len-9)
				tmp = AUXBUFSIZE-cnx[cn]->in_len-10;	/* hysteresis */
			    if(tmp > TEMPBUFSIZE)
				tmp = TEMPBUFSIZE;
			    
			    get_ser(cnx[cn], tempbuf, tmp);
			    buf_ser(cnx[cn], tempbuf, tmp);
			    cnx[cn]->ser_len -= tmp;
			}
			/* because buffer is not full,start another read */
			/* In raw mode we wait for the next read request */
			if(cnx[cn]->Mode != MODE_RAW)
			    set_read(cnx[cn]);
		    }
		    /* If buf_ser returns false, the buffer is full and */
		    /* we don't start another read request. */
		    
		    /* Someone waiting for characters ? */
		    if(cnx[cn]->waitpkt)
		    {
			Disable();
			if(cnx[cn]->Timer->tr_Node.ln_Type!=NT_REPLYMSG)
			    Remove(&cnx[cn]->Timer->tr_Node);
			Enable();
			
			returnpkt(cnx[cn]->waitpkt,
				  DOS_TRUE,
				  cnx[cn]->in_len+cnx[cn]->ser_len);
			cnx[cn]->waitpkt = NULL;
		    }
		    if(cnx[cn]->rdpkt)
		    {
			/* Either returs the packet or not */
			ret_readpacket(cnx[cn], myproc);
		    }
		}
	    }
	}
	    
	/* Write returned, check which one.. */
	if((signals & WRITE_SIG))
	{
	    for(cn=0;cn<MAXCNX;cn++)
	    {
		if(cnx[cn] && (cnx[cn]->aux_stat & AUX_CMD_WRITE) &&
		   CheckIO(cnx[cn]->WriteSER))
		{
		    cc = cnx[cn];
		    
		    cc->aux_stat &= ~AUX_CMD_WRITE;
		    returnpkt(cc->writepkt,
			      cc->writepkt->dp_Arg3,
			      cc->writepkt->dp_Result2);
		    cc->writepkt = NULL;
		    if(cc->writepkt2)
		    {
			if(write_ser(cc, (UBYTE *)cc->writepkt2->dp_Arg2,
				     (int)cc->writepkt2->dp_Arg3))
			{
			    /* Write finished already */
			    returnpkt(cc->writepkt2,
				      cc->writepkt2->dp_Arg3,
				      cc->writepkt2->dp_Result2);
			    cc->writepkt2 = NULL;
			}
			else
			{
			    /* Asynchronous write */
			    cc->aux_stat |= AUX_CMD_WRITE;
			    cc->writepkt = cc->writepkt2;
			    cc->writepkt2 = NULL;
			}
		    }
		}
	    }
	}
	
	if((signals & MYPORT_SIG))	/* Reopen or action request */
	{
	    /* TODO: First see what connection, then process */
	    
	    while(mypkt = (struct DosPacket *)GetMsg(myport))
	    {
		switch(mypkt->dp_Type)
		{
		case ACTION_OPEN:		/* Someone made Open() */
		    fh = (struct FileHandle *)AllocMem(sizeof(struct FileHandle), MEMF_CLEAR);
		    if(!fh)
		    {
			returnpkt(mypkt, DOS_FALSE, ERROR_NO_FREE_STORE);
			break;
		    }

		    /*	Multiple writers are allowed. */
		    /* get file name and Lower case it */
		    ptr = (UBYTE *)mypkt->dp_Arg2;
		    strcpy(name, (char *)ptr);
		    for(s=name;*s;s++)
			*s = tolower(*s);
	    
		    s=name;
		    if(!strcmp(name, "*") || !strncmp(name, "console:", 8))
		    {
			/* the same channel ("*" or "console:") */
			/* We have to see what channel belongs to the task, */
			/* because we return the same msgport for all openers*/
			for(i=0;i<MAXCNX;i++)
			{
			    if(cnx[i] &&
			       cnx[i]->owner ==
			       mypkt->dp_Message.mn_ReplyPort->mp_SigTask)
			    {
				break;
			    }
			}
			if(i==MAXCNX)	/* Couldn't find the channel */
			{
			    returnpkt(mypkt, DOS_FALSE, ERROR_OBJECT_IN_USE);
			    break;
			}
			cnx[i]->aux_open++;
		    }
		    else
		    {
			while(*s && *s != ':')
			    s++;		/* skip the handler name  */
			if(*s != ':')		/* weird, no handler name */
			{
			    returnpkt(mypkt, DOS_FALSE, ERROR_OBJECT_IN_USE);
			    break;
			}
			else
			    s++;		/* skip the colon */
		
			if((cc = (struct Connection *)
			    AllocMem(sizeof(struct Connection), MEMF_CLEAR)))
			{
			    chk_file(s, cc);
			    for(i=0;i<MAXCNX;i++)
			    {
				if(cnx[i] && !strcmp(cnx[i]->devicename,
						     cc->devicename) &&
				   cnx[i]->deviceunit == cc->deviceunit)
				{
				    FreeMem(cc, sizeof(struct Connection));
				    break;
				}
			    }
			    if(i==MAXCNX)
			    {
				/* Open a new serial */
				if(!(cc->aux_open = open_stuff(cc)))
				{
				    FreeMem(cc, sizeof(struct Connection));
				    returnpkt(mypkt, DOS_FALSE, ERROR_OBJECT_IN_USE);
				    break;
				}
				cc->reader = cc->owner = mypkt->dp_Message.mn_ReplyPort->mp_SigTask;
				i=0;
				while(cnx[i])
				    i++;
				cnx[i] = cc;
				/* init the port */
				cnxs++;
			    }
			    else
			    {
				/* Reopen a channel */
				cnx[i]->aux_open++;
			    }
			}
			else
			{
			    returnpkt(mypkt, DOS_FALSE, ERROR_OBJECT_IN_USE);
			    break;
			}
		    }
	    
		    fh->fh_Arg1 = (LONG)cnx[i];
		    fh->fh_Port = myport;
		    fh->fh_Pos = -1;
	    
		    returnpkt(mypkt, (LONG)fh, mypkt->dp_Result2);
		    break;
		    
		case ACTION_READ:	/* Someone made a Read() */
		    fh = (struct FileHandle *)mypkt->dp_Arg1;
		    cc = (struct Connection *)fh->fh_Arg1;
		    cn = 0;
		    while(cnx[cn] != cc && cn<MAXCNX)
			cn++;
		    cc->reader = mypkt->dp_Message.mn_ReplyPort->mp_SigTask;

		    if(cc->rdpkt)
		    {
			if(mypkt->dp_Arg3>=15)
			{
			    memcpy((UBYTE *)mypkt->dp_Arg2,
				   "Another reader\n", 15);
			    returnpkt(mypkt, 15, mypkt->dp_Result2);
			}
			else
			{
			    returnpkt(mypkt, DOS_FALSE, mypkt->dp_Result2);
			}
			break;
		    }
		    
		    cc->rdpkt = mypkt;
		    cc->ser_len = query_ser(cc);
		    (void)ret_readpacket(cc, myproc); /* Either returs or not */
		    break;
		    
		case ACTION_WRITE:	/* tell 'em we wrote them all */
		    fh = (struct FileHandle *)mypkt->dp_Arg1;
		    cc = (struct Connection *)fh->fh_Arg1;
		    cn = 0;
		    while(cnx[cn] != cc && cn<MAXCNX)
			cn++;
		    if(!cc->reader)
		    {
			cc->reader=mypkt->dp_Message.mn_ReplyPort->mp_SigTask;
		    }
		    /* we have a pending write */
		    if(cc->writepkt)
		    {
			if(cc->writepkt2)
			{
			    returnpkt(mypkt, DOS_FALSE, mypkt->dp_Result2);
			}
			else
			{
			    cc->writepkt2 = mypkt;
			    mypkt=NULL;
			}
			break;
		    }
		    if(write_ser(cc, (UBYTE *)mypkt->dp_Arg2,
				 (int)mypkt->dp_Arg3))
		    {
			/* Write finished already */
			returnpkt(mypkt, mypkt->dp_Arg3, mypkt->dp_Result2);
		    }
		    else
		    {
			/* Asynchronous write */
			cc->aux_stat |= AUX_CMD_WRITE;
			cc->writepkt = mypkt;
		    }
		    break;
		    
		case ACTION_WAITFORCHAR: /* just queue to wait for data */
		    fh = (struct FileHandle *)mypkt->dp_Arg1;
		    cc = (struct Connection *)fh->fh_Arg1;
		    cn = 0;
		    while(cnx[cn] != cc && cn<MAXCNX)
			cn++;
		    if(!cc->reader)
		    {
			cc->reader=mypkt->dp_Message.mn_ReplyPort->mp_SigTask;
		    }
		    if(cc->waitpkt)	/* Someone already waiting.. */
		    {
			returnpkt(mypkt, DOS_TRUE, 15);
			break;
		    }
		    
		    cc->waitpkt = mypkt;
		    cc->ser_len = query_ser(cc);
		    
		    /* if we have something, reply immediately */
		    if(cc->in_len+cc->ser_len)
		    {
			returnpkt(cc->waitpkt, DOS_TRUE,
				  cc->in_len+cc->ser_len);
			cc->waitpkt = NULL;
			break;
		    }
		    /* if there is no wait, reply immediately */
		    if(!cc->waitpkt->dp_Arg2)
		    {
			returnpkt(cc->waitpkt, DOS_FALSE,
				  cc->waitpkt->dp_Result2);
			cc->waitpkt = NULL;
			break;
		    }
		    
		    cc->Timer->tr_Node.ln_Type = NT_DELAY;/* Signal only */
		    cc->Timer->tr_Task = FindTask(NULL);
		    cc->Timer->tr_Ticks = cc->waitpkt->dp_Arg2;
		    SendTimerReq(cc->Timer);
		    set_read(cc);
		    break;
		    
		case ACTION_SCREEN_MODE:
		    fh = (struct FileHandle *)mypkt->dp_Arg1;
		    cc = (struct Connection *)fh->fh_Arg1;
		    cn = 0;
		    while(cnx[cn] != cc && cn<MAXCNX)
			cn++;
		    if(!cc->reader)
		    {
			cc->reader=mypkt->dp_Message.mn_ReplyPort->mp_SigTask;
		    }
		    cc->aux_stat |= AUX_ECHO;
		    cc->Mode = MODE_CONSOLE;
		    /* -1 means 'normal' Raw mode */
		    if(mypkt->dp_Arg2 == -1)
		    {
			cc->Mode = MODE_RAW;
			cc->aux_stat |= AUX_CRLF;
			cc->aux_stat &= ~AUX_ECHO;
		    }
		    else
		    {
			/* raw means RAW and NOECHO */
			if(mypkt->dp_Arg2 & MODE_RAW)
			{
			    cc->aux_stat &= ~AUX_ECHO;
			    cc->Mode = MODE_RAW;
			}
			/* NOCRLF possible in console mode too */
			if(mypkt->dp_Arg2 & MODE_NOCRLF)
			{
			    cc->aux_stat &= ~AUX_CRLF;
			}
			else
			{
			    cc->aux_stat |= AUX_CRLF;
			}
		    }
		    returnpkt(mypkt, DOS_TRUE, mypkt->dp_Result2);
		    /* Start reading again if we changed to cooked mode */
		    if(cc->Mode == MODE_CONSOLE &&
		       !(cc->aux_stat & AUX_CMD_READ))
			set_read(cc);
		    break;
		    
		case ACTION_CLOSE:
		    fh = (struct FileHandle *)mypkt->dp_Arg1;
		    cc = (struct Connection *)fh->fh_Arg1;
		    cn = 0;
		    while(cnx[cn] != cc && cn<MAXCNX)
			cn++;
		    /* The reader closing ? */
		    if(cc->reader==mypkt->dp_Message.mn_ReplyPort->mp_SigTask)
			cc->reader = (struct Task *)NULL;
		    
		    /* Decrease the open count. Any users left? NO! */
		    if(--cc->aux_open == 0)
		    {
			/* Close the serial and free ports etc. */
			close_ser(cc);
			
			/* TODO: return possible requests */
			FreeMem(cc, sizeof(struct Connection));
			cc = (struct Connection *)NULL;
			if(cn>=0 && cn<MAXCNX)
			    cnx[cn] = (struct Connection *)NULL;
			cnxs--;

			if(!cnxs)
			    run = FALSE;
		    }
		    FreeMem(fh, sizeof(struct FileHandle));
		    returnpkt(mypkt, DOS_TRUE, mypkt->dp_Result2);
		    break;
		    
		default:
		    returnpkt(mypkt, DOS_FALSE, ERROR_ACTION_NOT_KNOWN);
		    break;
		}
	    }
	}
    }
    /* Delete shared replyports before exiting.. */

    DeletePort(tempreply);
    DeletePort(writereply);
    return 0;
}


/* Return the read request */
int ret_readpacket(struct Connection *cc, struct Process *myproc)
{
    UBYTE *ptr;
    int i, tmp;
    
    if(cc->Mode == MODE_RAW)
    {
	/*  Ok, nothing in our buffer, nothing in the serial buffer, 
	    so just wait for the next byte to arrive */
	if((cc->in_len + cc->ser_len) == 0)
	{
	    set_read(cc);
	    return 0;
	}
	
	/* We have less than wanted! */
	if(cc->in_len < cc->rdpkt->dp_Arg3)
	{
	    if(cc->in_len)
		memcpy((UBYTE *)cc->rdpkt->dp_Arg2, cc->auxbuf, cc->in_len);
	    i = cc->in_len;
	    cc->aux_avail = cc->in_len = 0;
	    
	    if(cc->ser_len)
	    {
		/* How much still fits ? */
		tmp = cc->rdpkt->dp_Arg3-i;
		
		/* serial buffer has less than needed */
		if(cc->ser_len < tmp)
		{
		    tmp = cc->ser_len;
		    get_ser(cc, (UBYTE *)cc->rdpkt->dp_Arg2+i, tmp);
		    cc->ser_len = 0;
		}
		else
		{
		    get_ser(cc, (UBYTE *)cc->rdpkt->dp_Arg2+i, tmp);
		    cc->ser_len -= tmp;
		}
		i += tmp;
	    }
	}
	else	/* Enough in buffer to satisfy the request */
	{
	    i = cc->rdpkt->dp_Arg3;
	    memcpy((UBYTE *)cc->rdpkt->dp_Arg2, cc->auxbuf, i);
	    
	    cc->in_len -= i;
	    memmove(cc->auxbuf, cc->auxbuf+i, cc->in_len);
	    cc->aux_avail = tmp = 0;
	    while(tmp<cc->in_len)
		if(cc->auxbuf[tmp++] == '\n')
		    cc->aux_avail++;
	}
	returnpkt(cc->rdpkt, (long)i, cc->rdpkt->dp_Result2);
	cc->rdpkt = NULL;
	return -1;
    }
    else
    {
	/* Line waiting ? */
	if(cc->aux_avail)
	{
	    cc->aux_avail--;
	    i = 0;
	    ptr = (UBYTE *)cc->rdpkt->dp_Arg2;
	    while(i<cc->rdpkt->dp_Arg3 && i<cc->in_len)
	    {
		if((ptr[i] = cc->auxbuf[i]) == '\n')
		{
		    i++;
		    break;
		}
		i++;
	    }
		
	    cc->in_len -= i;
	    memmove(cc->auxbuf, cc->auxbuf+i, cc->in_len);
	    
	    if(ptr[i-1] != '\n')
		cc->aux_avail++;
	    
	    if(!(cc->aux_stat & AUX_CMD_READ) && cc->in_len<MAXLINESIZE-1)
	    {
		set_read(cc);	/* start waiting for reads again */
	    }
	    returnpkt(cc->rdpkt, (long)i, cc->rdpkt->dp_Result2);
	    cc->rdpkt = NULL;
	    return -1;
	}
    }
    return 0;
}

/*  Start an asynchronous Read request */
void set_read(struct Connection *cnx)
{
    if(!(cnx->aux_stat & AUX_CMD_READ))
    {
	/* Clear the signal, so that we don't misinterpret it */
	SetSignal(0, (1<<cnx->ReadSER->io_Message.mn_ReplyPort->mp_SigBit));
	
	cnx->ReadSER->io_Command= CMD_READ;
	cnx->ReadSER->io_Data	= (APTR)&cnx->in_c;
	cnx->ReadSER->io_Length = 1L;
	SendIO(cnx->ReadSER);
	cnx->aux_stat |= AUX_CMD_READ;
    }
}

int query_ser(struct Connection *cnx)	/* we use a dedicated IORequest */
{
    int wasactive=0;
    
#if 0
    /* TODO: If concurrent requests are not possible, abort read */
    if((cnx->aux_stat & AUX_CMD_READ) && !CheckIO(cnx->ReadSER))
    {
	AbortIO(cnx->ReadSER);
	WaitIO(cnx->ReadSER);
	SetSignal(0, (1L<<cnx->ReadSER->io_Message.mn_ReplyPort->mp_SigBit));
	
	cnx->aux_stat &= ~AUX_CMD_READ;
	wasactive = 1;
    }
#endif
    
    cnx->TempSER->io_Actual = 0; /* Hack for non-QUERY-serial */
    cnx->TempSER->io_Command = SDCMD_QUERY;
    DoIO(cnx->TempSER);
    
    /* TODO: If concurret requests are not possible, restart read */
    if(wasactive)
    {
	set_read(cnx);
    }
    return (int)cnx->TempSER->io_Actual;
}

void get_ser(struct Connection *cnx, UBYTE *destbuf, int size)
{
    cnx->ReadSER->io_Command = CMD_READ;
    cnx->ReadSER->io_Data    = (APTR)destbuf;
    cnx->ReadSER->io_Length  = size;
    DoIO(cnx->ReadSER);

    /* Clear the signal, so that we don't misinterpret it */
    SetSignal(0, (1<<cnx->ReadSER->io_Message.mn_ReplyPort->mp_SigBit));
}


int buf_ser(struct Connection *cnx, UBYTE *s, int num)
{
    UBYTE c;
    int i;
    
    if(cnx->Mode == MODE_RAW)
    {
	/* Copy the characters */
	memcpy(cnx->auxbuf+cnx->in_len, s, num);
	cnx->in_len += num;
	
	/* Check if any breaks or LF's were sent */
	while(num--)
	{
	    switch(*s++)
	    {
	    case 0x03:
		/* ^C typed so immediately send the signal (if it is allowed)*/
		if(cnx->reader)
		    Signal(cnx->reader, SIGF_CTRL_C);
		break;
	    case 0x04:
		/* ^D typed so immediately send the signal (if it is allowed)*/
		if(cnx->reader)
		    Signal(cnx->reader, SIGF_CTRL_D);
		break;
	    case 0x0a:
		cnx->aux_avail++;  /* always done when CR received */
		break;
	    }
	}
    }
    else
    {
	for(i=0;i<num;i++)
	{
	    switch(c = s[i])
	    {
	    case 0x03:
		c = 0;
		if(cnx->reader)
		    Signal(cnx->reader, SIGF_CTRL_C);
		break;
	    case 4:
		c = 0;
		if(cnx->reader)
		    Signal(cnx->reader, SIGF_CTRL_D);
		/* Fall through */
	    case 28:		/* ^\ so wipe out line and force EOF */
		cnx->in_len = c = 0;
		++cnx->aux_avail;
		break;
	    case 13:
		/* CR convert to LF if CRLF turned on */
		if(cnx->aux_stat & AUX_CRLF)
		    c = 0x0a;
		break;
	    case 8:			/* BS */
	    case 127:		/* DEL */
		if(cnx->in_len && cnx->auxbuf[cnx->in_len-1] != 0x0a)
		{
		    --cnx->in_len;
		    puts_ser(cnx, (UBYTE *)"\010 \010", 3);
		}
		c = 0;
		break;
	    case 24: 		/* ^X */
	    case 21:		/* ^U */
		while(cnx->in_len && cnx->auxbuf[cnx->in_len-1] != 0x0a)
		{
		    --cnx->in_len;
		    puts_ser(cnx, (UBYTE *)"\010 \010", 3);
		}
		c = 0;
		break;
	    case 27:		/* <ESC> */
		c = 0;
	    default:
		break;
	    }
	    if(c)
	    {
		cnx->auxbuf[cnx->in_len++] = c;
	    
		if(cnx->aux_stat & AUX_ECHO)
		{
		    if((cnx->aux_stat & AUX_CRLF) && c == 0x0a)
			puts_ser(cnx, (UBYTE *)"\r", 1);
		    puts_ser(cnx, &c, 1);
		}
	    }
	    if(c == 0x0a)
		cnx->aux_avail++;	/* always done when CR received */
	}
    }
    
    if(cnx->in_len >= MAXLINESIZE)
    {
	if(!cnx->aux_avail && (cnx->Mode == MODE_CONSOLE))
	{
	    /* Force a newline so that the line can be finally read.. */
	    cnx->aux_avail++;
	    cnx->auxbuf[cnx->in_len++] = 0x0a;
	}
	return 0;	/* buffer is full */
    }
    return -1;
}

/* Write 'em out one by one converting to CR LF if enabled. */
int write_ser(struct Connection *cnx, UBYTE *buf, int len)
{
    int i, count = 0;
    UBYTE c;

    if((cnx->aux_stat & AUX_CRLF) || (cnx->Mode != MODE_RAW))
    {
	for(i=0;i<len;i++)
	{
	    c = buf[i];
	    if(cnx->aux_stat & AUX_CRLF)
	    {
		if(c == 0x9b)
		{
		    cnx->last = 0x9b;
		    continue;
		}
		if(cnx->last == 0x0a && c == 0x0d) continue;	/* we already sent CR */
		if(cnx->last == 0x9b)
		{
		    cnx->outbuf[count++] = '\033';	/* <csi> is converted to ANSI-format */
		    cnx->outbuf[count++] = '[';
		}
	    }
	    cnx->outbuf[count++] = cnx->last = c;	/* if we didn't have CR after LF, we add it */
	    if(cnx->aux_stat & AUX_CRLF)
	    {
		if(cnx->last == 0x0a)
		    cnx->outbuf[count++] = 0x0d;
	    }
	    if(count+4 > OUTBUFSIZE)	/* 3 bytes may be added in one pass */
	    {
		puts_ser(cnx, cnx->outbuf, count);
		count = 0;
	    }
	}
	if(count)	/* something still to send */
	{
	    /* puts_ser(cnx, cnx->outbuf, count); */
	    
	    cnx->WriteSER->io_Command = CMD_WRITE;
	    cnx->WriteSER->io_Length = count;
	    cnx->WriteSER->io_Data = (ULONG)cnx->outbuf;
	    SendIO(cnx->WriteSER);
	    
	    return 0;	/* Not finished */
	}
	return 1;		/* Finished */
    }
    else
    {
	/* puts_ser(cnx, buf, len);	in raw and nocrlf -mode just send the buffer */
	
	cnx->WriteSER->io_Command = CMD_WRITE;
	cnx->WriteSER->io_Length = len;
	cnx->WriteSER->io_Data = (ULONG)buf;
	SendIO(cnx->WriteSER);
	
	cnx->last = buf[len-1];
	return 0;	 	/* Write not finished */
    }
}

/* write whole buffer */
void puts_ser(struct Connection *cnx, UBYTE *buf, int len)
{
    cnx->TempSER->io_Command = CMD_WRITE;
    cnx->TempSER->io_Length = len;
    cnx->TempSER->io_Data = (APTR)buf;
    DoIO(cnx->TempSER);
}


int open_stuff(struct Connection *cnx)
{
    cnx->ReadSER = (struct StdIOReq *)AllocMem(sizeof(struct StdIOReq), MEMF_CLEAR);
    cnx->WriteSER= (struct StdIOReq *)AllocMem(sizeof(struct StdIOReq), MEMF_CLEAR);
    cnx->TempSER = (struct StdIOReq *)AllocMem(sizeof(struct StdIOReq), MEMF_CLEAR);
    
    if(!cnx->ReadSER || !cnx->WriteSER || !cnx->TempSER)
    {
	goto endser;
    }
    
    cnx->ReadSER->io_Message.mn_ReplyPort = (struct MsgPort *)CreatePort(0);
    cnx->ReadSER->io_Data = 38400; /* Speed */
    if(OpenDevice(cnx->devicename, cnx->deviceunit, cnx->ReadSER, 0))
    {
	cnx->ReadSER->io_Device = NULL;
	goto endser;
    }
    
    cnx->ReadSER->io_Command = CMD_SETPARAMS;
    /* if speed is set in the filename part */
    if(cnx->devicespeed != -1)
    {
	cnx->ReadSER->io_Length = (ULONG)cnx->devicespeed;
    }
    cnx->ReadSER->io_Data = 0; /* Shared mode, nothing extra */
    DoIO(cnx->ReadSER);	/* if it didn't work, let the user worry :-) */
    
    /* Make copies of the structs (after the SETPARAMS!) */
    memcpy((UBYTE *)cnx->WriteSER, (UBYTE *)cnx->ReadSER, sizeof(struct StdIOReq));
    cnx->WriteSER->io_Message.mn_ReplyPort = writereply;
    
    memcpy((UBYTE *)cnx->TempSER, (UBYTE *)cnx->ReadSER, sizeof(struct StdIOReq));
    cnx->TempSER->io_Message.mn_ReplyPort = tempreply;

    if((cnx->Timer = (struct TimerReq *)AllocMem(sizeof(struct TimerReq), MEMF_CLEAR)))
    {
	set_read(cnx);	/* start reading immediately */
	return 1;
    }
 endser:
    close_ser(cnx);
    return 0;
}

void close_ser(struct Connection *cnx)
{
    /* Read pending? YES, abort it */
    if((cnx->aux_stat & AUX_CMD_READ))
    {
	AbortIO(cnx->ReadSER);
	WaitIO(cnx->ReadSER);
    }
    if(cnx->TempSER && cnx->TempSER->io_Device)
    {
	cnx->TempSER->io_Length = 38400;
	cnx->TempSER->io_Data = 6; /* Shared mode, shortret, cooked */
	cnx->TempSER->io_Command = CMD_SETPARAMS;
	DoIO(cnx->TempSER);	/* if it didn't work, let the user worry :-) */
    }

    /* Write pending? YES, wait for it to complete (TODO:LOCKUP possibility?) */
    if((cnx->aux_stat & AUX_CMD_WRITE))
    {
	/* AbortIO(cnx->WriteSER); */
	WaitIO(cnx->WriteSER);
    }
    if(cnx->ReadSER)
    {
	if(cnx->ReadSER->io_Device)
	    CloseDevice(cnx->ReadSER);
	if(cnx->ReadSER->io_Message.mn_ReplyPort)
	    DeletePort(cnx->ReadSER->io_Message.mn_ReplyPort);
	FreeMem(cnx->ReadSER, sizeof(struct StdIOReq));
	cnx->ReadSER = NULL;
    }
    if(cnx->WriteSER)
    {
	FreeMem(cnx->WriteSER, sizeof(struct StdIOReq));
	cnx->WriteSER = NULL;
    }
    if(cnx->TempSER)
    {
	FreeMem(cnx->TempSER, sizeof(struct StdIOReq));
	cnx->TempSER = NULL;
    }
}


void chk_file(char *name, struct Connection *cnx)
{
    char *s = name;

    strcpy(cnx->devicename, SERIALNAME);	/* set the defaults */
    cnx->deviceunit = 0;
    cnx->aux_stat = AUX_CRLF | AUX_ECHO;
    cnx->devicespeed = -1;
    cnx->Mode = MODE_CONSOLE;
    
    if(!*s)
	return;	/* no params */
    
    while(*s && *s != '/')
	s++;
    if(!*s)
    {
	strcpy(cnx->devicename, name);
	return;	/* no unit number */
    }
    
    cnx->deviceunit = atoi(s+1);
    strncpy(cnx->devicename, name, (long)(s-name));
    cnx->devicename[(long)(s-name)] = '\0';
    
    s++;
    
    while(1)
    {
	while(*s && *s != '/')
	    s++;
	if(!*s)
	    break;
	++s;
	if(!strncmp(s, "raw", 3))
	{
	    cnx->aux_stat &= ~AUX_ECHO;
	    cnx->Mode = MODE_RAW;
	    s += 3;
	}
	else if(!strncmp(s, "con", 3))
	{
	    cnx->aux_stat |= AUX_ECHO;
	    cnx->Mode = MODE_CONSOLE;
	    s += 3;
	}
	else if(!strncmp(s, "speed=", 6))
	{
	    cnx->devicespeed = atol(s += 6);
	}
	else if(!strncmp(s, "speed=", 5))
	{
	    cnx->devicespeed = atol(s += 5);
	}
    }
}

void returnpkt(struct DosPacket *packet, ULONG res, ULONG res2)
{
    packet->dp_Result = res;
    packet->dp_Result2 = res2;
    ReplyMsg((struct Message *)packet);
}
