#include <exec/exec_protos.h>
#include <dos/dos.h>
#include <dos/dos_protos.h>

#include <string.h>

LONG handler(void);

const UWORD kludgejump = 0x4ef9;
const ULONG kludgeaddr = (ULONG)handler;

static int launched = 0;
struct DOSLibrary *DOSBase = NULL;

#define ENTRY_DIR   1
#define ENTRY_FILE -1

int IsDirectory(const char *name)
{
    if(!*name)
	return ENTRY_DIR;
    while(*++name);
    if(name[-1]==':' || name[-1]=='/')
	return ENTRY_DIR;
    return ENTRY_FILE;
}


LONG handler()
{
    struct Process *myself = (struct Process *)FindTask(NULL);
    struct DosList *dol;
    struct DosPacket *packet;
    int done = 0;
    struct MsgPort *myport;

    /* Detaching */
    if(!launched)
    {
	if((DOSBase = (struct DOSLibrary *)OpenLibrary(DOSL_NAME)))
	{
	    struct SegList *segl = (struct SegList *)myself->pr_SegList;

	    myself->pr_SegList = NULL;
	    launched = 1;
	    if(CreateProc("NIL:", 10, segl, 1024, NULL ))
		return RETURN_OK;
	    CloseLibrary((struct Library *)DOSBase);
	}
	return RETURN_FAIL;
    }
    /* Ok, we are now running freely.. */

    Forbid();
    if(FindName(&DOSBase->dl_Handlers, (UBYTE *)"NIL:") ||
       !(myport = CreatePort(NULL)))
    {
	Permit();
	return RETURN_FAIL;
    }

    /* First create a devicenode */
    if((dol = (struct DosList *)AllocMem(sizeof(struct DosList), MEMF_CLEAR)))
    {
	ULONG opens = 0;

	dol->dol_Task = myport;
	dol->dol_Node.ln_Name = "NIL:";
	dol->dol_Node.ln_Type = DLT_HANDLER;

	dol->dol_misc.dol_handler.dol_SegList =
	    (struct SegList *)myself->pr_SegList;

	/* add to dos list ! */
	AddTail(&DOSBase->dl_Handlers, (struct Node *)dol);
	Permit();

	while(!done || opens)
	{
	    struct FileHandle *fh;
	    struct FileLock *fl;
	    struct FileInfoBlock *fib;

	    WaitPort(myport);
	    while((packet = (struct DosPacket *)GetMsg(myport)))
	    {
		packet->dp_Result = DOS_FALSE;
		switch(packet->dp_Type)
		{
		case ACTION_OPEN:
		    fh = (struct FileHandle *)
			AllocMem(sizeof(struct FileHandle), MEMF_CLEAR);
		    if(fh)
		    {
			fh->fh_Port = myport;
			fh->fh_Buf  = NULL; /* does not point to a buffer */
			fh->fh_Pos  = 0;
			opens++;
		    }
		    else
		    {
			packet->dp_Result2 = ERROR_NO_FREE_STORE;
		    }
		    packet->dp_Result = (LONG)fh;
		    break;

		case ACTION_READ:
		    break;

		case ACTION_WRITE:
		    packet->dp_Result = packet->dp_Arg3;
		    break;

		case ACTION_LOCK:
		    fl = (struct FileLock *)
			AllocMem(sizeof(struct FileLock), MEMF_CLEAR);
		    if(fl)
		    {
			fl->fl_Key = IsDirectory((char *)packet->dp_Arg2);
			fl->fl_Access = (LONG)packet->dp_Arg3;
			fl->fl_Task = myport;
			opens++;
		    }
		    else
		    {
			packet->dp_Result2 = ERROR_NO_FREE_STORE;
		    }
		    packet->dp_Result = (LONG)fl;
		    break;

		case ACTION_EXAMINE:
		    fl = (struct FileLock *)packet->dp_Arg1;
		    fib = (struct FileInfoBlock *)packet->dp_Arg2;
		    fib->fib_DirEntryType = fl->fl_Key;
		    fib->fib_OwnerID = fib->fib_GroupID = fib->fib_Size = 0;
		    DateStamp(&fib->fib_Date);
		    strcpy(fib->fib_Comment, "Entry generated by the handler");
		    if(fl->fl_Key == ENTRY_FILE)
		    {
			strcpy(fib->fib_FileName, "null-file");
			fib->fib_Protection = 0xddd;
		    }
		    else
		    {
			strcpy(fib->fib_FileName, "null-dir");
			fib->fib_Protection = 0xfff;
		    }
		    packet->dp_Result = DOS_TRUE;
		    break;

		case ACTION_EXNEXT:
		    packet->dp_Result2 = ERROR_NO_MORE_ENTRIES;
		    break;

		case ACTION_UNLOCK:
		    FreeMem((char *)packet->dp_Arg1, sizeof(struct FileLock));
		    opens--;
		    packet->dp_Result = DOS_TRUE;
		    break;

		case ACTION_CLOSE:
		    FreeMem((char *)packet->dp_Arg1, sizeof(struct FileHandle));
		    opens--;
		    /* Fall through! */

		case ACTION_EXAMINE_END:
		case ACTION_DELETE:
		    packet->dp_Result = DOS_TRUE;
		    break;

		case ACTION_DIE:
		    if(!done)
		    {
			done = 1;
			Remove((struct Node *)dol); /* No new accesses */
		    }
		    break;

		default:
		    packet->dp_Result2 = ERROR_ACTION_NOT_KNOWN;
		}
		ReplyMsg((struct Message *)packet);
	    }
	}
	FreeMem(dol, sizeof(struct DosList));
    }
    else
    {
	Permit();
    }
    DeletePort(myport);
    CloseLibrary((struct Library *)DOSBase);
    return 0;
}

