#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define ULONG unsigned long
#define LONG signed long

struct BootBlock
{
    ULONG bb_DiskType;
    ULONG bb_CheckSum;
    ULONG bb_LowBlock;
    ULONG bb_HiBlock;
    ULONG bb_BlockSize;
    ULONG bb_RootBlock;
};

struct BitMapBlock
{
    LONG  bmb_Type;      /* 128 */
    ULONG bmb_CheckSum;  /* BitMap checksum */
    ULONG bmb_Extension; /* Extension of the bitmap or NULL */
    ULONG bmb_StartBlock;/* Block number for bit 0 in BitMap[0] */
    ULONG bmb_BitMap[60];/* for 256-byte blocksize */
};

struct BootBlock *bb;
struct BitMapBlock *bmb;

LONG AllocBlock(struct BitMapBlock *bmb, int type);
void FreeBlock(struct BitMapBlock *bmb, ULONG block);

LONG AllocBlock(struct BitMapBlock *bmb, int down)
{
    ULONG entries = (bb->bb_BlockSize/4)-4;
    ULONG poslong;
    ULONG posval,posbit;

    if(down)
    {
	poslong = (bb->bb_RootBlock-bmb->bmb_StartBlock)/32-1;

	while(poslong && !bmb->bmb_BitMap[poslong])
	    poslong--;
printf("poslong: %d, bmb->bmb_BitMap[poslong]: 0x%08x\n",
       poslong, bmb->bmb_BitMap[poslong]);
	if(!bmb->bmb_BitMap[poslong])
	{
	    poslong = (bb->bb_HiBlock-bmb->bmb_StartBlock)/32;
	    while(poslong && !bmb->bmb_BitMap[poslong])
		poslong--;
	    if(!bmb->bmb_BitMap[poslong])
		return -1;
	}
	posval = bmb->bmb_BitMap[poslong];
	posbit = 31;
	while(!(posval&(1U<<31)))
	{
	    posbit--;
	    posval <<= 1;
	}
    }
    else
    {
	poslong = (bb->bb_RootBlock-bmb->bmb_StartBlock)/32;
	if(entries > (bb->bb_HiBlock-bmb->bmb_StartBlock)/32 + 1)
	    entries = (bb->bb_HiBlock-bmb->bmb_StartBlock)/32 + 1;

	while(poslong<entries && !bmb->bmb_BitMap[poslong])
	    poslong++;
printf("entries: %d poslong: %d, bmb->bmb_BitMap[poslong]: 0x%08x\n",
       entries, poslong, bmb->bmb_BitMap[poslong]);
	if(poslong==entries)
	{
	    poslong = (bb->bb_RootBlock-bmb->bmb_StartBlock)/32 - 1;
	    while(poslong && !bmb->bmb_BitMap[poslong])
		poslong--;
	    if(!bmb->bmb_BitMap[poslong])
		return -1;

	    posval = bmb->bmb_BitMap[poslong];
	    posbit = 31;
	    while(!(posval&(1U<<31)))
	    {
		posbit--;
		posval <<= 1;
	    }
	}
	else
	{
	    posval = bmb->bmb_BitMap[poslong];
	    posbit = 0;
	    while(!(posval&1))
	    {
		posbit++;
		posval >>= 1;
	    }
	}
    }
    bmb->bmb_BitMap[poslong] &= ~(1U<<posbit);
printf("BitMap[%ld] = %08x\n", poslong, bmb->bmb_BitMap[poslong]);
    return bmb->bmb_StartBlock + poslong*32 + posbit;
}

void FreeBlock(struct BitMapBlock *bmb, ULONG block)
{
    ULONG entries = (bb->bb_BlockSize/4)-4;
    ULONG poslong = (block-bmb->bmb_StartBlock)/32;
    ULONG posbit  = (block-bmb->bmb_StartBlock)%32;

    if(block < bb->bb_LowBlock || block > bb->bb_HiBlock)
	return;

    if(poslong > entries)
    {
	if(bmb->bmb_Extension)
	    FreeBlock((struct BitMapBlock *)bmb->bmb_Extension, block);
	return;
    }
    bmb->bmb_BitMap[poslong] |= (1U<<posbit);
printf("BitMap[%ld] = %08x\n", poslong, bmb->bmb_BitMap[poslong]);
}


int main(int argc, char *argv[])
{
    int i;
    ULONG poslong, posbit;

    bb = (struct BootBlock *)calloc(1U, 0x400);
    bmb = (struct BitMapBlock *)calloc(1U, 0x400);

    bb->bb_LowBlock = 0 /*0x840000/0x400*/;
    bb->bb_HiBlock =  120-1 /*0xc00000/0x400-1*/;
    bb->bb_RootBlock = (bb->bb_LowBlock+bb->bb_HiBlock)/2+1;
    bb->bb_BlockSize = 0x400;

    memset(bmb, 0xff, sizeof(struct BitMapBlock));

    bmb->bmb_Type = 128;
    bmb->bmb_Extension = 0;
    bmb->bmb_StartBlock = bb->bb_LowBlock;
    /* Reserve RootBlock & BitMapBlock */
    bmb->bmb_BitMap[(bb->bb_RootBlock-bb->bb_LowBlock)/32] &= ~(3U<<((bb->bb_RootBlock-bb->bb_LowBlock)%32));
    poslong = (bb->bb_HiBlock-bb->bb_LowBlock)/32;
    bmb->bmb_BitMap[poslong] = 0;
    posbit = 0;
    while(bmb->bmb_StartBlock + poslong*32 + posbit <= bb->bb_HiBlock)
    {
	bmb->bmb_BitMap[poslong] |= (1U<<posbit);
	posbit++;
    }

    printf("%ld,%ld,%ld\n",
	   bb->bb_LowBlock, bb->bb_RootBlock, bb->bb_HiBlock);

    for(i=0;i<70;i++)
	printf("Allocated: %ld\n", AllocBlock(bmb, 0));

    printf("Changing direction\n");
    for(i=0;i<50;i++)
	printf("Allocated: %ld\n", AllocBlock(bmb, 1));


    FreeBlock(bmb,54);
    printf("Allocated: %ld (54?)\n", AllocBlock(bmb, 0));

    FreeBlock(bmb,65);
    printf("Allocated: %ld (65?)\n", AllocBlock(bmb, 0));

    return 0;
}

