#ifndef DOS_DEFINES_H
#define DOS_DEFINES_H

#define DOS_TRUE  (-1)
#define DOS_FALSE (0)

/* Mode parameter to SetMode() */
/* Console (cooked) mode has line-editing and echo (line-buffered),
   raw-mode passes all received characters to the reader (non-buffered),
   it NOCRLF is set, LF is not appended for sent CR's
 */
#define MODE_CONSOLE	0	/* normal buffered console mode */
#define MODE_RAW	1	/* raw mode, no line editing nor echo */
#define MODE_NOCRLF	2	/* CR not added to LF, all codes sent */

/* Mode parameter to Open() */
/* These flags define what kind of access the process should have for the
   Open() to succeed. The mode can also be used to remember what actions
   (Read/Write) are possible on a filehandle.
 */
#define MODE_READ      (1<<0) /* Need read access to open */
#define MODE_WRITE     (1<<1) /* Need write access to open */
#define MODE_EXECUTE   (1<<2) /* Need execute access to open */
#define MODE_DELETE    (1<<3) /* Delete if exists (and needs delete access)*/
#define MODE_CREATE    (1<<4) /* Create if doesn't exist */

#define MODE_OLDFILE     (MODE_READ | MODE_WRITE)
 /* Open existing file read/write (shared lock) */
#define MODE_NEWFILE     (MODE_READ | MODE_WRITE | MODE_DELETE | MODE_CREATE)
 /* Open freshly created file (delete old file) read/write, exclusive lock. */
#define MODE_READWRITE   (MODE_READ | MODE_WRITE | MODE_CREATE)
 /* Open old file w/shared lock, creates file if doesn't exist. */


/* Relative position to Seek() */
#define OFFSET_BEGINNING    -1      /* relative to Beginning Of File */
#define OFFSET_CURRENT       0      /* relative to Current file position */
#define OFFSET_END           1      /* relative to End Of File    */

/* Passed as type to Lock() */
#define SHARED_LOCK          -2     /* File is readable by others */
#define ACCESS_READ          -2     /* Synonym */
#define EXCLUSIVE_LOCK       -1     /* No other access allowed    */
#define ACCESS_WRITE         -1     /* Synonym */

/* These are the return codes used by convention by AmigaDOS commands */
/* See FAILAT and IF for relvance to EXECUTE files                    */
#define RETURN_OK                           0  /* No problems, success */
#define RETURN_WARN                         5  /* A warning only */
#define RETURN_ERROR                       10  /* Something wrong */
#define RETURN_FAIL                        20  /* Complete or severe failure*/

/* Values returned by SameLock() */
#define LOCK_DIFFERENT          -1
#define LOCK_SAME               0
#define LOCK_SAME_VOLUME        1       /* locks are on same volume */
#define LOCK_SAME_HANDLER       LOCK_SAME_VOLUME
/* LOCK_SAME_HANDLER was a misleading name, def kept for src compatibility */

/* Errors from IoErr(), etc. */
#define ERROR_NO_FREE_STORE               100
#define ERROR_NO_DEFAULT_DIR              201
#define ERROR_OBJECT_IN_USE               202
#define ERROR_OBJECT_EXISTS               203
#define ERROR_DIR_NOT_FOUND               204
#define ERROR_OBJECT_NOT_FOUND            205
#define ERROR_BAD_STREAM_NAME             206
#define ERROR_OBJECT_TOO_LARGE            207
#define ERROR_ACTION_NOT_KNOWN            209
#define ERROR_INVALID_COMPONENT_NAME      210
#define ERROR_INVALID_LOCK                211
#define ERROR_OBJECT_WRONG_TYPE           212
#define ERROR_DISK_NOT_VALIDATED          213
#define ERROR_DISK_WRITE_PROTECTED        214
#define ERROR_RENAME_ACROSS_DEVICES       215
#define ERROR_DIRECTORY_NOT_EMPTY         216
#define ERROR_TOO_MANY_LEVELS             217
#define ERROR_DEVICE_NOT_MOUNTED          218
#define ERROR_SEEK_ERROR                  219
#define ERROR_COMMENT_TOO_BIG             220
#define ERROR_DISK_FULL                   221
#define ERROR_DELETE_PROTECTED            222
#define ERROR_WRITE_PROTECTED             223
#define ERROR_READ_PROTECTED              224
#define ERROR_EXECUTE_PROTECTED           225
#define ERROR_NO_DISK                     226
#define ERROR_NO_MORE_ENTRIES             232
#define ERROR_IS_SOFT_LINK                233
#define ERROR_OBJECT_LINKED               234
#define ERROR_BAD_HUNK                    235
#define ERROR_NOT_IMPLEMENTED             236
#define ERROR_NOT_A_DOS_DISK              237
#define ERROR_PARENT_PROTECTED            238
#define ERROR_NOT_OWNER                   239

/* Packet types */
#define ACTION_NIL              0
#define ACTION_STARTUP          0
#define ACTION_DIE              2
#define ACTION_LOCK             3
#define ACTION_UNLOCK           4
#define ACTION_DUPLOCK          5
#define ACTION_CURRENTDIR       6
#define ACTION_RELABEL          9
#define ACTION_WRITE            10
#define ACTION_READ             11
#define ACTION_DELETE           16
#define ACTION_RENAME           17
#define ACTION_MORE_CACHE       18
#define ACTION_COPY_DIR         19
#define ACTION_WAITFORCHAR      20
#define ACTION_SETPROTECTION    21
#define ACTION_SETCOMMENT       22
#define ACTION_SETOWNER         23
#define ACTION_CREATE_DIR       24
#define ACTION_EXAMINE          25
#define ACTION_EXNEXT           26
#define ACTION_EXAMINE_END      27
#define ACTION_EXAMINE_FH       28
#define ACTION_DISKINFO         29
#define ACTION_INFO             30
#define ACTION_FLUSH            31
#define ACTION_PARENT           32
#define ACTION_TIMER            33
#define ACTION_INHIBIT          34
#define ACTION_DISK_TYPE        35
#define ACTION_DISK_CHANGE      36
#define ACTION_SET_DATE         37
#define ACTION_SCREEN_MODE      38
#define ACTION_SAME_LOCK        40

#define ACTION_SEEK             1008
#define ACTION_OPEN             1004
#define ACTION_CLOSE            1007
#define ACTION_FORMAT           1020
#define ACTION_SET_FILE_SIZE    1022
#define ACTION_WRITE_PROTECT    1023 /* Write-protect volume */

#define ACTION_CHANGE_SIGNAL    995
#define ACTION_IS_FILESYSTEM    1027
#define ACTION_COPY_DIR_FH      1030
#define ACTION_PARENT_FH        1031

/* FIBB are bit definitions, FIBF are field definitions */
/* RWED bits are 1 == allowed. */
#define FIBB_SCRIPT    14       /* program is a script (execute) file */
#define FIBB_PURE      13       /* program is reentrant and rexecutable */
#define FIBB_ARCHIVE   12       /* cleared whenever file is changed */

#define FIBB_OTR_DELETE    11   /* Other: prevent file from being deleted */
#define FIBB_OTR_READ      10   /* Other: file is readable */
#define FIBB_OTR_WRITE      9   /* Other: file is writable */
#define FIBB_OTR_EXECUTE    8   /* Other: file is executable */
#define FIBB_GRP_DELETE     7   /* Group: prevent file from being deleted */
#define FIBB_GRP_READ       6   /* Group: file is readable */
#define FIBB_GRP_WRITE      5   /* Group: file is writable */
#define FIBB_GRP_EXECUTE    4   /* Group: file is executable */

#define FIBB_DELETE    3        /* prevent file from being deleted */
#define FIBB_READ      2        /* ignored by old filesystem */
#define FIBB_WRITE     1        /* ignored by old filesystem */
#define FIBB_EXECUTE   0        /* ignored by system, used by Shell */

#define FIBF_OTR_DELETE    (1<<FIBB_OTR_DELETE)
#define FIBF_OTR_READ      (1<<FIBB_OTR_READ)
#define FIBF_OTR_WRITE     (1<<FIBB_OTR_WRITE)
#define FIBF_OTR_EXECUTE   (1<<FIBB_OTR_EXECUTE)
#define FIBF_GRP_DELETE    (1<<FIBB_GRP_DELETE)
#define FIBF_GRP_READ      (1<<FIBB_GRP_READ)
#define FIBF_GRP_WRITE     (1<<FIBB_GRP_WRITE)
#define FIBF_GRP_EXECUTE   (1<<FIBB_GRP_EXECUTE)

#define FIBF_SCRIPT    (1<<FIBB_SCRIPT)
#define FIBF_PURE      (1<<FIBB_PURE)
#define FIBF_ARCHIVE   (1<<FIBB_ARCHIVE)
#define FIBF_DELETE    (1<<FIBB_DELETE)
#define FIBF_READ      (1<<FIBB_READ)
#define FIBF_WRITE     (1<<FIBB_WRITE)
#define FIBF_EXECUTE   (1<<FIBB_EXECUTE)

/* ID stands for InfoData */
        /* Disk states */
#define ID_WRITE_PROTECTED 80    /* Disk is write protected */
#define ID_VALIDATING      81    /* Disk is currently being validated */
#define ID_VALIDATED       82    /* Disk is consistent and writeable */

        /* Disk types */
#define ID_NO_DISK_PRESENT      (-1)
#define ID_UNREADABLE_DISK      (0x42414400L)   /* 'BAD\0' */
#define ID_DOS_DISK             (0x444F5300L)   /* 'DOS\0' */
#define ID_NOT_REALLY_DOS       (0x4E444F53L)   /* 'NDOS'  */

#endif
