#ifndef EXEC_PROTOS_H
#define EXEC_PROTOS_H

#include <exec/types.h>
#include <exec/memory.h>
#include <exec/execbase.h>
#include <exec/tasks.h>

/*------ interrupts ----------------------------------------------------*/
void	Disable(void);
void	Enable(void);
void	Forbid(void);
void	Permit(void);
void	AddIntServer(long /* D0 */, struct Interrupt * /* A1 */ );
void	RemIntServer(long /* D0 */, struct Interrupt * /* A1 */);
void	Cause(long vector /* D0 */);
/*------ memory allocation: ---------------------------------------------*/
void	*AllocMem(ULONG size /* D0 */, ULONG flags /* D1 */);
void	FreeMem(void *chunk /* A0 */, ULONG size /* D0 */);
ULONG	AvailMem(ULONG flags /* D0 */);
void    *AllocRemember(struct List *memlist /* A0 */, ULONG size /* D0 */, ULONG flags /* D1 */);
void    FreeRemember(void *chunk /* A0 */);
void    FreeRememberAll(struct List *memlist /* A0 */);
/* List routines in lists.h */
/*------ tasks: ---------------------------------------------------------*/
/* tc_Stack,tc_StackSize,tc_UnLoad should be initialized prior to AddTask() */
void	AddTask(struct Task * /*A1*/,long InitPC /*D0*/,long FinalPC /*D1*/);
void    RemTask(struct Task * /*A1*/);
struct Task *FindTask(char *name /* A1 */);
WORD	SetTaskPri(struct Task * /* A0 */, WORD /* D1 */);
ULONG	SetSignal(ULONG newsignals /* D0 */, ULONG sigmask /* D1 */);
ULONG	Wait(ULONG sigmask /* D0 */);
void	Signal(struct Task * /* A0 */, ULONG sigmask /* D0 */);
ULONG	AllocSignal(void);
VOID	FreeSignal(ULONG signum /* D0 */);
/*------ 'Resident' programs: -------------------------------------------*/
void    AddProgram(struct Program * /* A1 */);   /* Does nor alloc mem */
struct Program *RemProgram(char *name /* A1 */); /* Does not free mem */
int     FindProgram(char *name /* A1 */);        /* returns 0=not found */
/* Port/message/device routines in io.h */
/*------ libraries: -----------------------------------------------------*/
void    AddLibrary(struct Library * /* A1 */);
struct Library *RemLibrary(struct Library * /* A1 */);
void    InitLibFuncs(struct Library */*A0*/,ULONG *fncs/*A1*/,ULONG num/*D0*/);
struct Library *OpenLibrary(char *name /* A1 */);
void    CloseLibrary(struct Library * /* A1 */);
ULONG   SetFunction(struct Library */*A1*/,LONG offset/*D0*/,ULONG val/*D1*/);
/*------ resources: -----------------------------------------------------*/
void    AddResource(struct Node *);
struct Node *RemResource(struct Node *);
struct Node *AllocResource(char *name);
void    FreeResource(struct Node *);
/*------ other: ---------------------------------------------------------*/
void	DateStamp(struct DateStamp * /* A0 */);
void	SetDate(struct DateStamp * /* A0 */);

#endif

