#ifndef	EXEC_IO_H
#define	EXEC_IO_H

#include <exec/lists.h>
#include <exec/libraries.h>

struct MsgPort
{
    struct  Node mp_Node;
    UBYTE   mp_Flags;
    UBYTE   mp_SigBit;          /* signal bit number    */
    void   *mp_SigTask;         /* object to be signalled */
    struct  List mp_MsgList;    /* message linked list  */
};

#define mp_SoftInt mp_SigTask   /* Alias */

/* mp_Flags: Port arrival actions (PutMsg) */
#define PF_ACTION       3       /* Mask */
#define PA_SIGNAL       0       /* Signal task in mp_SigTask */
#define PA_SOFTINT      1       /* Signal SoftInt in mp_SoftInt/mp_SigTask */
#define PA_IGNORE       2       /* Ignore arrival */


struct Message
{
    struct  Node mn_Node;
    struct  MsgPort *mn_ReplyPort;  /* message reply port */
    UWORD   mn_Length;              /* total message length, in bytes */
};

struct Device
{
    struct Library dd_Library;
};

#define DEV_BEGINIO (-18)
#define DEV_ABORTIO (-24)

struct StdIOReq
{
    struct Message io_Message;
    struct Device *io_Device; /* Filled in by OpenDevice */
    struct Unit *io_Unit; /* Device-private data */
    UWORD io_Command; /* What to do */
    UWORD io_Error;
    LONG io_Actual;  /* Actual written/read data */
    LONG io_Length;  /* Data to write/read */
    LONG io_Data;    /* Pointer to data */
    ULONG io_Offset;  /* Block number for block-oriented devices */
};

#define CMD_INVALID	0
#define CMD_RESET	1
#define CMD_READ	2
#define CMD_WRITE	3
#define CMD_UPDATE	4
#define CMD_CLEAR	5
#define CMD_STOP	6
#define CMD_START	7
#define CMD_FLUSH	8
#define CMD_SETPARAMS   9
#define CMD_WAITFORCHAR 10

struct TimerReq
{
    struct Node tr_Node;
    struct Task *tr_Task;
    ULONG tr_Ticks;
};

/*---------------------------------------------*/
void    AddPort(struct MsgPort *);
void    RemPort(struct MsgPort *);
struct MsgPort *CreatePort(char *name);
void DeletePort(struct MsgPort *port);
struct MsgPort *FindPort(char *name);
void    WaitPort(struct MsgPort *);

void	PutMsg(struct MsgPort *, struct Message *);
struct	Message *GetMsg(struct MsgPort *);
void    ReplyMsg(struct Message *);
void	SendTimerReq(struct TimerReq *);

void AddDevice( struct Device *device );
void RemDevice( struct Device *device );
LONG OpenDevice( char *devName, ULONG unit, struct StdIOReq *req, ULONG flags);
void CloseDevice( struct StdIOReq *req );
LONG DoIO( struct StdIOReq *req );
LONG SendIO( struct StdIOReq *req );
struct StdIOReq *CheckIO( struct StdIOReq *req );
LONG WaitIO( struct StdIOReq *req );
void AbortIO( struct StdIOReq *req );
/*---------------------------------------------*/

#endif	/* EXEC_IO_H */
