#ifndef EXEC_LISTS_H
#define EXEC_LISTS_H

#include <exec/types.h>

/* node types */
#define NT_UNKNOWN      0
#define NT_TASK         1       /* Exec task */
#define NT_TIMER	2
#define NT_DELAY	3	/* send only SIGF_DELAY */
#define NT_DEVICE       4
#define NT_PORT         5
#define NT_MESSAGE      6       /* Indicates message currently pending */
#define NT_FREEMSG      7
#define NT_REPLYMSG     8       /* Message has been replied */
#define NT_RESOURCE     9
#define NT_LIBRARY      10
#define NT_SEMAPHORE    11
#define NT_RESIDENT     12	/* struct Program */
#define NT_STARTUP      13      /* run at startup */

#define NT_USER         254     /* User node types work down from here */
#define NT_EXTENDED     255


struct List
{ 
    struct Node *lh_Head;
    struct Node *lh_Tail;
    struct Node *lh_TailPred;
};

struct Node
{
    struct Node *ln_Next;	/* 0 */
    struct Node *ln_Prev;	/* 4 */
    UBYTE	ln_Type;	/* 8 */
    BYTE	ln_Pri;		/* 9 */
    char        *ln_Name;       /* 10 */
};

struct MinNode
{
    struct MinNode *lmn_Next;
    struct MinNode *lmn_Prev;
};

void Insert( struct List *list, struct Node *node, struct Node *pred );
void AddHead( struct List *list, struct Node *node );
void AddTail( struct List *list, struct Node *node );
void Remove( struct Node *node );
struct Node *RemHead( struct List *list );
struct Node *RemTail( struct List *list );
void Enqueue( struct List *list, struct Node *node );
struct Node *FindName( struct List *list, UBYTE *name );
void NewList( struct List *list );

struct Node *HeadNode(struct List *list);
struct Node *TailNode(struct List *list);
struct Node *NextNode(struct Node *node);
struct Node *PrevNode(struct Node *node);

#endif
