#include "stdio.h"
#include "string.h"
#include "stdlib.h"
#include "ctype.h"
#include <dos/dos.h>
#include <dos/dos_protos.h>
#include <exec/exec_protos.h>

int __main(void);
int ___init(void);

const UWORD __KludgeCommand = 0x4ef9;         /* jmp.l absolute      */
const ULONG __KludgeAddress = (ULONG) ___init;
/* address of the _sp.s function ___init() */


int main(int argc, char **argv);

/* The following two functions are required for exit() to work */
void __exitasm(int result);
void (*__exitfunc)(void) = NULL;

struct DOSLibrary *DOSBase = NULL;
FILE *stdin = NULL, *stdout = NULL, *stderr = NULL;
FILE *stddata;
struct List __openfiles;
struct List _memlist;
unsigned long __starttime;
static struct ArgList *AL = NULL;

static void cleanupmain();

#define MAX_ARG_LEN 256



/*
 *   __init gets called every time this startup module is used.
 */

int __init(void) {
    int result;
    int i;
    char *args;

    /* HERE ARE THE OPERATIONS MADE BEFORE MAIN() IS CALLED */

    if(!(DOSBase = (struct DOSLibrary *)OpenLibrary("dos.library")))
        return RETURN_FAIL;

    /* Initialize time base */
    __starttime = FindTask(NULL)->tc_Elapsed;

    /* Initialize memory allocation list */
    NewList(&_memlist);

    /* initialize open files list */
    NewList(&__openfiles);

    /* Allocate space for standard file handles */
    stddata = calloc(3, sizeof(FILE));
    if (!stddata)
    {
	cleanupmain();
        return RETURN_FAIL;
    }

    /* Setup standard file streams */
    for (i=0; i<3; i++)
    {
	stddata[i].bufsize = __DEFAULT_BUFSIZE;
	stddata[i].flags = __FF_SYSFILE | __FF_SYSBUF;
	stddata[i].bufmode = _IOLBF;
	stddata[i].unc = -1;
	stddata[i].buf = malloc(__DEFAULT_BUFSIZE);
	if (!(stddata[i].buf))
	{
	    cleanupmain();
	    return RETURN_FAIL;
	}
    }
    stddata[0].filehandle = Input();
    stddata[1].filehandle = Output();
    stddata[2].filehandle = Error();
    stdin =  stddata;
    stdout = stddata + 1;
    stderr = stddata + 2;

    /* Parse the arguments */
    args = ((struct Process *) FindTask(NULL))->pr_Arguments;

    AL = ParseArgs(args);
    if (!AL)
    {
	cleanupmain();
	return RETURN_FAIL;
    }

    result = main(AL->argc, AL->argv);

    cleanupmain();

    return result;
}




/*
 *   cleanupmain() calls __exitfunc() if needed and
 *   frees all resources allocated by __setupmain()
 */

static void cleanupmain(void)
{
    FILE *currfile;
    /* Here begins the cleanup after calling main() */

    /* Check is DOSBase is valid */
    if (!DOSBase)
	return;

    /* If user has defined an exit function, call it */
    if (__exitfunc)
	__exitfunc();

    /* Let's flush standard outputs */
    if (stdout) fflush(stdout);
    if (stderr) fflush(stderr);

    /* Close all user-opened files */
    while((currfile = (FILE *)HeadNode(&__openfiles)))
    {
	fclose(currfile);
    }

    /* Free parameter list, if any */
    if (AL)
	FreeArgs(AL);

    /* Free all memory allocated by us */
    FreeRememberAll(&_memlist);

    CloseLibrary((struct Library *)DOSBase);
}


/*
 *   Here is exit(), actually defined in stdlib.h
 */

void exit(int status)
{
    cleanupmain();
    __exitasm(status);
}

/*
 *   Added to satisfy new GCC which thinks that main() should
 *   call __main().
 */
int __main(void)
{
    return 0;
}
