#ifndef STDIO_H
#define STDIO_H

#include "stddef.h"
#include "stdarg.h"

#define EOF -1L
#define FILENAME_MAX 108
#define FOPEN_MAX 0x7FFFFFFF /* This system has dynamic file open tables */

#define _IOFBF 0 /* full buffering */
#define _IOLBF 1 /* line buffering */
#define _IONBF 2 /* no buffering   */

#define SEEK_SET 0 /* Set offset from beginning of file */ 
#define SEEK_CUR 1 /* Set offset from current location  */
#define SEEK_END 2 /* Set offset from end of file       */

#define __DEFAULT_BUFSIZE 512
#define BUFSIZ __DEFAULT_BUFSIZE

/* This structures and typedefs are system private. You may NOT   */
/* use their contents for anything if you wish your programs will */
/* work also with the next versions of the library or if you wish */
/* to write clean ANSI-C programs.                                */
struct __MinNode
{
    struct __MinNode *lmn_Next;
    struct __MinNode *lmn_Prev;
};

typedef struct {
    struct __MinNode mn;
    unsigned long filehandle;
    char *buf;
    long bufsize;             /* Total size of buffer                     */
    long bufpos;              /* Position in buffer                       */
    long bufused;             /* How much of the read buffer in use       */
    long flags;
    int error;
    int eof;
    short bufmode;
    short unc;
    char tmpfilename[FILENAME_MAX+1];
} FILE;
/* System private typedefs and structures end here */

typedef long fpos_t;

#define __FF_TMPFILE    0x1  /* Temporary file to be deleted after close */
#define __FF_BUFISREAD  0x2  /* The buffer is a read buffer */
#define __FF_SYSFILE    0x4  /* The file opened by system */
#define __FF_SYSBUF     0x8  /* The file's buffer allocated by system */

extern struct List __openfiles;
extern FILE *stdin;
extern FILE *stdout;
extern FILE *stderr;

int sprintf(char *d, const char *fmt, ...);
int fprintf(FILE *fp, const char *fmt, ...);
int printf(const char *fmt, ...);

int vsprintf(char *d, const char *fmt, va_list arg);
int vfprintf(FILE *fp, const char *fmt, va_list arg);
int vprintf(const char *fmt, va_list arg);

FILE *fopen(const char *filename, const char *mode);
int fclose(FILE *stream);
int fflush(FILE *stream);
int fputc(int c, FILE *stream);
#define putc(c, stream) (fputc((c),(stream)))
#define putchar(c) (putc((c),stdout))
int fseek(FILE *stream, long offset, int origin);
long ftell(FILE *stream);
#define rewind(stream) ((void) {(fseek((stream), 0, SEEK_SET)); clearerr(stream)}) 
int fgetpos(FILE *stream, fpos_t *ptr);
int fsetpos(FILE *stream, const fpos_t *ptr);

void clearerr(FILE *stream);
int feof(FILE *stream);
#define feof(stream) ((stream)->feof)
#define ferror(stream) ((stream)->ferror)
#define cleararr(stream) ((void) ((stream)->feof = (stream)->ferror = 0))
#define perror(s) ((void) (fprintf(stderr, "%s: %s\n", (s), "error message")))
int setvbuf(FILE *stream, char *buf, int mode, size_t size);
void setbuf(FILE *stream, char *buf);
int fputs(const char *s, FILE *stream);
int fgetc(FILE *stream);
#define getc(stream) fgetc(stream)
#define getchar() getc(stdin)
char *fgets(char *s, int n, FILE *stream);
char *gets(char *s);
int puts(const char *s);
int ungetc(int c, FILE *stream);
size_t fread(void *ptr, size_t size, size_t nobj, FILE *stream);
size_t fwrite(const void *ptr, size_t size, size_t nobj, FILE *stream);
int remove(const char *filename);
int rename(const char *oldname, const char *newname);

    /* System private functions. You may NOT use them in your ANSI-C code! */
    /* returns true if abort is active */
void __chkabort(void);
int __doscanf(va_list ap, FILE *fp, int (*cf)(int, FILE *), const char *fmt);

#endif
