/*	ls.c for BOAR by Pasi Ojala  */

#include <exec/types.h>
#include <exec/execbase.h>
#include <exec/memory.h>
#include <dos/dos.h>
#include <exec/exec_protos.h>
#include <dos/dos_protos.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#ifndef FIBF_HIDDEN
#define FIBB_HIDDEN	15
#define FIBF_HIDDEN	(1<<FIBB_HIDDEN)
#define FIBB_SUID	31
#define FIBF_SUID	(1<<FIBB_SUID)
#endif

void PutStr(char *a)
{
    Write(Output(), a, strlen(a));
}


#define NAME_LENGTH	108
#define COMM_LENGTH	80
#define MAXCOLS		20
#define BUFSIZE		1024

char *vers="\0$VER: ls 1.1\n";
char *usage=
"Usage: ls [-hRaldogncC] [directory/file list]\n"\
"           -h, -?  help\n"\
"           -R      recursive\n"\
"           -a      show all\n"\
"           -l      long listing\n"\
"           -d      show the directory itself\n"\
"           -o      show file owner info (sets -l too)\n"\
"           -g      show file owner and group info (sets -lo too)\n"\
"           -n      no sorting\n"\
"           -c      show comments (sets -l too)\n"\
"           -C      one-column mode\n";


struct base
{
    struct Node node;
    char   filename[NAME_LENGTH];/* name of the file */
    LONG   dirtype;		/* dirtype */
    LONG   protection;		/* extented protection bits */
    LONG   size;		/* size of the file in bytes */
    LONG   numblocks;		/* size of the file in blocks */
    struct DateStamp date;	/* last changed */
    char   comment[COMM_LENGTH];/* comment line */
    UWORD  OwnerUID;
    UWORD  OwnerGID;
};

#define FLAG_HELP		(1<<0)
#define FLAG_ALL		(1<<1)
#define FLAG_LONG		(1<<2)
#define FLAG_OWNER		(1<<3)
#define FLAG_GROUP		(1<<4)
#define FLAG_RECURSIVE	(1<<5)
#define FLAG_FILENOTES	(1<<6)
#define FLAG_ONECOLUMN	(1<<7)
#define FLAG_DIRECTORY	(1<<8)
#define FLAG_NOSORT		(1<<9)
#define FLAG_SINCE		(1<<10)
#define FLAG_UPTO		(1<<11)

int  readdir(char *directory, long flags, long outputwidth, long curdays);
char *convertdate(long curdays, struct DateStamp *ds, char *string, int mode);
int date2stamp(struct DateStamp *ds, char *string);
int listfiles(struct List *filelist, char *directory, int curdays,
	      long flags, long outputwidth);
int addfiles(struct List *filelist, struct FileInfoBlock *fib, long flags);


int readdir(char *directory, long flags, long outputwidth, long curdays)
{
    struct	base *help;
    int		buflen  = sizeof(struct FileInfoBlock)+85+180, error = FALSE;
    struct	FileInfoBlock *fib =
	(struct FileInfoBlock *)AllocMem(buflen, MEMF_CLEAR);
    char	*temp, *apu;
    APTR	mylock;
    struct  List filelist;
    
    if(!fib)
    {
	PutStr("Memory allocation error!\n");
	return TRUE;
    }
    temp = (char *)fib + sizeof(struct FileInfoBlock);
    apu  = temp + 180;
    NewList(&filelist);
    
    if((mylock = Lock(directory, ACCESS_READ)))
    {
	if(Examine(mylock, fib))
	{
#define ST_ROOT         1
#define ST_USERDIR      2
#define ST_SOFTLINK     3       /* looks like dir, but may point to a file! */
#define ST_LINKDIR      4       /* hard link to dir */
#define ST_FILE         -3      /* must be negative for FIB! */
#define ST_LINKFILE     -4      /* hard link to file */
#define ST_PIPEFILE     -5      /* for pipes that support ExamineFH */

	    if(fib->fib_DirEntryType < 0 /*ST_FILE*/ ||
	       (flags & FLAG_DIRECTORY))
	    {
		UnLock(mylock);
		if(addfiles(&filelist, fib, flags))
		{
		    error = TRUE;
		    goto errorexit;
		}
		error= listfiles(&filelist, NULL, curdays, flags, outputwidth);
		goto errorexit;
	    }
	    else
	    {
		while(ExNext(mylock, fib))
		{
		    if( ((flags & FLAG_ALL) && fib->fib_FileName[0]) ||
		       (!(flags & FLAG_ALL) &&
			(fib->fib_FileName[0] != '.') &&
			stricmp((fib->fib_FileName+strlen(fib->fib_FileName)-5),
				".info") &&
			fib->fib_FileName[0] &&
			!(fib->fib_Protection & FIBF_HIDDEN) ) )
		    {
			if(addfiles(&filelist, fib, flags))
			{
			    UnLock(mylock);
			    error = TRUE;
			    goto errorexit;
			}
			
		    }
		    if(SetSignal(0L, SIGF_CTRL_C) & SIGF_CTRL_C)
		    {
			UnLock(mylock);
			PutStr("\n***Break\n");
			error = TRUE;
			goto errorexit;
		    }
		}
		ExamineEnd(mylock, fib);
	    }
	}
	else
	{
	    PutStr("Examine failed\n");
	}
	UnLock(mylock);
    }
    else
    {
	PutStr("Can't open ");
	PutStr(directory);
	PutStr("\n");
	error = TRUE;
	goto errorexit;
    }
    error= listfiles(&filelist, directory, curdays, flags, outputwidth);
    
 errorexit:
    while(help=(struct base *)RemHead(&filelist))
	FreeMem(help, sizeof(*help));
    
    if(fib)
	FreeMem(fib, buflen);
    
    return error;
}

int addfiles(struct List *filelist, struct FileInfoBlock *fib, long flags)
{
    struct base *new, *help;
    
    if(!(new = (struct base *)AllocMem(sizeof(struct base), MEMF_CLEAR)))
    {
	return RETURN_ERROR;
    }
    new->dirtype = fib->fib_DirEntryType;
    strncpy(new->filename, fib->fib_FileName, NAME_LENGTH-2);
    new->protection = fib->fib_Protection;
    new->size = fib->fib_Size;
    new->numblocks = fib->fib_NumBlocks;
    new->date = fib->fib_Date;
    strcpy(new->comment, fib->fib_Comment);
    new->OwnerUID = fib->fib_OwnerID;
    new->OwnerGID = fib->fib_GroupID;

    if(new->dirtype > 0)
    {
	new->size = 0;
	new->numblocks = 0;
    }
    
    if(flags & FLAG_NOSORT)
    {
	AddTail(filelist, (struct Node *)new);
    }
    else
    {
	help = (struct base *)filelist->lh_Head;
	if(help->node.ln_Next && stricmp(new->filename, help->filename) > 0)
	{
	    while(help->node.ln_Next->ln_Next &&
		  stricmp(new->filename,
			  ((struct base *)(help->node.ln_Next))->filename) > 0)
		help=(struct base *)help->node.ln_Next;
	    
	    if(help->node.ln_Next->ln_Next)
	    {
		Insert(filelist, (struct Node *)new, (struct Node *)help);
	    }
	    else
	    {
		AddTail(filelist, (struct Node *)new);
	    }
	}
	else
	{
	    AddHead(filelist, (struct Node *)new);
	}
    }
    return RETURN_OK;
}


void putbits(char *buffer, ULONG protection)
{
    int i;

    strcpy(buffer, "hspadrwedrwedrwe");

    for(i=0;i<4;i++)
    {
	if(!(protection & (FIBF_ARCHIVE<<i)))
	    buffer[3-i] = '-';
	if(!(protection & (FIBF_EXECUTE<<i)))
	    buffer[7-i] = '-';
	if(!(protection & (FIBF_GRP_EXECUTE<<i)))
	    buffer[11-i] = '-';
	if(!(protection & (FIBF_OTR_EXECUTE<<i)))
	    buffer[15-i] = '-';
    }
}

int listfiles(struct List *filelist, char *directory, int curdays,
	      long flags, long outputwidth)
{
    struct	base *help;
    int		i, buflen=strlen(directory)+85+200;
    APTR	mylock = NULL;
    char	*buffer=(char *)AllocMem(buflen, MEMF_CLEAR);
    char	*temp, *apu;
    
    if(!buffer)
    {
	PutStr("Memory allocation error!\n");
	return RETURN_ERROR;
    }
    
    temp = buffer+buflen-200;
    apu  = temp+160;
    
    if(directory && *directory)
    {
	PutStr(directory);
	PutStr(" (dir):");
	
	if(!filelist->lh_Head->ln_Next)
	    PutStr(" is empty");
	
	PutStr("\n");
    }
    
    if((flags & FLAG_LONG))
    {
	help = (struct base *)filelist->lh_Head;
	while(help->node.ln_Next)
	{
	    putbits(buffer+2, help->protection);
	    
	    if((help->protection & FIBF_SUID))
	    {
		if(buffer[8] != '-')
		    buffer[8] = 'S';
	    }
	    buffer[0] = '-';
	    if(help->dirtype > 0)
		buffer[0] = 'd';

	    if(help->dirtype == ST_LINKFILE)
		buffer[0] = 'l';
	    else if(help->dirtype == ST_LINKDIR)
		buffer[0] = 'D';
	    else if(help->dirtype == ST_SOFTLINK)
		buffer[0] = 's';
	    
	    buffer[1] = 'c';
	    if(!help->comment[0])
		buffer[1] = '-';

	    if((flags & FLAG_GROUP))
	    {
		sprintf(apu, "id=%d", help->OwnerUID);
		sprintf(temp, "%s %-8s ", buffer, apu);
		sprintf(apu, "id=%d", help->OwnerGID);
		sprintf(temp + 28, "%-8s ", apu);
		sprintf(temp + 36,
			" %8d %s %s",
			help->size,
			convertdate(curdays, &help->date, apu, 0),
			help->filename);
	    }
	    else if(flags & FLAG_OWNER)
	    {
		sprintf(apu, "id=%d", help->OwnerUID);
		sprintf(temp, "%s %-8s ", buffer, apu);
		sprintf(temp + 27,
			" %8d %s %s",
			help->size,
			convertdate(curdays, &help->date, apu, 0),
			help->filename);
	    }
	    else
	    {
		sprintf(temp,
			"%18s %5d %8d %s %s",
			buffer,
			help->numblocks,
			help->size,
			convertdate(curdays, &help->date, apu, 1),
			help->filename);
	    }
	    PutStr(temp);
	    
	    if(help->dirtype == ST_SOFTLINK)
	    {
		struct FileInfoBlock *fib = (struct FileInfoBlock *)
		    AllocMem(sizeof(struct FileInfoBlock), NULL);
		
		if(*directory)
		    sprintf(temp, "%s/%s", directory, help->filename);
		else
		    strcpy(temp, help->filename);
		
		if(fib && (mylock = Lock(temp, ACCESS_READ)) &&
		   Examine(mylock, fib))
		{
		    UnLock(mylock);
		    sprintf(temp, "  %s\n", fib->fib_FileName);
		    if(fib->fib_DirEntryType > 0)
			strcat(temp, "/");
		    PutStr(temp);
		}
		else
		{
		    if(mylock)
			UnLock(mylock);
		    PutStr("  (Can't examine)\n");
		}
		if(fib)
		    FreeMem(fib, sizeof(struct FileInfoBlock));
	    }
	    else
	    {
		PutStr("\n");
	    }
	    
	    if((flags & FLAG_FILENOTES) && help->comment[0] )
	    {
		PutStr("\"");
		PutStr(help->comment);
		PutStr("\"\n");
	    }
	    
	    help = (struct base *)help->node.ln_Next;
	    if(SetSignal(0L, SIGF_CTRL_C) & SIGF_CTRL_C)
	    {
		PutStr("\n***Break\n");
		if(buffer)
		    FreeMem(buffer, buflen);
		return TRUE;
	    }
	}
	/*PutStr("\n");*/
    }
    else if(flags & FLAG_ONECOLUMN)
    {
	help = (struct base *)filelist->lh_Head;
	while(help->node.ln_Next)
	{
	    strcpy(buffer, help->filename);
	    if(help->dirtype >= 0)
	    {
		if(help->dirtype == ST_SOFTLINK)
		    strcat(buffer, "");
		else
		    strcat(buffer, "/");
	    }
	    PutStr(buffer);
	    PutStr("\n");
	    if(SetSignal(0L, SIGF_CTRL_C) & SIGF_CTRL_C)
	    {
		PutStr("\n***Break\n");
		if(buffer)
		    FreeMem(buffer, buflen);
		return TRUE;
	    }
	    help = (struct base *)help->node.ln_Next;
	}
    }
    else
    {
	char *temp2="                                ";
	int	longest=0, files=0, cols=0, rows, j=0, len=strlen(temp2);
	struct base *rs[MAXCOLS];
	
	help = (struct base *)filelist->lh_Head;
	while(help->node.ln_Next)
	{
	    if(strlen(help->filename) > longest)
		longest = strlen(help->filename);
	    files++;
	    help = (struct base *)help->node.ln_Next;
	}
	longest += 2;	/* two spaces apart */
	
	if((cols = (outputwidth+2)/longest) > MAXCOLS)
	    cols = MAXCOLS;
	rows = files/cols + 1;
	
	for(i=0; i<MAXCOLS; rs[i++] = NULL);
	
	help = (struct base *)filelist->lh_Head;
	for(j=i=0; i<rows*cols; i++)
	{
	    if(help->node.ln_Next)
	    {
		if(!(i%rows))
		{
		    rs[j++] = help;
		}
		help = (struct base *)help->node.ln_Next;
	    }
	    else
	    {
		if(!(i%rows))
		{
		    rs[j++] = NULL;
		}
	    }
	}
	
	for(i=0; i<rows; i++)
	{
	    j = 0;
	    buffer[0] = '\0';
	    while(rs[j] && rs[j]->node.ln_Next && j<cols)
	    {
		strcat(buffer, rs[j]->filename);
		switch(rs[j]->dirtype)
		{
		case ST_ROOT:
		case ST_USERDIR:
		    strcat(buffer, "/");
		    break;
		case ST_LINKDIR:
		    strcat(buffer, "@");
		    break;
		case ST_SOFTLINK:
		    strcat(buffer, "");
		    break;
		default:
		    strcat(buffer, " ");
		}
		if(j<cols-1 && rs[j+1])
		    strcat(buffer,
			   temp2+len-longest+(strlen(rs[j]->filename)+1));
		
		rs[j] = (struct base *)rs[j]->node.ln_Next;
		j++;
	    }
	    PutStr(buffer);
	    PutStr("\n");
	    if(SetSignal(0L, SIGF_CTRL_C) & SIGF_CTRL_C)
	    {
		PutStr("\n***Break\n");
		if(buffer)
		    FreeMem(buffer, buflen);
		return TRUE;
	    }
	}
	/*PutStr("\n");*/
    }
    
    if((flags & FLAG_RECURSIVE) && !(flags & FLAG_DIRECTORY))
    {
	help = (struct base *)filelist->lh_Head;

	while(help->node.ln_Next)
	{
	    if(help->dirtype==ST_ROOT || help->dirtype==ST_USERDIR)
	    {
		i=*(directory+strlen(directory)-1);
		if(*directory && i!='/' && i!=':')
		    sprintf(buffer, "%s/%s", directory, help->filename);
		else
		    sprintf(buffer, "%s%s", directory, help->filename);
		if(readdir(buffer, flags, outputwidth, curdays))
		{
		    if(buffer)
			FreeMem(buffer, buflen);
		    return TRUE;
		}
	    }
	    help = (struct base *)help->node.ln_Next;
	}
    }
    
    if(buffer)
	FreeMem(buffer, buflen);
    return FALSE;
}

static LONG timezone = 0;
static char *m[]={"", "Jan", "Feb", "Mar", "Apr",
		      "May", "Jun", "Jul", "Aug",
		      "Sep", "Oct", "Nov", "Dec", ""};
char *convertdate(long curdays, struct DateStamp *ds, char *temp, int mode)
{
    long n;
    int months, days, years;

    ds->ds_Minutes += timezone;
    if(ds->ds_Minutes < 0)
    {
	ds->ds_Minutes += 60*24;
	ds->ds_Days--;
    }
    else if(ds->ds_Minutes > 60*24)
    {
	ds->ds_Minutes -= 60*24;
	ds->ds_Days++;
    }
    n = ds->ds_Days-2251	/*+2922*/;		/* Make it 28th Feb 76 */

    years = (4*n+3)/1461;	/* Handle leap years */
    n -= 1461*(long)years/4;
    months = (5*n+2)/153;	/* Get the # of months after Feb */
    days = n-(153*months+2)/5+1;
    months += 3;
    if(months > 12)
    {
	years++;
	months -= 12;
    }
    years += 84;

    if(days<1 || days>31 || months<1 || months>12 || years<78)
    {
	days = months = 1;
	years = 78;
    }
    years += 1900;

    if(mode == 0)
    {
	if(ds->ds_Days == curdays)
	{
	    sprintf(temp, "Today  %02d:%02d",
		    ds->ds_Minutes/60, ds->ds_Minutes%60);
	}
	else if(ds->ds_Days + 180 < curdays)
	{
	    sprintf(temp, "%02d-%3s-%04d ",
		    days, m[months], years);
	}
	else
	{
	    sprintf(temp, "%02d-%3s %02d:%02d",
		    days, m[months], ds->ds_Minutes/60, ds->ds_Minutes%60);
	}
    }
    else
    {
	if(ds->ds_Days == curdays-1)
	{
	    sprintf(temp, "Yesterday %02d:%02d",
		    ds->ds_Minutes/60, ds->ds_Minutes%60);
	}
	else if(ds->ds_Days == curdays)
	{
	    sprintf(temp, "Today     %02d:%02d",
		    ds->ds_Minutes/60, ds->ds_Minutes%60);
	}
	else
	{
	    sprintf(temp, "%02d-%3s-%02d %02d:%02d",
		    days, m[months], years%100, ds->ds_Minutes/60, ds->ds_Minutes%60);
	}
    }
    return temp;
}

int main(int argc, char *argv[])
{
    int i, ulostettu=FALSE;
    long flags=0, outputwidth = 79, curdays;
    char *str;
    char temp[100];
    struct DateStamp now;
    struct List filelist;

    NewList(&filelist);
    DateStamp(&now);
    curdays = now.ds_Days;

#if 0
    if(GetVar("TIMEZONE", temp, sizeof(temp), LV_VAR) != -1)
    {
	timezone = atol(temp);
	if(timezone < 0)
	{
	    timezone = -(60*(-timezone/100) + (-timezone%100));
	}
	else
	{
	    timezone = 60*(timezone/100) + (timezone%100);
	}
    }
#endif
    timezone = tzminutes();

    if(GetVar("columns", temp, sizeof(temp), LV_VAR) != -1)
    {
	outputwidth = atol(temp) - 1;
    }
    else if(GetVar("cols", temp, sizeof(temp), LV_VAR) != -1)
    {
	outputwidth = atol(temp) - 1;
    }
    if(outputwidth < 2)
	outputwidth = 79;

    for(i=1; i<argc; i++)
    {
	if(argv[i][0] == '-')
	{
	    str = argv[i]+1;
	    while(*str)
	    {
		switch(*str++)
		{
		case 'R':
		    flags |= FLAG_RECURSIVE;
		    break;
		case 'a':
		    flags |= FLAG_ALL;
		    break;
		case 'l':
		    flags |= FLAG_LONG;
		    break;
		case 'o':
		    flags |= FLAG_LONG | FLAG_OWNER;
		    break;
		case 'g':
		    flags |= FLAG_LONG | FLAG_GROUP | FLAG_OWNER;
		    break;
		case 'c':
		    flags |= FLAG_FILENOTES | FLAG_LONG;
		    break;
		case 'C':
		    flags |= FLAG_ONECOLUMN;
		    break;
		case 'd':
		    flags |= FLAG_DIRECTORY;
		    break;
		case 'n':
		    flags |= FLAG_NOSORT;
		    break;
		case 'h':
		case '?':
		default:
		    flags |= FLAG_HELP;
		    break;
		}
	    }
	}
	else
	{
	    if(flags & FLAG_HELP)
	    {
		break;
	    }
	    if(readdir(argv[i], flags, outputwidth, curdays))
		PutStr("ERROR\n");
	    ulostettu = TRUE;
	}
    }

    if(flags & FLAG_HELP)
    {
	PutStr(usage);
    }
    else
    {
	if(!ulostettu)
	    readdir("", flags, outputwidth, curdays);
    }
    return 0;
}
