#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include <exec/tasks.h>
#include "dp83932b.h"
#include "ether.h"
#include <exec/execbase.h>
#include <exec/exec_protos.h>

/* These are called from the device library-base */
LONG EtherOpen(ULONG unit, struct StdIOReq *req, ULONG flags);
void EtherClose(struct StdIOReq *req);
void EtherBeginIO(struct StdIOReq *req);
void EtherAbortIO(struct StdIOReq *req);

/*-------------------------------------------*/
static struct Unit *OpenEther(UWORD channel,struct StdIOReq *req);
static void	CloseEther(struct Unit *ether);
/*-------------------------------------------*/


struct Unit
{
    struct MsgPort un_Port;
    ULONG un_Flags;
    UWORD un_OpenCnt;

    struct Interrupt Int;

    struct List ReadRequests;
    struct List WriteRequests;

    struct StdIOReq *writereq;
    struct StdIOReq *readreq;

    ULONG MP_CounterMSL;
    ULONG FAE_CounterMSL;
    ULONG CRC_CounterMSL;
};

struct EtherDevice
{
    struct Device sd_Device;
    struct Unit *sd_Unit;
};


static LONG EtherOpen(ULONG unit, struct StdIOReq *req, ULONG flags)
{
    struct Unit *temp;

/* TODO: locking */

    /* Is the unit currently in use ? */
    temp = ((struct EtherDevice *)(req->io_Device))->sd_Unit;
    if(temp)
    {
	req->io_Unit = ((struct EtherDevice *)req->io_Device)->sd_Unit;
	req->io_Unit->un_OpenCnt++;
	req->io_Error = 0;
	req->io_Device->dd_Library.lib_OpenCnt++;
    }
    else
    {
	if(temp = OpenEther(unit, req))
	{
	    ((struct EtherDevice *)req->io_Device)->sd_Unit = temp;
	    req->io_Unit = temp;
	    req->io_Unit->un_Flags = flags;
	    req->io_Unit->un_OpenCnt++;
	    req->io_Error = 0;
	    req->io_Device->dd_Library.lib_OpenCnt++;
	}
	else
	{
	    req->io_Error = 2;
	}
    }
    return req->io_Error;
}

static void EtherClose(struct StdIOReq *req)
{
    if(req->io_Unit)
    {
	req->io_Device->dd_Library.lib_OpenCnt--;
	req->io_Unit->un_OpenCnt--;
	if(!req->io_Unit->un_OpenCnt)
	{
	    CloseEther(req->io_Unit);
	}
	req->io_Unit = NULL;
    }
}

static void EtherBeginIO(struct StdIOReq *req)
{
    PutMsg(&req->io_Unit->un_Port, &req->io_Message);
}

static void EtherAbortIO(struct StdIOReq *req)
{
    struct Unit *ether = (struct Unit *)req->io_Unit;

    Disable();
    if(req->io_Message.mn_Node.ln_Type == NT_REPLYMSG)
    {
	Enable();
	return;
    }
    if(req == ether->readreq)
    {
	req->io_Actual = 0;
	ReplyMsg( &req->io_Message);
	ether->readreq = (struct StdIOReq *)RemHead(&ether->ReadRequests);
	Enable();
	return;
    }
    if(req == ether->writereq)
    {
	/* Well, let the user just wait until the block is transmitted */
	Enable();
	return;
    }
    Remove((struct Node *)req);
    req->io_Actual = 0;
    ReplyMsg(&req->io_Message);
    Enable();
    return;
}


static void EtherHandler(ULONG is_Data, UWORD *StackFrame);
static void EtherHandler(ULONG is_Data, UWORD *StackFrame)
{
    struct Unit *ether=(struct Unit *)is_Data;
    struct StdIOReq *message;
    struct sonicregs * const SONIC= (struct sonicregs *)0x600000;
    UWORD status = SONIC->InterruptStatus;

    if((status & ISR_MP))
    {
	ether->MP_CounterMSL++;
    }
    if((status & ISR_FAE))
    {
	ether->FAE_CounterMSL++;
    }
    if((status & ISR_CRC))
    {
	ether->CRC_CounterMSL++;
    }
    if((status & ISR_TC))
    {
	/* Timer complete -- ? */
    }
    if((status & ISR_TXER))
    {
	/* Transmit error for any packet */
    }
    if((status & ISR_TXDN))
    {
	/* Transmit complete for n packets */
    }
    if((status & ISR_PKTRX))
    {
	/* Packet(s) received */
    }
    if((status & ISR_PINT))
    {
	/* Programmed interrupt */
    }
    if((status & ISR_HBL))
    {
	/* Heartbeat lost */
    }
    if((status & ISR_BR))
    {
	/* Bus retry occurred */
    }
    SONIC->InterruptStatus = status; /* Clear handled events */

    while( (message = (struct StdIOReq *)GetMsg(&ether->un_Port)) )
    {
	message->io_Error = 10;
	switch(message->io_Command)
	{
	case CMD_READ:
	    break;
	case CMD_WRITE:
	    break;
	case CMD_SETPARAMS:
	    break;
	case CMD_FLUSH:
	    break;
	case CMD_RESET:
	    break;
	default:
	    break;
	}
    }
}

static struct Unit *OpenEther(UWORD channel,struct StdIOReq *req)
{
    struct Unit *ether = (struct Unit *)AllocMem(sizeof(struct Unit), 0);

    if(!ether)
	return NULL;

    ether->readreq = NULL;
    ether->writereq = NULL;
    NewList( &ether->ReadRequests );
    NewList( &ether->WriteRequests );

    ether->Int.is_Code = &EtherHandler;
    ether->Int.is_Data = (ULONG)ether;
    AddIntServer(81, &ether->Int);

    NewList(&ether->un_Port.mp_MsgList);
    ether->un_Port.mp_Flags = PA_SOFTINT;
    ether->un_Port.mp_SoftInt = (void *)(81);  /* The same handler */

    ether->MP_CounterMSL = 0;
    ether->FAE_CounterMSL = 0;
    ether->CRC_CounterMSL = 0;
    /* Start receiving ? */

    return ether;
}

static void CloseEther(struct Unit *ether)
{
    RemIntServer(81, &ether->Int);
    FreeMem(ether, sizeof(struct Unit));
}

void AddEtherDevice()
{
    ULONG memchunk = (ULONG)AllocMem(sizeof(struct EtherDevice)+24, 0);
    /* allocate struct device and lib_negsize */
    struct EtherDevice *etherdev;
    static const ULONG vectors[4] = {
	(ULONG)EtherOpen,
	(ULONG)EtherClose,
	(ULONG)EtherBeginIO,
	(ULONG)EtherAbortIO
    };

    if(memchunk)
    {
	etherdev = (struct EtherDevice *)(memchunk+24);

	InitLibFuncs(&etherdev->sd_Device.dd_Library, (ULONG *)vectors, 4);
	etherdev->sd_Device.dd_Library.lib_Node.ln_Name = "ether.device";
	etherdev->sd_Device.dd_Library.lib_Version = 1;
	etherdev->sd_Device.dd_Library.lib_OpenCnt = 0;
	etherdev->sd_Unit = NULL;

	AddDevice( &etherdev->sd_Device );
    }
}
