#include <exec/types.h>
#include <exec/exec_protos.h>
#include <dos/dos_protos.h>
#include <stdlib.h>
#include <stdio.h>
#include "dp83932b.h"

#define MIN_FRAGMENT_SIZE 64
#define MTU 1518 /* in bytes, should be even quantity */
#define MY_ADDRESS 0x82e624d0 /* 130.230.36.208 (sigx8.cs.tut.fi) */
#define PHYS_ADDRESS "\010\000\011\042\003\135" /*"\135\003\042\011\000\010"*/ /* LSB first */
#define BROADCAST "\377\377\377\377\377\377"
#define MULTICAST "\377\377\377\377\0\0"
/*
sigx8.cs.tut.fi:\
        tc=sigx.defaults\
	ha=08000922035d:\
        ip=130.230.36.208
*/
#define SONIC_MEM_ALIGN (64*1024)
#define SONIC_MEM_SIZE  (8*1024)

struct TDA
{
    UWORD status;
    UWORD config;
    UWORD pkt_size;
    UWORD frag_count; /**  3 **/
    UWORD frag0_ptr0; /** /  **/
    UWORD frag0_ptr1; /**< physical addresses **/
    UWORD frag0_size; /** \  **/
    UWORD frag1_ptr0; /** /  **/
    UWORD frag1_ptr1; /**< the packet itself **/
    UWORD frag1_size; /** \  **/
    UWORD frag2_ptr0; /** /  **/
    UWORD frag2_ptr1; /**< padding **/
    UWORD frag2_size; /** \  **/
    UWORD link;
};

#define NUM_RDA    256 /* Number of receive descriptors */
struct RDA
{
    UWORD status;
    UWORD byte_count;
    UWORD pkt_ptr0;
    UWORD pkt_ptr1;
    UWORD seq_no;
    UWORD pkt_link;
    UWORD in_use;
};

#define NUM_RRA     96
#define NUM_RBA     16 /* Number of buffers */
#define RBA_SIZE  8192 /* size of one buffer */
struct RRA
{
    UWORD buff_ptr0;
    UWORD buff_ptr1;
    UWORD buff_wc0;
    UWORD buff_wc1;
};

#define MIN_TX_SIZE (16*sizeof(struct TDA))
#define MIN_RX_SIZE (NUM_RBA*sizeof(struct RRA)+64*sizeof(struct RDA))

struct CAMEntry
{
    UWORD EntryPointer;
    UWORD AddressPort0;
    UWORD AddressPort1;
    UWORD AddressPort2;
};

struct SCORE
{
    UBYTE *buffer; /* pointer to the buffer area */
    UWORD seq;     /* RRA sequence number */
    UWORD packets; /* Number of unprocessed packets left in this buffer */
};


void show_received(struct RDA *crda)
{
    struct sonicregs * const SONIC= (struct sonicregs *)0x600000;
    UWORD status = SONIC->ReceiveControl;

    if((status & 1))
    {
	UBYTE *packet = (UBYTE *)(crda->pkt_ptr0 | (crda->pkt_ptr1<<16));
	int i;

	printf("Packet received\n");
	printf("\tStatus: 0x%04x (0x%04x)\n",
	       crda->status,
	       status);
	printf("\tBytes:  0x%04x\n", crda->byte_count);
	printf("\tSeq_no: 0x%04x\n", crda->seq_no);
	printf("\tInuse:  0x%04x\n", crda->in_use);
	printf("\tData:   ");

	for(i=0;i<crda->byte_count;i++)
	{
	    if(!(i&31))
		printf("\n\t");
	    if(!(i&7))
		printf(" ");
	    printf("%02x", *(packet+i));
	}
	printf("\n");
    }
    if((status & 2))
    {
	printf("Loopback packet received\n");
    }
    if((status & 4))
    {
	printf("Frame Alignment Error\n");
    }
    if((status & 8))
    {
	printf("CRC Error\n");
    }
    if((status & 16))
    {
	printf("Collision Activity\n");
    }
    if((status & 32))
    {
	printf("CRS Activity\n");
    }
    if((status & 64))
    {
	printf("Last Packet In RBA\n");
    }
    if((status & 128))
    {
	printf("Broadcast Packet\n");
    }
    if((status & 256))
    {
	printf("Multicast Packet\n");
    }
}


static struct SCORE scoreBoard[NUM_RBA];
static int sequence_num;

#define STAT_BROADCAST 1
#define STAT_MULTICAST 2
#define STAT_TLOOPBACK 4
#define STAT_ELOOPBACK 8
#define STAT_MLOOPBACK 16
#define STAT_WAIT      32

int main(int argc, char *argv[])
{
    UBYTE *sonicmem, *rxstart, *txstart;
    LONG txsize, rxsize, status = 0;
    struct sonicregs * const SONIC= (struct sonicregs *)0x600000;
    int i, tmp;

    struct TDA *tailtda, *ctda; /* transmit descriptor addresses */
    struct RDA *lrda, *urda, *crda; /* low/up/curr receive desc.addr */
    struct RRA *rra; /* resource area addr */
    struct CAMEntry *CAM;
    int num_rra = NUM_RRA, num_rda = NUM_RDA, rba_size = RBA_SIZE;

    static char *packet = "A testpacket\n";
    const static char padding[MIN_FRAGMENT_SIZE] = "\0";
    static char address[24] = PHYS_ADDRESS PHYS_ADDRESS;

    for(i=1;i<argc;i++)
    {
	if(!stricmp(argv[i],"broadcast"))
	{
	    status |= STAT_BROADCAST;
	}
	else if(!stricmp(argv[i],"multicast"))
	{
	    status |= STAT_MULTICAST;
	}
	else if(!stricmp(argv[i],"tloopback"))
	{
	    status |= STAT_TLOOPBACK;
	}
	else if(!stricmp(argv[i],"eloopback"))
	{
	    status |= STAT_ELOOPBACK;
	}
	else if(!stricmp(argv[i],"mloopback"))
	{
	    status |= STAT_MLOOPBACK;
	}
	else if(!stricmp(argv[i],"wait"))
	{
	    status |= STAT_WAIT;
	}
	else if(!strncmp(argv[i],"data=",5))
	{
	    packet = argv[i]+5;
	}
	else
	{
	    fprintf(stderr, "Usage: %s [broadcast] [multicast] [tloopback] [eloopback] [mloopback] [wait]\n", argv[0]);
	    exit(10);
	}
    }

    if((status & STAT_BROADCAST))
    {
	memcpy(address, BROADCAST, 6);
    }
    if((status & STAT_MULTICAST))
    {
	memcpy(address, MULTICAST, 6);
    }


    /**************************************************************/
    /*** 0) Allocate Memory for transmit and receive structures ***/
    /**************************************************************/

    txstart = sonicmem = (UBYTE *)calloc(SONIC_MEM_SIZE, 1);
    rxstart = (UBYTE *)((((ULONG)sonicmem)+SONIC_MEM_ALIGN-1)&
			~(SONIC_MEM_ALIGN-1));

    if(rxstart > sonicmem + SONIC_MEM_SIZE)
    {
	printf("The whole block is inside the alignment\n");
	rxstart = sonicmem + SONIC_MEM_SIZE/2;
    }
    txsize = (ULONG)rxstart - (ULONG)sonicmem;
    rxsize = SONIC_MEM_SIZE - txsize;

/*
   .  X----------------X----------------X
   .         |---------X------|
   .           rxstart ^ rxsize
   . txstart ^ txsize
*/

    if(txsize < MIN_TX_SIZE)
    {
	printf("txsize too small, stealing rxbuffer's beginning\n");
	/*
	   . X----------------X----------------X
	   .                |-X--------------|
	   .          rxstart ^
	   .        txstart ^
	   */
	txstart = rxstart;
	txsize = MIN_TX_SIZE;
	rxstart += MIN_TX_SIZE;
	rxsize -= MIN_TX_SIZE;
	/*
	   .       |-X---|----------|
	   .     rxstart ^
	   . txstart ^
	 */
    }
    if(rxsize < MIN_RX_SIZE)
    {
	printf("rxsize too small, stealing txbuffer's end\n");
	/*
	   .       X----------------X----------------X
	   .         |--------------X-|
	   .                rxstart ^
	   . txstart ^
	 */
	rxstart -= MIN_RX_SIZE;
	rxsize = MIN_RX_SIZE;
	txsize -= MIN_RX_SIZE;
	/*
	   .       X----------------X----------------X
	   .         |----------|---X-|
	   .            rxstart ^
	   . txstart ^
	 */
    }

    printf("Mem: 0x%06x, txstart: 0x%06x, txsize: 0x%06x, rxstart: 0x%06x, rxsize: 0x%06x\nRev: 0x%04x\n",
	   sonicmem,
	   txstart, txsize,
	   rxstart, rxsize,
	   SONIC->SiliconRevision);

    /***********************************/
    /*** 1&2) Reset SONIC, configure ***/
    /***********************************/

    /* START of set_reg_value() */
    SONIC->Command = 0x80; /* Reset */
    SONIC->DataConfiguration = 0x8000; /* Extended bus mode */
    SONIC->InterruptMask = 0;
    SONIC->DataConfiguration2 = 0; /* LEDs on for bus master cycles */

    SONIC->ReceiveControl = 0x3800;
    /* *SONIC->InterruptMask = 0x3fff; */
    SONIC->InterruptStatus = 0xffff;
    /* END of set_reg_value() */

    SONIC->EndOfBufferWordCount = MTU/2;

    SONIC->WatchdogTimer0 = 0xb5;
    SONIC->WatchdogTimer1 = 0xb5;
    SONIC->Command = 0x0020; /* start timer */

    /********************************/
    /*** 3) Setup receive filters ***/
    /********************************/

    /****************************/
    /*** 4) Enable Interrupts ***/
    /****************************/

    /****************************/
    /*** 5) Initialize Memory ***/
    /****************************/

    /* CAM init */
    CAM = (struct CAMEntry *)rxstart;
    SONIC->UpperReceiveResourceAddress   = (((ULONG)CAM)>>16);
    SONIC->CAMDescriptorPointer = (ULONG)CAM;
    SONIC->CAMDescriptorCount = 1;
    CAM->EntryPointer = 0;
    CAM->AddressPort0 = (((int)PHYS_ADDRESS[1])<<8)|((int)PHYS_ADDRESS[0]);
    CAM->AddressPort1 = (((int)PHYS_ADDRESS[3])<<8)|((int)PHYS_ADDRESS[2]);
    CAM->AddressPort2 = (((int)PHYS_ADDRESS[5])<<8)|((int)PHYS_ADDRESS[4]);
    CAM[1].EntryPointer = 0x1; /* CAM Enable Mask */
    SONIC->Command = 512; /* Load CAM */

    for(i=0;i<10;i++)
    {
	if(!(SONIC->Command & 512))
	{
	    printf("Load CAM performed\n");
	    break;
	}
	Delay(10);
    }
    SONIC->Command = 0x80;
    SONIC->CAMEntryPointer = 0;
    printf("CAM0: 0x%04x 0x%04x 0x%04x\nCAMEnable: 0x%04x\n",
	   SONIC->CAMAddressPort0,
	   SONIC->CAMAddressPort1,
	   SONIC->CAMAddressPort2,
	   SONIC->CAMEnable
	   );
    SONIC->Command = 0;

    /* Transmit descriptor init */
    /* ctda    = current TDA, where packets to send are added (EOL = 1) */
    /* tailtda = last linked TDA, where used TDA's are linked */

    tmp = txsize/sizeof(struct TDA);

    ctda = (struct TDA *)txstart;
    for(i=1;i<tmp;i++)
    {
	ctda[i-1].link = ((ULONG)&ctda[i]) | 1;
    }
    tailtda = ctda + tmp;
    tailtda->link = 1;

    printf("ctda: 0x%08x  tailtda: 0x%08x number of TDA's: %d\n",
	   ctda, tailtda, tmp);


    /* Receive resource area init */
    rra = (struct RRA *)rxstart;

    tmp = num_rra * sizeof(struct RRA) + num_rda * sizeof(struct RDA);
    while(tmp > rxsize)
    {
	if(num_rra > NUM_RBA)
	{
	    --num_rra;
	}
	--num_rda;
	tmp = num_rra * sizeof(struct RRA) + num_rda * sizeof(struct RDA);
    }
    printf("%d receive resources, %d receive descriptors\n",
	   num_rra, num_rda);

    crda = lrda = (struct RDA *)(rra + num_rra);
    urda = lrda + num_rda;

    printf("lrda: 0x%08x  urda: 0x%08x\n", lrda, urda);

    /* Initialize receive buffer accounting system */
    sequence_num = 0;
    for(i=0;i<NUM_RBA;i++)
    {
	scoreBoard[i].buffer  = malloc(rba_size);
	scoreBoard[i].packets = 0;
	rra[i].buff_ptr0 = (ULONG)scoreBoard[i].buffer;
	rra[i].buff_ptr1 = (((ULONG)scoreBoard[i].buffer)>>16);
	rra[i].buff_wc0 = (rba_size>>1);
	rra[i].buff_wc1 = (rba_size>>17);
	scoreBoard[i].seq = sequence_num++;
    }

    /* Receive descriptor area init */
    for(i=0;i<num_rda-1;i++)
    {
	lrda[i].pkt_link = (ULONG)&lrda[i+1]; /* low 16 bits */
	lrda[i].in_use = i+1; /* non-zero */
    }
    lrda[num_rda-1].pkt_link = (ULONG)&lrda[0] | 1; /* low 16 bits */
    lrda[num_rda-1].in_use = num_rda; /* non-zero */


    /*************************************************/
    /*** 6) Initialize Buffer Management Registers ***/
    /*************************************************/

    SONIC->UpperTransmitDescriptorAddress = (((ULONG)ctda)>>16);
    SONIC->CurrentTransmitDescriptorAddress = (((ULONG)ctda) & 0xffff);

    SONIC->UpperReceiveDescriptorAddress   = (((ULONG)&lrda[0])>>16);
    SONIC->CurrentReceiveDescriptorAddress = (ULONG)crda;
    SONIC->UpperReceiveResourceAddress   = (((ULONG)&rra[0])>>16);
    SONIC->ResourceStartAddress = (ULONG)&rra[0];
    SONIC->ResourceEndAddress   = (ULONG)&rra[num_rra];
    SONIC->ResourceReadPointer  = (ULONG)&rra[0];
    SONIC->ResourceWritePointer = (ULONG)&rra[NUM_RBA];

    SONIC->Command = 256; /* Read RRA */
    for(i=0;i<10;i++)
    {
	if(!(SONIC->Command & 256))
	{
	    printf("Read RRA performed\n");
	    break;
	}
	Delay(10);
    }

    /*************************************************/
    /*** 7) Start Reception by issuing RRA command ***/
    /*************************************************/

    /*******************************/
    /*** 8) Clear Tally Counters ***/
    /*******************************/
    SONIC->CRCErrorTally = 0xffff;
    SONIC->FAETally = 0xffff;
    SONIC->MissedPacketTally = 0xffff;

    /***********************/
    /*** 9) Enable SONIC ***/
    /***********************/

    SONIC->UpperTransmitDescriptorAddress = (((ULONG)ctda)>>16);
    SONIC->CurrentTransmitDescriptorAddress = (((ULONG)ctda) & 0xffff);

    SONIC->Command = 0;
    /* try to send */
    /* Next TDA is linked already */
    ctda->status = 0;
    ctda->config = 0;
    ctda->frag_count = 3;
    ctda->frag0_ptr0 = (((ULONG)address)&0xffff);
    ctda->frag0_ptr1 = (((ULONG)address)>>16);
    ctda->frag0_size = 12;
    ctda->frag1_ptr0 = (((ULONG)packet)&0xffff);
    ctda->frag1_ptr1 = (((ULONG)packet)>>16);
    ctda->frag1_size = strlen(packet);
    ctda->frag2_ptr0 = (((ULONG)padding)&0xffff);
    ctda->frag2_ptr1 = (((ULONG)padding)>>16);
    ctda->frag2_size = 1; /* for now we always need 1 pad byte */
    ctda->pkt_size = ctda->frag0_size + ctda->frag1_size + ctda->frag2_size;

    if(ctda->pkt_size < MIN_FRAGMENT_SIZE)
    {
	ctda->frag2_size += MIN_FRAGMENT_SIZE - ctda->pkt_size;
	printf("Adding %d pad bytes\n", MIN_FRAGMENT_SIZE - ctda->pkt_size);
	ctda->pkt_size = MIN_FRAGMENT_SIZE;
    }

    ctda->link |= 1; /* set EOL */

    printf("Transmit Control: 0x%04x\n", SONIC->TransmitControl);
    printf("Receive Control:  0x%04x\n", SONIC->ReceiveControl);


    if((status & STAT_TLOOPBACK))
    {
	SONIC->ReceiveControl = 0x3800 | 0x600; /* 0x600 Tranc. loopback*/
	printf("Tranceiver loopback\n");
    }
    else if((status & STAT_MLOOPBACK))
    {
	SONIC->ReceiveControl = 0x3800 | 0x200; /* 0x200 MAC loopback*/
	printf("MAC loopback\n");
    }
    else if((status & STAT_ELOOPBACK))
    {
	SONIC->ReceiveControl = 0x3800 | 0x400; /* 0x400 Endec loopback*/
	printf("Endec loopback\n");
    }

    SONIC->Command = 8; /* start reception */
    SONIC->Command = 2; /* start transmit */
    printf("Command Register: 0x%04x\n", SONIC->Command);

    for(i=0;i<3;i++)
    {
	if(!(SONIC->Command & 2))
	{
	    printf("CTDA: 0x%06x, ctda: 0x%06x\n",
		   (SONIC->UpperTransmitDescriptorAddress<<16) +
		   SONIC->CurrentTransmitDescriptorAddress,
		   ctda);
	    ctda = (struct TDA *)(((ULONG)ctda & 0xffff0000)|(ctda->link&~1));

	    printf("\nTransmit completed:");
	    if((SONIC->TransmitControl & 2))
	    {
		printf(" Byte count mismatch");
	    }
	    if((SONIC->TransmitControl & 4))
	    {
		printf(" FIFO underrun");
	    }
	    if((SONIC->TransmitControl & 8))
	    {
		if(!(SONIC->InterruptStatus & 1)) /* Receiver FIFO Overrun */
		    printf(" Packet monitored bad");
		else
		    printf(" Packet not monitored");
	    }
	    if((SONIC->TransmitControl & 16))
	    {
		printf(" ???");
	    }
	    if((SONIC->TransmitControl & 32))
	    {
		printf(" Out of window collision");
	    }
	    if((SONIC->TransmitControl & 64))
	    {
		printf(" Excessive collisions");
	    }
	    if((SONIC->TransmitControl & 128))
	    {
		printf(" CRS lost");
	    }
	    if((SONIC->TransmitControl & 256))
	    {
		printf(" No CRS");
	    }
	    if((SONIC->TransmitControl & 512))
	    {
		printf(" Deferred transmission");
	    }
	    if((SONIC->TransmitControl & 1))
	    {
		printf(" No internal errors");
	    }
	    printf("\n");
	    break;

	}
	if(i)
	{
	    printf("---\n");
	    printf("Transmit not completed!\n");
	    printf("Command Register: 0x%04x\n", SONIC->Command);
	    printf("Transmit Control: 0x%04x\n", SONIC->TransmitControl);
	    printf("Receive Control:  0x%04x\n", SONIC->ReceiveControl);
	    printf("Interrupt Status: 0x%04x\n", SONIC->InterruptStatus);
	}
	Delay(100);
    }

    while(1)
    {
	/* Wait for a packet to arrive */
	if((SONIC->InterruptStatus & 16))
	{
	    SONIC->InterruptStatus = 16;
	    printf("Receive buffer area exceeded\n");
	}
	if((SONIC->InterruptStatus & 32))
	{
	    SONIC->InterruptStatus = 32;
	    printf("Receive buffer exhausted\n");
	}
	if((SONIC->InterruptStatus & 64))
	{
	    SONIC->InterruptStatus = 64;
	    printf("Receive descriptors exceeded\n");
	}
	if((SONIC->InterruptStatus & 1024))
	{
	    SONIC->InterruptStatus = 1024;
	    printf("Packet received!\n");
	    show_received(crda);
	    crda++;
	    break;
	}
	if(!(status & STAT_WAIT))
	    break;
	Delay(100);
	if((SetSignal(0,SIGF_CTRL_C) & SIGF_CTRL_C))
	{
	    printf("**break\n");
	    break;
	}
    }

    SONIC->Command = 0x8080; /* Reset */

    /* cleanup and exit */
    for(i=0;i<NUM_RBA;i++)
    {
	if(scoreBoard[i].buffer)
	    free(scoreBoard[i].buffer);
    }
    free(sonicmem);
    return 0;
}

