#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include <exec/execbase.h>
#include <exec/exec_protos.h>
#include <exec/tasks.h>
#include <exec/io.h>
#include <dos/dos.h>

#include "internal.h"

#define HISTORY  40
#define ARGLEN   256

#define DUMB	0
#define VT100	1
#define ANSI	2

#define MODE_NORMAL	0
#define MODE_ESC	1
#define MODE_CSI	2

void FPutC(APTR io, char a)
{
    Write(io, &a, 1);
}


static void unescapestring(char *str)
{
    while(*str)
    {
	switch(*str++)
	{
	case '\\':
	    switch(*str)
	    {
	    case 'n':
		*str = 0x0a;
		break;
	    case 't':
		*str = 0x09;
		break;
	    case 'a':
		*str = 0x07;
		break;
	    case 'e':
		*str = 0x1b;
		break;
	    case 'r':
		*str = 0x0d;
		break;
	    }
	    strcpy(str-1, str);
	    break;
	case '^':
	    if(*str)
	    {
		*str = (*str & 0x1f);
	    }
	    strcpy(str-1, str);
	    break;
	}
    }
}

static void escapestring(char *str)
{
    while(*str)
    {
	switch(*str)
	{
	case ' ':
	    if(!*(str+1)) /* do not escape trailing space */
	    {
		str++;
		break;
	    }
	case '\"':
	case '\\':
	case '<':
	case '>':
	case '$':
	    memmove(str+1, str, strlen(str)+1);
	    *str++ = '\\';
	    str++;
	    break;
	default:
	    str++;
	}
    }
}

int GetLine(APTR io, char *pr, char *a, int length, BOOL echo)
{
    static int func=-1, histc=0;	/* history count */
    static char history[HISTORY][ARGLEN]; /* table for history */
    static char funcstr[ARGLEN+1], killbuf[ARGLEN+1];
    char c=0, temp[ARGLEN+2];	/* temporary for expansions etc. */
    int done=0, term=DUMB, hist=histc, oldhist=histc, altc=0, mode=MODE_NORMAL;
    long i=0, k, l, j, error = 0;
    WORD oldpri = SetTaskPri(FindTask(NULL),2);

    static int commandcount=0;

    int len, Columns = 80;
    char tmp[80];

    if(GetVar("term", temp, sizeof(temp), LV_VAR) == -1)
    {
	term = DUMB;
    }
    else
    {
	if(!strncmp(temp,"vt100",5) || !strncmp(temp,"vt102",5))
	    term = VT100;
	else
	    if(!strncmp(temp,"ansi",4))
		term = ANSI;
    }
    if(GetVar("columns", temp, sizeof(temp), LV_VAR) != -1)
    {
	Columns = atoi(temp);
	if(Columns < 8)
	    Columns = 8;
    }

    SetMode(Input(), MODE_RAW);
    SetSignal(0, SIGF_CTRL_C | SIGF_CTRL_D);
    i = 0;
    a[0] = '\0';

    while(!done)
    {
	if(func != -1)
	{
	    len = 0;
	    while(funcstr[func])
	    {
		if((SetSignal(0L, SIGF_CTRL_C) & SIGF_CTRL_C))
		{
		    PutStr("^C");
		    mode = 0;
		    func = -1;
		    continue;
		}
		switch(c = funcstr[func++])
		{
		case '\\':
		    if(funcstr[func])
		    {
			switch(c = funcstr[func++])
			{
			case 'n':
			    c=0x0a;
			    break;
			case 't':
			    c=0x09;
			    break;
			case 'a':
			    c=0x07;
			    break;
			case 'e':
			    c=0x1b;
			    break;
			case 'r':
			    c=0x0d;
			    break;
			}
		    }
		    break;
		case '^':
		    if(funcstr[func])
		    {
			c = (funcstr[func++] & 0x1f);
		    }
		    break;
		}
		tmp[len++] = c;
	    }
	    func = -1;
	}
	else
	{
	    Delay(2); /* wait for 10-20ms -- makes it more efficient to
			 read big chunks of data but does not hurt the
			 response times so much. */
	    len = Read(io, tmp, 78);
	    if(!len)
	    {
		done = 1;
		SetTaskPri(FindTask(NULL), oldpri);
		SetMode(Input(), MODE_CONSOLE);
		return -1;  /* EOF */
	    }
	}

	for(k=0;k<len;k++)
	{
	    c = tmp[k];
	    if(mode == MODE_NORMAL && c == 0x9b)
	    {
		mode = MODE_CSI;
		altc = 0;
		continue;
	    }

	    if((SetSignal(0, SIGF_CTRL_C) & SIGF_CTRL_C) || c==0x03)
	    {
		PutStr("^C");
		c = 0x0c;	/* refresh line */
		mode = MODE_NORMAL;
	    }
	    if(mode == MODE_NORMAL)
	    {
		switch(c)
		{
		case 0x09:		/* TAB (expand command/filename) */
		    if(i && echo)
		    {
			j = i-1;
			while(j>=0 && (a[j]!=' ' || (j>0 && a[j-1]=='\\')) &&
			      a[j]!='>' && a[j]!='\"')
			    j--;
			strcpy(temp,a+j+1);
			temp[i-j-1] = '\0';

			oldpri = SetTaskPri(FindTask(NULL),oldpri);

			unescapestring(temp);
			if(j >= 0)
			    fcomp(temp);
			else
			    ccomp(temp);

			oldpri = SetTaskPri(FindTask(NULL),oldpri);
			escapestring(temp);

			if(strlen(temp)+strlen(a)-i+j+2<length)
			{
			    l = strlen(temp)+j+1;
			    if(a[i])
			    {
				if(a[i] == ' ' && temp[strlen(temp)-1] == ' ')
				{
				    strcat(temp,a+i+1);
				}
				else
				{
				    strcat(temp,a+i);
				}
				strcpy(a+j+1,temp);
			    }
			    else
			    {
				strcpy(a+j+1,temp);
			    }
			    PutStr(a+i);
			    
			    i=strlen(a+l);
			    if(i)
			    {
				if(term == VT100)
				{
				    sprintf(temp,"\033[%ldD",i);
				    PutStr(temp);
				}
				else
				{
				    if(term == DUMB)
				    {
					for(j=i;j>0;j--)
					    PutStr("\b");
				    }
				    else
				    {
					for(j=i;j>0;j--)
					    PutStr("\033[D");
				    }
				}
			    }
			    i=l;
			    oldhist=hist=histc;
			}
		    }
		    break;
		case 0x0c:	/* ^L - refresh*/
		    PutStr("\n");
		    PutStr(pr);
		    if(term != DUMB)
			PutStr("\033[K");
		    if(echo)
			PutStr(a);
		    if(j = strlen(a)-i)
		    {
			if(term == VT100)
			{
			    sprintf(temp,"\033[%ldD",j);
			    PutStr(temp);
			}
			else
			{
			    if(term == ANSI)
			    {
				for(;j>0;j--)
				    PutStr("\033[D");
			    }
			    else
			    {
				for(;j>0;j--)
				    FPutC(Output(),'\b');
			    }
			}
		    }
		    break;
		case 0x0a:	/* LF or CR */
		case 0x0d:
		    PutStr(a+i);
		    done=1;
		    break;
		case 0x1b:
		    mode = MODE_ESC;
		    break;
		case 0x9b:
		    mode = MODE_CSI;
		    altc = 0;
		    break;
		case 0x10:
		    if(hist)
			hist--;
		    break;
		case 0x0e:
		    if(hist < histc)
			hist++;
		    break;
		case 0x06:	/* crsr right */
		    if(i < strlen(a))
		    {
			if(echo)
			{
			    if(term == DUMB)
				FPutC(Output(),a[i]);
			    else
				PutStr("\033[C");
			}
			i++;
		    }
		    break;
		case 0x02:	/* crsr left */
		    if(i)
		    {
			if(echo)
			{
			    if(term == DUMB)
				FPutC(Output(),'\b');
			    else
				PutStr("\033[D");
			}
			i--;
		    }
		    break;
		case 0x01:		/* ^a - goto start of line */
		    if(echo)
		    {
			if(term == VT100)
			{
			    if(i)
			    {
				sprintf(temp,"\033[%ldD",i);
				PutStr(temp);
			    }
			}
			else
			{
			    while(i)
			    {
				if(term == DUMB)
				    FPutC(Output(),'\b');
				else
				    PutStr("\033[D");
				i--;
			    }
			}
		    }
		    i = 0;
		    break;
		case 0x05:		/* ^e - goto eol */
		    if(echo)
			PutStr(a+i);
		    i = strlen(a);
		    break;
		case 0x04:		/* ^D */	
		    /* If on the EOL, show completions */
		    if(i && a[i] == '\0' && echo)
		    {
			j = i-1;
			while(j>=0 && (a[j]!=' ' || (j>0 && a[j-1]=='\\')) &&
			      a[j]!='>' && a[j]!='\"')
			    j--;
			strcpy(temp,a+j+1);
			temp[i-j-1]=0;
			PutStr("\n");

			/* See if we have spaces before the command */
			l=j;
			while(l>=0 && a[l] == ' ')
			    l--;
			oldpri = SetTaskPri(FindTask(NULL), oldpri);

			if(l>=0)
			    fcomplist(temp, Columns);
			else
			    ccomplist(temp, Columns);
			
			oldpri = SetTaskPri(FindTask(NULL), oldpri);

			PutStr(pr);
			if(term != DUMB)
			    PutStr("\033[K");
			if(echo)
			    PutStr(a);
			break;
		    }
		    /* Fall through */
		case 0x7f:		/* DEL & ^D */
		    if(a[i])
		    {
			strcpy(a+i,a+i+1);
			if(echo)
			{
			    if(term == VT100)
			    {
				PutStr("\033[P");
			    }
			    else
			    {
				if(term == DUMB)
				{
				    PutStr(a+i);
				    PutStr(" ");
				    for(j=strlen(a+i)+1;j>0;j--)
					FPutC(Output(),'\b');
				}
				else
				{
				    sprintf(temp,"%s\033[%ldD",a+i,strlen(a+i)+1);
				    PutStr(temp);
				}
			    }
			}
			oldhist=hist=histc;
		    }
		    break;
		case 0x08:		/* BS */
		    if(i>0)
		    {
			i--;
			j=i;
			while(a[j]=a[j+1])
			    j++;
			if(echo)
			{
			    PutStr("\b");
			    PutStr(a+i);
			    PutStr(" ");
			    if(term == VT100)
			    {
				sprintf(temp,"\033[%ldD",strlen(a+i)+1);
				PutStr(temp);
			    }
			    else
			    {
				if(term == ANSI)
				{
				    for(j=strlen(a+i)+1;j>0;j--)
					PutStr("\033[D");
				}
				else
				{
				    for(j=strlen(a+i)+1;j>0;j--)
					FPutC(Output(),'\b');
				}
			    }
			}
			oldhist=hist=histc;
		    }
		    break;
		case 0x0b:	/* ^K - delete to the end of line */
		    if(term != DUMB)
		    {
			PutStr("\033[K");
		    }
		    else
		    {
			for(j=strlen(a+i);j>0;j--)
			    FPutC(Output(),' ');
			for(j=strlen(a+i);j>0;j--)
			    FPutC(Output(),'\b');
		    }
		    strcpy(killbuf,a+i);
		    a[i]='\0';
		    oldhist=hist=histc;
		    break;
		case 0x15:	/* ^U - clear start of the line */
		    if(i)
		    {
			if(term == VT100)
			{
			    sprintf(temp, "\033[%ldD", i);
			    PutStr(temp);
			}
			else
			{
			    if(term == DUMB)
			    {
				for(j=i;j>0;j--)
				    FPutC(Output(),'\b');
			    }
			    else
			    {
				for(j=i;j>0;j--)
				    PutStr("\033[D");
			    }
			}
			j=strlen(a);	/* remember old length */
			strncpy(killbuf,a,i);
			killbuf[i]='\0';
			strcpy(a,a+i);
			
			if(term == VT100)
			{
			    if(a[0])
			    {
				sprintf(temp,"%s\033[K\033[%ldD",(LONG)a,strlen(a));
				PutStr(temp);
			    }
			    else
				PutStr("\033[K");
			}
			else
			{
			    if(term == DUMB)
			    {
				PutStr(a);
				for(i=strlen(a);i<j;i++)
				    FPutC(Output(),' ');
				for(;i>0;i--)
				    FPutC(Output(),'\b');
			    }
			    else
			    {
				PutStr(a);
				PutStr("\033[K");
				for(j=strlen(a);j>0;j--)
				    PutStr("\033[D");
			    }
			}
			i=0;
			oldhist=hist=histc;
		    }
		    break;
		case 0x17:	/* ^W - delete previous word */
		    if(i)
		    {
			j=i;
			if(a[i-1] == ' ')
			{
			    FPutC(Output(),'\b');
			    j--;
			}
			else
			{
			    while(j && a[j-1] != ' ')
			    {
				FPutC(Output(),'\b');
				j--;
			    }
			}
			strncpy(killbuf,a+j,i-j);
			killbuf[i-j]='\0';
			strcpy(a+j,a+i);
			
			if(term == VT100)
			{
			    if(a[j])
			    {
				sprintf(temp,"%s\033[K\033[%ldD",(LONG)a+j,strlen(a+j));
				PutStr(temp);
			    }
			    else
				PutStr("\033[K");
			}
			else
			{
			    if(term == DUMB)
			    {
				PutStr(a+j);
				for(c=i-j;c>0;c--)
				    FPutC(Output(),' ');
				for(i=strlen(a+j)+i-j;i>0;i--)
				    FPutC(Output(),'\b');
				c=0x17;	/* Just be sure it has no side effects */
			    }
			    else
			    {
				PutStr(a+j);
				PutStr("\033[K");
				for(i=strlen(a+j);i>0;i--)
				    PutStr("\033[D");
			    }
			}
			i=j;
			oldhist=hist=histc;
		    }
		    break;
		case 0x18:	/* ^X - clear line */
		    if(strlen(a))
		    {
			if(term == DUMB)
			{
			    for(j=strlen(a+i);j>0;j--)
				FPutC(Output(),' ');
			    for(j=strlen(a);j>0;j--)
				FPutC(Output(),'\b');
			    for(j=strlen(a);j>0;j--)
				FPutC(Output(),' ');
			    for(j=strlen(a);j>0;j--)
				FPutC(Output(),'\b');
			}
			else
			{
			    if(i)
			    {
				if(term == VT100)
				{
				    sprintf(temp,"\033[%ldD",i);
				    PutStr(temp);
				}
				else
				{
				    for(j=i;j>0;j--)
					PutStr("\033[D");
				}
			    }
			    PutStr("\033[K");
			}
			strcpy(killbuf,a);
			a[0]='\0';
			i=0;
			oldhist=hist=histc;
		    }
		    break;
		case 0x19:	/* ^Y - yank killed text back */
		    strcpy(funcstr, killbuf);
		    func=0;
		    break;
		case 0x1c:	/* ^\, EOF */
		    error = -1;
		    SetTaskPri(FindTask(NULL), oldpri);
		    SetMode(Input(), MODE_CONSOLE);
		    return -1;
		    
		default:
		    if(strlen(a)<length-1 && c>0x1f && (c<0x80 || c>0x9f)) /* we don't want control chars */
		    {
			for(j=strlen(a) ; j>=i ; j--)
			    a[j+1] = a[j];
			a[i] = c;
			if(echo)
			{
			    PutStr(a+i);
			    if(a[i+1])
			    {
				if(term == VT100)
				{
				    sprintf(temp,"\033[%ldD",strlen(a+i+1));
				    PutStr(temp);
				}
				else
				{
				    if(term == DUMB)
				    {
					for(j=strlen(a+i+1);j>0;j--)
					    PutStr("\b");
				    }
				    else
				    {
					for(j=strlen(a+i+1);j>0;j--)
					    PutStr("\033[D");
				    }
				}
			    }
			}
			i++;
			oldhist = hist = histc;
		    }
		}
	    }
	    else
	    {
		if(mode == MODE_ESC)
		{
		    if(c == 0x5b)
		    {
			mode=MODE_CSI;	/* if it is [, we are in the sequence mode */
			altc=0;
		    }
		    else
		    {
			mode=MODE_NORMAL;
		    }
		}
		else
		if(mode == MODE_CSI)
		{
		    switch(c)
		    {
		    case 'A':
			temp[altc]=0;
			if(altc == 0)
			{
			    if(hist)		/* crsr up */
				hist--;
			}
			else
			{
			    PutStr("<csi>");
			    PutStr(temp);
			    PutStr("A\n");
			}
			mode = MODE_NORMAL;
			break;
		    case 'B':
			if(hist < histc)	/* crsr down */
			    hist++;
			mode = MODE_NORMAL;
			break;
		    case 'C':
			if(i < strlen(a))	/* crsr right */
			{
			    PutStr("\033[C");
			    i++;
			}
			mode = MODE_NORMAL;
			break;
		    case 'D':
			if(i)			/* crsr left */
			{
			    PutStr("\033[D");
			    i--;
			}
			mode = MODE_NORMAL;
			break;
		    case '~':
			temp[altc] = '\0';
			altc = 0;
			if(temp[0] == '?')
			{
			    strcpy(funcstr,"HELP");
			    if(GetVar(funcstr,funcstr,ARGLEN,LV_VAR) != -1)
				func = 0;
			    /* HELP */
			}
			else
			{
			    func = atol(temp);	/* get the key name */
			    if(func>22)
			    {
				func -= 12;
			    }
			    else if(func>16)
			    {
				func -= 11;
			    }
			    else if(func>10)
			    {
				func -= 10;
			    }
			    sprintf(funcstr,"F%ld", func);

			    if(GetVar(funcstr,funcstr,ARGLEN,LV_VAR) != -1)
				func = 0;
			    else
				func = -1;
			}
			mode = MODE_NORMAL;
			break;
		    default:
			if(altc < ARGLEN)
			    temp[altc++] = c;
			break;
		    }
		}
		
		if(hist != oldhist && echo)
		{
		    /* if we have selected an event */
		    strcpy(a, history[hist]);
		    if(i)
		    {
			if(term == VT100)
			{
			    /* print new event and delete the rest of line */
			    sprintf(temp,"\033[%ldD",i);
			    PutStr(temp);
			}
			else
			{
			    if(term == DUMB)
			    {
				for(j=i; j>0; j--)
				    PutStr("\b");
			    }
			    else
			    {
				for(j=i; j>0; j--)
				    PutStr("\033[D");
			    }
			}
		    }
		    PutStr(a);
		    PutStr("\033[K");		/* print new event */
		    i = strlen(a);
		    oldhist = hist;
		}
	    }
	}
    }
    PutStr("\n");

    if((hist == histc || hist<histc-1) && echo && a[0])
    {	/* if the line was a new one or too far away and non-blank, add event*/
	commandcount++;
	strcpy(history[histc++], a);
	if(histc == HISTORY) /* max # of events reached, let's forget the oldest */
	{
	    for(i=0; i<histc; i++)
		strcpy(history[i], history[i+1]);
	    history[--histc][0] = '\0';
	}
    }
    SetTaskPri(FindTask(NULL), oldpri);
    SetMode(Input(), MODE_CONSOLE);
    return 0;
}

int fcomp(char *keyname)
{
    struct	FileInfoBlock *fib;
    APTR 	mylock;
    char	*found, *temppi, *name1, *name2;
    int		len, i=0, q;
    BOOL	done=FALSE, dir=FALSE;
    const	int size=4*ARGLEN+sizeof(struct FileInfoBlock);

    if((fib=(struct FileInfoBlock *)AllocMem(size,MEMF_CLEAR)))
    {
	found = (char *)fib+sizeof(struct FileInfoBlock);
	name2=(name1=(temppi=found+ARGLEN)+ARGLEN)+ARGLEN;

	found[0] = '\0';
	strcpy(temppi, keyname);
	i = strlen(temppi);
	while(i>=0 && temppi[i] != '/' && temppi[i] != ':')
	    i--;
	q = i+1;
	strcpy(name1, temppi+q);
	len = strlen(name1);
	temppi[q] = '\0';
	
	if(mylock=Lock(temppi,ACCESS_READ))
	{
	    Examine(mylock,fib);
	    while(ExNext(mylock,fib))
	    {
		if(!strncmp(fib->fib_FileName,name1,len))
		{
		    if(found[0])
		    {
			i=len;
			while(found[i] == fib->fib_FileName[i] && found[i])
			    i++;
			if(found[i])	/* filenames aren't identical */
			{
			    found[i] = '\0';
			    strcpy(name2, fib->fib_FileName);
			    name2[i] = '\0';
			    done = FALSE;
			}
		    }
		    else								/* the first match */
		    {
			strcpy(name2, fib->fib_FileName);
			strcpy(found, name2);
			if(fib->fib_DirEntryType > 0)
			    dir = TRUE;
			done = TRUE;
			if(!name1[0])
			    break;
		    }
		}
	    }
	    UnLock(mylock);
	    if(dir && done)
		strcat(name2,"/");	/* we have a directory */
	    if(!dir && done)
		strcat(name2," ");	/* we have a complete filename */
	    if(found[0])
		strcpy(keyname+q, name2);
	}
	FreeMem(fib, size);
    }
    if(!done)
    {
	PutStr("\a");
	return 0;
    }
    else
    {
	return -1;
    }
}


void fcomplist(char *keyname,int Columns)
{
    struct	FileInfoBlock *fib;
    APTR 	mylock;
    char	*temppi, *name1;
    int		len, i, q, num=0;
    const	int size=2*ARGLEN+sizeof(struct FileInfoBlock);

    if((fib=(struct FileInfoBlock *)AllocMem(size,MEMF_CLEAR)))
    {
	temppi = (char *)fib + sizeof(struct FileInfoBlock);
	name1 = temppi + ARGLEN;

	strcpy(temppi, keyname);
	i = strlen(temppi);
	while(i>=0 && temppi[i] != '/' && temppi[i] != ':')
	    i--;
	q = i+1;
	strcpy(name1, temppi+q);
	len = strlen(name1);
	temppi[q] = '\0';
	
	if(mylock=Lock(temppi,ACCESS_READ))
	{
	    Examine(mylock, fib);
	    while(ExNext(mylock, fib))
	    {
		if(!strncmp(fib->fib_FileName, name1, len))
		{
		    num += strlen(fib->fib_FileName)+2;
		    if(fib->fib_DirEntryType>0)
		    {
			if(num > Columns)
			{
			    PutStr("\n");
			    PutStr(fib->fib_FileName);
			    PutStr("/ ");
			    num = strlen(fib->fib_FileName)+2;
			}
			else
			{
			    PutStr(fib->fib_FileName);
			    PutStr("/ ");
			}
		    }
		    else
		    {
			if(num > Columns)
			{
			    PutStr("\n");
			    PutStr(fib->fib_FileName);
			    PutStr("  ");
			    num = strlen(fib->fib_FileName)+2;
			}
			else
			{
			    PutStr(fib->fib_FileName);
			    PutStr("  ");
			}
		    }
		}
	    }
	    UnLock(mylock);
	    if(num)
		PutStr("\n");
	}
	FreeMem(fib, size);
    }
}


