#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include <exec/execbase.h>
#include <exec/exec_protos.h>
#include <exec/tasks.h>
#include <exec/io.h>
#include <dos/dos.h>

#include "internal.h"

#define HASH_BUFINITIAL	2048	/* Buffer size at first allocation */
#define HASH_NUMINITIAL	256	/* Name pointers at first allocation */
#define HASH_BUFDELTA	1024	/* Buffer growth at overflow */
#define HASH_NUMDELTA	128	/* Name pointer increase at overflow */

struct CmdCache
{
    char *buffer;
    long  buffersize;
    long  bufferindex;
    long  commands;
    long  maxcommands;
    short *indices;
};

static struct CmdCache CommandCache={NULL,0,0,0,NULL};

LONG SystemInfo(int argc, char *argv[]);
LONG RemoveProgram(int argc, char *argv[]);
LONG UploadProgram(int argc, char *argv[]);
ULONG ReceiveBase64(char *buffer, ULONG size);

LONG DoAlias(int argc, char *argv[]);
LONG DoAssign(int argc, char *argv[]);
LONG DoAvail(int argc, char *argv[]);
LONG DoCd(int argc, char *argv[]);
LONG DoChmod(int argc, char *argv[]);
LONG DoCopy(int argc, char *argv[]);
LONG DoEcho(int argc, char *argv[]);
LONG DoFilenote(int argc, char *argv[]);
LONG DoFormat(int argc, char *argv[]);
LONG DoInfo(int argc, char *argv[]);
LONG DoInternal(int argc, char *argv[]);
LONG DoLogin(int argc, char *argv[]);
LONG DoList(int argc, char *argv[]);
LONG DoMakedir(int argc, char *argv[]);
LONG DoMonitor(int argc, char *argv[]);
LONG DoPath(int argc, char *argv[]);
LONG DoPrgtofile(int argc, char *argv[]);
LONG DoRehash(int argc, char *argv[]);
LONG DoRelabel(int argc, char *argv[]);
LONG DoRename(int argc, char *argv[]);
LONG DoReset(int argc, char *argv[]);
LONG DoRm(int argc, char *argv[]);
LONG DoSet(int argc, char *argv[]);
LONG DoTop(int argc, char *argv[]);
LONG DoType(int argc, char *argv[]);
LONG DoUnalias(int argc, char *argv[]);
LONG DoUnset(int argc, char *argv[]);
LONG DoWhich(int argc, char *argv[]);

struct Internal Internals[]=
{
    "v", &SystemInfo,
    "D", &RemoveProgram,
    "L", &UploadProgram,
    "alias", &DoAlias,
    "assign", &DoAssign,
    "avail", &DoAvail,
    "cd", &DoCd,
    "chmod", &DoChmod,
    "copy", &DoCopy,
    "echo", &DoEcho,
    "filenote", &DoFilenote,
    "format", &DoFormat,
    "info", &DoInfo,
    "internal", &DoInternal,
    "login", &DoLogin,
    "list", &DoList,
    "mkdir", &DoMakedir,
    "mon", &DoMonitor,
    "path", &DoPath,
    "prgtofile", &DoPrgtofile,
    "rehash", &DoRehash,
    "relabel", &DoRelabel,
    "rename", &DoRename,
    "reset", &DoReset,
    "rm", &DoRm,
    "set", &DoSet,
    "top", &DoTop,
    "type", &DoType,
    "unalias", &DoUnalias,
    "unset", &DoUnset,
    "which", &DoWhich,
    NULL, NULL
};

static char *FilePart(char *src)
{
    char *a = src;

    while(*src)
    {
	if(*src==':' || *src=='/')
	{
	    a = src+1;
	}
	src++;
    }
    return a;
}



void putbits(char *buffer, ULONG protection)
{
    int i;

    strcpy(buffer, "hspadrwedrwedrwe");

    for(i=0;i<4;i++)
    {
	if(!(protection & (FIBF_ARCHIVE<<i)))
	    buffer[3-i] = '-';
	if(!(protection & (FIBF_EXECUTE<<i)))
	    buffer[7-i] = '-';
	if(!(protection & (FIBF_GRP_EXECUTE<<i)))
	    buffer[11-i] = '-';
	if(!(protection & (FIBF_OTR_EXECUTE<<i)))
	    buffer[15-i] = '-';
    }
}

LONG SystemInfo(int argc, char *argv[])
{
    struct ExecBase *SysBase = *((struct ExecBase **)4);
    struct DateStamp time;
    char out[80], *a;
    UBYTE fi[3];

    DateStamp(&time);
    if(time.ds_Days < 0)
    {
	sprintf(out, "BOAR time: not set\n");
    }
    else
    {
	sprintf(out, "BOAR time: %ld days, %2ld:%02ld:%02ld\n",
		time.ds_Days,
		time.ds_Minutes/60,
		time.ds_Minutes%60,
		time.ds_Ticks/100);
    }
    PutStr(out);
    sprintf(out, "MPU version: %d\n", (long)*(UWORD *)0x65067e);
    PutStr(out);

    /* Display FLASH data */
    PutStr("FLASH Information:\n");
    Forbid();
    /* FLASH Autoselect command */
    *((volatile UWORD *)0x84aaaa) = 0xaaaa;
    *((volatile UWORD *)0x845554) = 0x5555;
    *((volatile UWORD *)0x84aaaa) = 0x9090;
    /* Only take the info from the odd (MSB) memory chips */
    fi[0] = *((volatile UBYTE *)0x840000);
    fi[1] = *((volatile UBYTE *)0x840002);
    fi[2] = *((volatile UBYTE *)0x840004);
    /* FLASH Read/Reset command */
    *((volatile UWORD *)(0x84aaaa)) = 0xaaaa;
    *((volatile UWORD *)(0x845554)) = 0x5555;
    *((volatile UWORD *)(0x84aaaa)) = 0xf0f0;
    Permit();
    sprintf(out, "\tCompany: %d Device: 0x%02x Protect: 0x%02x\n\n",
	    fi[0], fi[1], fi[2]);
    PutStr(out);

#define BLOCKSIZE 2048
    if((a = (char *)AllocMem(BLOCKSIZE, 0)))
    {
	char *b;
	struct Program *prg;
	struct Library *lib;
	struct DosList *dol;

	Forbid();
	b = a;
	b += sprintf(b, "Resident Programs:\n");
	b += sprintf(b, "Name                    Start   Code   Data    BSS RelText RelData\n");
	prg = (struct Program *)SysBase->ProgramList.lh_Head;
	while((prg->pm_Node.ln_Next) && (b-a<BLOCKSIZE-80))
	{
	    if((prg->pm_Exec->exe_Magic & 0xffff) == 0x0107 &&
	       prg->pm_Exec->exe_Text)
	    {
		b += sprintf(b,"%-20s 0x%06x %6ld %6ld %6ld %7ld %7ld\n",
			     prg->pm_Node.ln_Name,
			     (ULONG)prg->pm_Exec,
			     prg->pm_Exec->exe_Text,
			     prg->pm_Exec->exe_Data,
			     prg->pm_Exec->exe_BSS,
			     prg->pm_Exec->exe_TRSize,
			     prg->pm_Exec->exe_DRSize);
	    }
	    else
	    {
		b += sprintf(b,"%-20s 0x%06x (Not a valid executable)\n",
			     prg->pm_Node.ln_Name,
			     (ULONG)prg);
	    }
	    prg = (struct Program *)prg->pm_Node.ln_Next;
	}
	Permit();
	Write(Output(), a, b-a);
	PutStr("\n");

	Forbid();
	b = a;
	b += sprintf(b,"Libraries:\n");
	b += sprintf(b,"Name                  LibBase Vers Opens\n");
	lib = (struct Library *)SysBase->LibList.lh_Head;
	while((lib->lib_Node.ln_Next) && (b-a<BLOCKSIZE-80))
	{
	    b += sprintf(b, "%-20s 0x%06x %4ld %5ld\n",
			 lib->lib_Node.ln_Name,
			 (ULONG)lib,
			 lib->lib_Version,
			 lib->lib_OpenCnt);
	    lib = (struct Library *)lib->lib_Node.ln_Next;
	}
	Permit();
	Write(Output(), a, b-a);
	PutStr("\n");

	Forbid();
	b = a;
	b += sprintf(b,"Handlers:\n");
	b += sprintf(b,"Name                 DosList  MsgPort\n");
	dol = (struct DosList *)DOSBase->dl_Handlers.lh_Head;
	while((dol->dol_Node.ln_Next) && (b-a<BLOCKSIZE-80))
	{
	    b += sprintf(b, "%-20s 0x%06x 0x%06x\n",
			 dol->dol_Node.ln_Name,
			 (ULONG)dol,
			 dol->dol_Task);
	    dol = (struct DosList *)dol->dol_Node.ln_Next;
	}
	Permit();
	Write(Output(), a, b-a);
	PutStr("\n");

	Forbid();
	b = a;
	b += sprintf(b, "Devices:\n");
	b += sprintf(b, "Name                 Vers Opens\n");
	lib = (struct Library *)SysBase->DeviceList.lh_Head;
	while((lib->lib_Node.ln_Next) && (b-a<BLOCKSIZE-80))
	{
	    b += sprintf(b, "%-20s %4ld %5ld\n",
			 lib->lib_Node.ln_Name,
			 lib->lib_Version,
			 lib->lib_OpenCnt);
	    lib = (struct Library *)lib->lib_Node.ln_Next;
	}
	Permit();
	Write(Output(), a, b-a);
	PutStr("\n");

	FreeMem(a, BLOCKSIZE);
    }
    return 0;
}

LONG RemoveProgram(int argc, char *argv[])
{
    int i;
    LONG error = 0;
    struct Program *prg;

    /* D <name> */

    for(i=1;i<argc;i++)
    {
	Write(Output(), argv[i], strlen(argv[i]));
	if((prg = RemProgram(argv[i])))
	{
	    FreeMem(prg->pm_Exec, prg->pm_Size);
	    FreeMem((char *)prg,
		    sizeof(struct Program) + strlen(prg->pm_Name));
	    PutStr(": Program deleted\n");
	}
	else
	{
	    PutStr(": Program not found (or general error)!\n");
	    error = 10;
	}
    }
    return error;
}

LONG UploadProgram(int argc, char *argv[])
{
    ULONG size;
    LONG error = 0;
    struct Executable *exe;
    APTR file;

    if(argc!=3)
    {
	PutStr("Usage: L <name> <size>\n");
	return 20;
    }

    /* L <name> <size> */
    /* <name>  is used when the program is added to the system */
    /* <size>  is the program size in dec or 0x<hex> */

    if(argv[2][0] == '0' && argv[2][1] == 'x')
	size = xtol(argv[2]+2);
    else
	size = atol(argv[2]);
    if(size>0)
    {
	if((exe = (struct Executable *)AllocMem(size, 0)))
	{
	    if(!ReceiveBase64((char *)exe, size))
	    {
		if((file = Open(argv[1], MODE_NEWFILE)))
		{
		    if(Write(file, (char *)exe, size) != size)
		    {
			PrintFault(IoErr(), argv[1]);
			error = 20;
		    }
		    if(!Close(file))
		    {
			PrintFault(IoErr(), argv[1]);
			error = 20;
		    }
		}
		else
		{
		    PrintFault(IoErr(), argv[1]);
		    error = 20;
		}
	    }
	    else
	    {
		PutStr("\nProgram NOT! added\n");
		error = 20;
	    }
	    FreeMem(exe, size);
	}
	else
	{
	    PutStr("Memory allocation error!\n");
	    error = 30;
	}
    }
    else
    {
	PutStr("Too short program! :-)\n");
	error = 10;
    }
    return error;
}


ULONG ReceiveBase64(char *buffer, ULONG size)
{
    /* The Base64 alphabet (big and small letters, numbers, plus and slash) */
    static const unsigned char Base64[]=
	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";

    WORD B2B[256], *b;
    int i, len;
    int symbols = 0, pad = 0;
    char buf2[4];
    long val;
    ULONG curaddr;
    char line[80],out[80];

    /* Form a Base64->Binary translation table. If a code is not in the */
    /* base64 alphabet, the corresponding value is -1. Otherwise the table */
    /* gives the bit pattern of the original 6 bits directly. */
    b = &B2B[0];
    i = 256;
    while(--i>=0)
    {
	*b++ = -1;
    }
    /* for(i=0;i<256;B2B[i++]=-1); */
    i=0;
    while(Base64[i])
    {
        B2B[Base64[i]]=i;
        i++;
    }

    /* Receive program */
    curaddr = (ULONG)buffer;
    SetMode(Input(), MODE_RAW);
    sprintf(out, "%6ld bytes left", size);
    PutStr(out);
    while(size>0)
    {
	len = Read(Input(), line, 80); /* Base64 line len 78 + \n*/
	i = 0;
	while(size>0 && i<len)
	{
	    if(B2B[line[i]]!=-1)
	    {
		buf2[symbols++]=B2B[line[i]] & 0x3f;
		if(B2B[line[i]]==64)
		    pad++;
	    }
	    if(symbols==4)
	    {
		val = (buf2[0]<<18) | (buf2[1]<<12) | (buf2[2]<<6) | buf2[3];
		if(pad)
		{
		    if(pad==1)
		    {
			*(UBYTE *)curaddr++ = (val>>16)&0xff;
			*(UBYTE *)curaddr++ = (val>>8)&0xff;
			size -= 2;
		    }
		    else
		    {
			*(UBYTE *)curaddr++ = (val>>16)&0xff;
			size--;
		    }
		}
		else
		{
		    *(UBYTE *)curaddr++ = (val>>16)&0xff;
		    *(UBYTE *)curaddr++ = (val>>8)&0xff;
		    *(UBYTE *)curaddr++ = (val)&0xff;
		    size -= 3;
		}
		symbols = 0;
	    }
	    i++;
	}
	sprintf(out, "\r%6ld bytes left", size);
	PutStr(out);
    }
    PutStr("\n");
    SetMode(Input(), MODE_CONSOLE);
    return size;
}

LONG DoAlias(int argc, char *argv[])
{
    struct LocalVar *var;

    if(argc==1)
    {
	/* Display all aliases */
	var = (struct LocalVar *)
	    ((struct Process *)FindTask(NULL))->pr_LocalVars.lh_Head;
	while(var->lv_Node.ln_Next)
	{
	    if(var->lv_Flags == LV_ALIAS)
	    {
		PutStr(var->lv_Node.ln_Name);
		PutStr(" \t\"");
		PutStr((char *)var->lv_Value);
		PutStr("\"\n");
	    }
	    var = (struct LocalVar *)var->lv_Node.ln_Next;
	}
	return RETURN_OK;
    }
    if(argc==2)
    {
	if((var = FindVar(argv[1], LV_ALIAS)))
	{
	    PutStr(var->lv_Node.ln_Name);
	    PutStr(" \t\"");
	    PutStr((char *)var->lv_Value);
	    PutStr("\"\n");
	    return RETURN_OK;
	}
	return RETURN_ERROR;
    }
    if(argc==3)
    {
	if(!SetVar(argv[1], argv[2], strlen(argv[2])+1, LV_ALIAS))
	{
	    PutStr("Not set!\n");
	    return RETURN_ERROR;
	}
	return RETURN_OK;
    }
    PutStr("Usage: alias [<name> [<value>]]\n");
    return RETURN_ERROR;
}

LONG DoUnalias(int argc, char *argv[])
{
    int i;
    LONG error = RETURN_OK;

    for(i=1;i<argc;i++)
    {
	if(!DeleteVar(argv[i], LV_ALIAS))
	{
	    PutStr(argv[i]);
	    PutStr(": not found!\n");
	    error = RETURN_WARN;
	}
    }
    return error;
}

LONG ListAssigns(char *name)
{
    struct DosList *dol;
    struct AssignList *al;
    char temp[160];

    PutStr("Assigns:\n");

    /* TODO: doslist locking */
    dol = (struct DosList *)DOSBase->dl_Handlers.lh_Head;
    while(dol->dol_Node.ln_Next)
    {
	if(dol->dol_Node.ln_Type == DLT_ASSIGN &&
	   (!name || !stricmp(dol->dol_Node.ln_Name, name)))
	{
	    if(dol->dol_misc.dol_assign.dol_AssignName)
	    {
		sprintf(temp, "%-12s ", dol->dol_Node.ln_Name);
		PutStr(temp);
		PutStr(dol->dol_misc.dol_assign.dol_AssignName);
		PutStr(" [late]\n");
	    }
	    else
	    {
		al = (struct AssignList *)
		    dol->dol_misc.dol_assign.dol_AssignList.lh_Head;
		if(al->al_Node.lmn_Next)
		{
		    sprintf(temp, "%-12s ", dol->dol_Node.ln_Name);
		    PutStr(temp);
		    if(!getpath(temp, 160, al->al_Lock))
		    {
			PutStr("..");
		    }
		    PutStr(temp);
		    PutStr("\n");
		    al = (struct AssignList *)al->al_Node.lmn_Next;
		}
		else
		{
		    PutStr(dol->dol_Node.ln_Name);
		    PutStr("\n");
		}
		while(al->al_Node.lmn_Next)
		{
		    PutStr("             ");
		    if(!getpath(temp, 160, al->al_Lock))
		    {
			PutStr("..");
		    }
		    PutStr(temp);
		    PutStr("\n");
		    al = (struct AssignList *)al->al_Node.lmn_Next;
		}
	    }
	}
	dol = (struct DosList *)dol->dol_Node.ln_Next;
    }
    /* TODO: doslist unlocking */
    return RETURN_OK;
}

/* "assign" display assigns */
/* "assign name: ( [REMOVE] | directory [(LATE | ADD)] )" */
LONG DoAssign(int argc, char *argv[])
{
    APTR lock;

    if(argc<2)
    {
	return ListAssigns(NULL);
    }
    if(argc==2)
    {
	return ListAssigns(argv[1]);
    }
    if(argc==3)
    {
	if(!stricmp(argv[2], "REMOVE"))
	{
	    if(!AssignLock(argv[1], NULL))
	    {
		PrintFault(IoErr(), argv[1]);
		return RETURN_ERROR;
	    }
	    return RETURN_OK;
	}
	if((lock = Lock(argv[2], ACCESS_READ)))
	{
	    if(!AssignLock(argv[1], lock))
	    {
		PrintFault(IoErr(), argv[1]);
		UnLock(lock);
		return RETURN_ERROR;
	    }
	    return RETURN_OK;
	}
	PrintFault(IoErr(), argv[2]);
	return RETURN_ERROR;
    }
    if(argc==4)
    {
	if(!stricmp(argv[3], "LATE"))
	{
	    if(!AssignLate(argv[1], argv[2]))
	    {
		PrintFault(IoErr(), argv[1]);
		return RETURN_ERROR;
	    }
	    return RETURN_OK;
	}
	else if(!stricmp(argv[3], "ADD"))
	{
	    if((lock = Lock(argv[2], ACCESS_READ)))
	    {
		if(!AssignAdd(argv[1], lock))
		{
		    PrintFault(IoErr(), argv[1]);
		    UnLock(lock);
		    return RETURN_ERROR;
		}
		return RETURN_OK;
	    }
	    PrintFault(IoErr(), argv[2]);
	    return RETURN_ERROR;
	}
    }
    PutStr("Usage: assign <name>: <directory> [LATE | ADD]\n");
    PutStr("       assign <name>: REMOVE\n");
    return RETURN_ERROR;
}

LONG DoAvail(int argc, char *argv[])
{
    struct ExecBase *SysBase = *((struct ExecBase **)4);
    char *a, *b;
    ULONG alloc_size = 2048;
    struct Memory *mem = SysBase->MemoryList;

    if((a = (char *)AllocMem(alloc_size, 0)))
    {
	b = a;
	b += sprintf(b, "Free Memory: %ld bytes  Largest: %ld bytes\n",
			AvailMem(0),
			AvailMem(MEMF_LARGEST));
	b += sprintf(b, "Memory list:\n");
	Forbid();
	while(mem && (b-a<alloc_size-80))
	{
	    if(mem->mm_Size < 1024)
	    {
		b += sprintf(b, "\t0x%06x-0x%06x  (0x%06x = %d B)\n",
			     mem,
			     ((int)mem)+mem->mm_Size,
			     mem->mm_Size,
			     mem->mm_Size);
	    }
	    else if(mem->mm_Size < 10240)
	    {
		b += sprintf(b, "\t0x%06x-0x%06x  (0x%06x = %d.%02d kB)\n",
			     mem,
			     ((int)mem)+mem->mm_Size,
			     mem->mm_Size,
			     (mem->mm_Size>>10),
			     ((mem->mm_Size&0x3ff)*100)>>10);
	    }
	    else
	    {
		b += sprintf(b, "\t0x%06x-0x%06x  (0x%06x = %d kB)\n",
			     mem,
			     ((int)mem)+mem->mm_Size,
			     mem->mm_Size,
			     (mem->mm_Size>>10));
	    }
	    mem = mem->mm_Next;
	}
	Permit();
	Write(Output(), a, b-a);
	FreeMem(a, alloc_size);
	return RETURN_OK;
    }
    PutStr("Memory allocation failed!\n");
    return RETURN_FAIL;
}

LONG DoCd(int argc, char *argv[])
{
    APTR lock, oldlock;

    if(argc>2)
    {
	PutStr("Usage: cd <directory>\n");
	return RETURN_ERROR;
    }
    if(argc<2)
    {
	char *buffer = (char *)AllocMem(1024, 0);

	if(buffer)
	{
	    if(getpath(buffer, 1024,
		       ((struct Process *)FindTask(NULL))->pr_CurrentDir))
	    {
		PutStr(buffer);
		PutStr("\n");
	    }
	    FreeMem(buffer, 1024);
	}
	return RETURN_OK;
    }

    if((lock = Lock(argv[1], ACCESS_READ)))
    {
	if(oldlock = CurrentDir(lock))
	{
	    UnLock(oldlock);
	    return RETURN_OK;
	}
	else
	{
	    UnLock(lock);
	    PrintFault(IoErr(), argv[1]);
	    return RETURN_ERROR;
	}
    }
    PrintFault(IoErr(), argv[1]);
    return RETURN_ERROR;
}

LONG DoChmod(int argc,char *argv[])
{
    ULONG protection, oldProtection, bit = 0;
    ULONG orMask=0, andMask=0xffffffff, xorMask=0, setProt=0xffffffff;
    char *a, buffer[20];
    int mode='+', group=0, error=RETURN_OK, set=0, addMode=0;
    long n;
    struct FileInfoBlock fib;
    APTR lock;

    if(argc > 2)
    {
	a=argv[1];
	if(isxdigit(*a))
	{
	    setProt = xtol(a);
	    while(isxdigit(*a))
		a++;
	    if(*a)
		error = RETURN_ERROR;

	    setProt = ((setProt&0xf)<<FIBB_OTR_EXECUTE) |
		(((setProt&0xf0)>>4)<<FIBB_GRP_EXECUTE) |
		(((setProt&0xf00)>>8)<<FIBB_EXECUTE) |
		(setProt&0xfffff000);
	}
	else while(*a && error==RETURN_OK)
	{
	    switch(*a)
	    {
	    case 'u':
		group   = addMode*group | 0x0001;
		addMode = 1;
		break;
	    case 'g':
		group   = addMode*group | 0x0010;
		addMode = 1;
		break;
	    case 'o':
		group   = addMode*group | 0x0100;
		addMode = 1;
		break;
	    case '+':
	    case '-':
	    case '=':
		mode = *a;
		break;
	    case 'h':
		bit   = (1<<15)/*FIBF_HIDDEN*/;
		group = 0x0001;
		set   = -1;
		break;
	    case 's':
		bit   = FIBF_SCRIPT;
		group = 0x0001;
		set   = -1;
		break;
	    case 'p':
		bit   = FIBF_PURE;
		group = 0x0001;
		set   = -1;
		break;
	    case 'a':
		bit   = FIBF_ARCHIVE;
		group = 0x0001;
		set   = -1;
		break;
	    case 'r':
		bit   = FIBF_READ;
		set   = -1;
		break;
	    case 'w':
		bit   = FIBF_WRITE;
		set   = -1;
		break;
	    case 'x':
	    case 'e':
		bit   = FIBF_EXECUTE;
		set   = -1;
		break;
	    case 'd':
		set   = -1;
		bit   = FIBF_DELETE;
		break;
	    default:
		error = RETURN_WARN;
		break;
	    }
	    if(*a != 'u' && *a != 'g' && *a != 'o')
		addMode = 0;

	    a++;
	    if(error==RETURN_OK && set)
	    {
		if(!group)
		    group = 0x0001;

		if(mode == '+')
		{
		    orMask |= (bit*group);
		}
		else if(mode == '-')
		{
		    andMask &= ~(bit*group);
		}
		else if(mode == '=')
		{
		    xorMask = (bit*group);
		}
		set = 0;
	    }
	}
    }
    else
    {
	error = RETURN_WARN;
    }
    if(error!=RETURN_OK)
    {
	PutStr("Usage: chmod [ugo][+-=][hsparwed] <files>\n");
	return error;
    }

    for(n=2; n<argc; n++)
    {
	if((lock = Lock(argv[n], ACCESS_READ)))
	{
	    if(Examine(lock, &fib))
	    {
		UnLock(lock);
		oldProtection = fib.fib_Protection;

		if(!~setProt)
		    protection =
			((oldProtection ^ xorMask) | orMask) & andMask;
		else
		    protection = setProt;

		putbits(buffer, protection);
		PutStr(buffer);
		PutStr(" ");
		PutStr(fib.fib_FileName);
		PutStr("\n");

		if(protection != oldProtection)
		{
		    if(!SetProtection(argv[n], protection))
		    {
			PrintFault(IoErr(), argv[n]);
			/*
			   return value is warn, whenever we return
			   */
			if(error==RETURN_OK)
			    error = RETURN_WARN;
		    }
		}
	    }
	    else
	    {
		UnLock(lock);
		PrintFault(IoErr(), argv[n]);
		error = RETURN_FAIL;
	    }
	}
	else
	{
	    PrintFault(IoErr(), argv[n]);
	    error = RETURN_FAIL;
	}
    }
    return error;
}


LONG DoCopy(int argc, char *argv[])
{
    APTR file, file2, lock, dirlock=NULL;
    LONG len, size, error = RETURN_OK;
    char *buffer, failed = 0, *dstname;
    struct FileInfoBlock fib;

    if(argc != 3)
    {
	PutStr("Usage: copy <file> <newfile>\n");
	return RETURN_ERROR;
    }

    dstname = argv[2];
    if((dirlock = Lock(argv[2], ACCESS_READ)))
    {
	Examine(dirlock, &fib);
	if(fib.fib_DirEntryType < 0)
	{
	    UnLock(dirlock);
	    dirlock = NULL;
	}
	else
	{
	    dstname = FilePart(argv[1]);
	}
    }

    if(!(lock = Lock(argv[1], ACCESS_READ)))
    {
	if(dirlock)
	    UnLock(dirlock);
	PrintFault(IoErr(), argv[1]);
	return RETURN_ERROR;
    }
    Examine(lock, &fib);
    UnLock(lock);
    if(fib.fib_DirEntryType > 0)
    {
	if(dirlock)
	    UnLock(dirlock);
	PutStr("Directories as sources not supported yet!\n");
	return RETURN_ERROR;
    }
    size = fib.fib_Size;

    while(!(buffer = (char *)AllocMem(size, 0)))
    {
	size /= 2;
	if(size<1024)
	{
	    if(dirlock)
		UnLock(dirlock);
	    PutStr("Memory allocation error!\n");
	    return RETURN_FAIL;
	}
    }

    if((file = Open(argv[1], MODE_OLDFILE)))
    {
	if(dirlock)
	{
	    if(!(lock = CurrentDir(dirlock)))
	    {
		PrintFault(IoErr(), argv[2]);
		UnLock(dirlock);
		dirlock = NULL;
		FreeMem(buffer, size);
		Close(file);
		DeleteFile(argv[1]);
		return RETURN_FAIL;
	    }
	}

	if((file2 = Open(dstname, MODE_NEWFILE)))
	{
	    while((len = Read(file, buffer, size)))
	    {
		if(Write(file2, buffer, len) != len)
		{
		    PrintFault(IoErr(), argv[2]);
		    failed = 1;
		    break;
		}
	    }
	    if(!failed && !len && IoErr())
	    {
		PrintFault(IoErr(), argv[1]);
		failed = 1;
	    }
	    Close(file2);

	    if(failed)
	    {
		if(DeleteFile(dstname))
		    PutStr("Destination removed.");
		error = RETURN_ERROR;
	    }
	    else
	    {
		if(fib.fib_Comment[0])
		    SetComment(dstname, fib.fib_Comment);
		SetProtection(dstname, fib.fib_Protection);
	    }
	}
	else
	{
	    PrintFault(IoErr(), argv[2]);
	    error = RETURN_ERROR;
	}
	Close(file);
	if(dirlock)
	{
	    UnLock(CurrentDir(lock));
	}
    }
    else
    {
	PrintFault(IoErr(), argv[1]);
	error = RETURN_ERROR;
    }
    FreeMem(buffer, size);
    return error;
}

LONG DoEcho(int argc,char *argv[])
{
    int i, start=1, stop=argc, noline=0;

    if(argc>start && !strcmp(argv[start],"-n"))
    {
	noline=1;
	++start;
    }
    else if(argc && !strcmp(argv[argc-1],"noline"))
    {
	noline=1;
	--stop;
    }

    if(start<stop)
    {
	PutStr(argv[start]);
	for(i=start+1;i<stop;i++)
	{
	    PutStr(" ");
	    PutStr(argv[i]);
	}
    }
    if(!noline)
	PutStr("\n");

    return RETURN_OK;
}


LONG DoFilenote(int argc, char *argv[])
{
    if(argc != 3)
    {
	PutStr("Usage: filenote <file> <note>\n");
	return RETURN_ERROR;
    }
    if(!SetComment(argv[1], argv[2]))
    {
	PrintFault(IoErr(), argv[1]);
	return RETURN_ERROR;
    }
    return RETURN_OK;
}

LONG DoFormat(int argc, char *argv[])
{
    if(argc != 2)
    {
	PutStr("Usage: format <drive>\n");
	return RETURN_ERROR;
    }
    PutStr("Formatting ");
    PutStr(argv[1]);
    PutStr("\n");
    if(Format(argv[1], 0))
    {
	return RETURN_OK;
    }
    else
    {
	PrintFault(IoErr(), argv[1]);
	return RETURN_ERROR;
    }
}

LONG DoInfo(int argc, char *argv[])
{
    struct DosList *dol;
    struct InfoData info;
    struct FileInfoBlock fib;
    APTR lock;
    char temp[160], usage[8], sizestr[8];
    ULONG size;
    int header = 0;

    if(argc>1)
    {
	PutStr("Usage: info\n");
	return RETURN_ERROR;
    }
    /* TODO: doslist locking */
    dol = (struct DosList *)DOSBase->dl_Handlers.lh_Head;
    while(dol->dol_Node.ln_Next)
    {
	if(dol->dol_Node.ln_Type != DLT_VOLUME)
	{
	    /* Not a volume */
	}
	else if((lock = Lock(dol->dol_Node.ln_Name, ACCESS_READ)))
	{
	    if(!header)
	    {
		PutStr("Mounted disks:\n");
		PutStr("Unit      Size    Used    Free Full BlkSz Name\n");
		header = 1;
	    }

	    Examine(lock, &fib);
	    if(DiskInfo(lock, &info))
	    {
		if(info.id_DiskType == ID_NO_DISK_PRESENT)
		{
		    sprintf(temp, "%-10s No disk present\n",
			    dol->dol_Node.ln_Name);
		}
		else if(info.id_DiskType == ID_NOT_REALLY_DOS)
		{
		    sprintf(temp, "%-10s Not a DOS disk\n",
			    dol->dol_Node.ln_Name);
		}
		else if(info.id_DiskType == ID_UNREADABLE_DISK)
		{
		    sprintf(temp, "%-10s Unreadable disk\n",
			    dol->dol_Node.ln_Name);
		}
		else
		{
		    strcpy(usage, " -- ");
		    if(info.id_NumBlocks)
			sprintf(usage, "%3d%%",
				info.id_NumBlocksUsed*100/info.id_NumBlocks);
		    size = info.id_NumBlocks * info.id_BytesPerBlock / 1024;
		    if(size<1000)
		    {
			sprintf(sizestr, "%3dK", size);
		    }
		    else if(size<10240)
		    {
			sprintf(sizestr, "%d.%dM", size/1024, (size&0x3ff)*10/1024);
		    }
		    else
		    {
			sprintf(sizestr, "%3dM", size/1024);
		    }
		    sprintf(temp, "%-9s %s %7d %7d %s %5d %-10s\n",
			    dol->dol_Node.ln_Name,
			    sizestr,
			    info.id_NumBlocksUsed,
			    info.id_NumBlocks - info.id_NumBlocksUsed,
			    usage,
			    info.id_BytesPerBlock,
			    fib.fib_FileName);
		}
		PutStr(temp);
	    }
	    else
	    {
	        PutStr(dol->dol_Node.ln_Name);
		PutStr("\n");
	    }
	    UnLock(lock);
	}
	else
	{
	    PrintFault(IoErr(), dol->dol_Node.ln_Name);
	}
	dol = (struct DosList *)dol->dol_Node.ln_Next;
    }

    header = 0;
    dol = (struct DosList *)DOSBase->dl_Handlers.lh_Head;
    while(dol->dol_Node.ln_Next)
    {
	if(dol->dol_Node.ln_Type == DLT_HANDLER)
	{
	    if(!header)
	    {
		PutStr("\nHandlers:\n");
		header = 1;
	    }
	    PutStr(dol->dol_Node.ln_Name);
	    PutStr("\n");
	}
	dol = (struct DosList *)dol->dol_Node.ln_Next;
    }
    return RETURN_OK;
}

LONG DoInternal(int argc,char *argv[])
{
    int i=0,j=0,Columns=80;
    char temp[40];

    PutStr("Internal commands:\n");
    while(Internals[i].name)
    {
	if(++j>Columns/12-1)
	{
	    PutStr(Internals[i].name);
	    PutStr("\n");
	    j=0;
	}
	else
	{
	    sprintf(temp, "%-12s", Internals[i].name);
	    PutStr(temp);
	}
	i++;
    }
    if(j)
	PutStr("\n");

    return RETURN_OK;
}

LONG DoLogin(int argc, char *argv[])
{
    if(argc!=2 && argc!=3)
    {
	PutStr("Usage: login <uid> [<gid>]\n");
	return RETURN_ERROR;
    }
    FindTask(NULL)->tc_UserID = atoi(argv[1]);
    if(argc==3)
    {
	FindTask(NULL)->tc_GroupID = atoi(argv[2]);
    }
    return RETURN_OK;
}


#define LS_FLAG_DIR         1  /* Show the directory itself */

static LONG showdir(char *name, LONG flags)
{
    APTR lock;
    struct FileInfoBlock fib;
    char bits[32], temp[32+108+60]; /* size for flags, filename and extra */
    LONG error = RETURN_OK;

    if((lock = Lock(name, ACCESS_READ)))
    {
	if(Examine(lock, &fib))
	{
	    if(fib.fib_DirEntryType < 0 || (flags & LS_FLAG_DIR))
	    {
		putbits(bits, fib.fib_Protection);
		if(fib.fib_DirEntryType < 0)
		{
		    sprintf(temp, "%s %-22s %7d %5d %5d\n",
			    bits,
			    fib.fib_FileName,
			    fib.fib_Size,
			    fib.fib_OwnerID,
			    fib.fib_GroupID);
		    PutStr(temp);
		}
		else
		{
		    sprintf(temp, "%s %-22s   <dir> %5d %5d\n",
			    bits,
			    fib.fib_FileName,
			    fib.fib_OwnerID,
			    fib.fib_GroupID);
		    PutStr(temp);
		}
		if(fib.fib_Comment[0])
		{
		    PutStr(": ");
		    PutStr(fib.fib_Comment);
		    PutStr("\n");
		}
	    }
	    else
	    {
		while(ExNext(lock, &fib))
		{
		    putbits(bits, fib.fib_Protection);
		    if(fib.fib_DirEntryType < 0)
		    {
			sprintf(temp, "%s %-22s %7d %5d %5d\n",
				bits,
				fib.fib_FileName,
				fib.fib_Size,
				fib.fib_OwnerID,
				fib.fib_GroupID);
			PutStr(temp);
		    }
		    else
		    {
			sprintf(temp, "%s %-22s   <dir> %5d %5d\n",
				bits,
				fib.fib_FileName,
				fib.fib_OwnerID,
				fib.fib_GroupID);
			PutStr(temp);
		    }
		    if(fib.fib_Comment[0])
		    {
			PutStr(": ");
			PutStr(fib.fib_Comment);
			PutStr("\n");
		    }
		}
		ExamineEnd(lock, &fib);
	    }
	}
	else
	{
	    PrintFault(IoErr(), name);
	    error = RETURN_ERROR;
	}
	UnLock(lock);
    }
    else
    {
	PrintFault(IoErr(), name);
	error = RETURN_ERROR;
    }
    return error;
}


LONG DoList(int argc, char *argv[])
{
    int i,n;
    LONG error = RETURN_OK, flags = 0, showed = 0;

    for(n=1;n<argc;n++)
    {
	if(argv[n][0]=='-')
	{
	    i=1;
	    while(argv[n][i])
	    {
		if(argv[n][i]=='d')
		    flags |= LS_FLAG_DIR;
		else
		{
		    PutStr("Invalid option\n");
		    return RETURN_ERROR;
		}
		i++;
	    }
	}
	else
	{
	    showed = 1;
	    error = showdir(argv[n], flags);
	}
	if(error)
	    break;
    }
    if(!showed)
	showdir("", flags);
    return error;
}

LONG DoMakedir(int argc, char *argv[])
{
    int i;
    LONG error = RETURN_OK;
    APTR lock;

    for(i=1;i<argc;i++)
    {
	if(lock = CreateDir(argv[i]))
	{
	    UnLock(lock);
	}
	else
	{
	    PrintFault(IoErr(), argv[i]);
	    error = RETURN_WARN;
	}
	if((SetSignal(0, SIGF_CTRL_C) & SIGF_CTRL_C))
	{
	    Write(Error(), "^C\n***aborted\n", 14);
	    error = RETURN_ERROR;
	    break;
	}
    }
    return error;
}

static const char intro[]=
"  _  _  _  __     _    \n"\
" /_)/ )/_)/_ /)/)/ )/|/\n"\
"/_)(_// |/_ / '/(_// / \n"\
"=-=-=-=-=-=-=-=-=-=-=-=\n";

static const char prompt[]="BOREMON > ";
static const char hex[16] ="0123456789abcdef";

static const char help[]=
"Monitor commands (case is significant)\n\r"\
"=-=-=-=-=-=-=-=-\n\r"\
"m <addr> [<addr2>]   Display memory\n\r"\
"i <addr> [<addr2>]   Display memory in ASCII\n\r"\
"f <a1> <a2> <val>    Fill memory range with the byte/2 bytes/word\n\r"\
"g <a1>               Jump to address\n\r"\
"* <a1> <val>         Write byte(s) into memory\n\r"\
"";
#define INLINESIZE 256

char *skipws(char *a);
char *skipnws(char *a);
int     Call(ULONG address);

LONG DoMonitor(int argc, char *argv[])
{
    int i;
    char line[INLINESIZE],*a;
    char out[80];
    ULONG addr1=0, addr2, curaddr=0, valuel;
    UBYTE valueb;

    PutStr((char *)intro);
    while(1)
    {
	PutStr((char *)prompt);
	if(GetLine(Input(), (char *)prompt, line, INLINESIZE, 1))
	{
	    /* EOF */
	    return -1;
	}
	a = skipws(line);

	if(!strncmp(line, "exit", 4) || !strncmp(line, "quit", 4))
	{
	    return 0;
	}

	switch(*a)
	{
	case 'h':
	    /* TODO: Hunt */
	case '?':
	    PutStr((char *)help);
	    break;
	case 'g':
	    a = skipws(a+1);
	    valuel = xtol(a);
	    sprintf(out, "Really jump to 0x%06x ? (Y/N): ", valuel);
	    PutStr(out);
	    SetMode(Input(), MODE_RAW);
	    if(Read(Input(), out, 1) && (out[0]=='y' || out[0]=='Y'))
	    {
		PutStr("Y\n\r");
		Call(valuel);
	    }
	    else
	    {
		PutStr("\n\r");
	    }
	    SetMode(Input(), MODE_CONSOLE);
	    break;
	case 'm':
	    addr2 = addr1 + 255; /* default */
	    a = skipws(a+1);
	    if(*a)
	    {
		addr1 = xtol(a);
		addr2 = addr1 + 255; /* default */
		a = skipws(skipnws(a));
		if(*a)
		{
		    addr2 = xtol(a);
		}
	    }
	    
	    for(curaddr=addr1;curaddr<=addr2;curaddr+=16)
	    {
		out[0] = hex[(curaddr>>20)&0xf];
		out[1] = hex[(curaddr>>16)&0xf];
		out[2] = hex[(curaddr>>12)&0xf];
		out[3] = hex[(curaddr>>8)&0xf];
		out[4] = hex[(curaddr>>4)&0xf];
		out[5] = hex[(curaddr)&0xf];
		out[6] = ' ';
		out[7] = ' ';

		if(curaddr&1)
		{
		    for(i=0;i<16;i++)
		    {
			valueb = *(UBYTE *)(curaddr+i);
			out[8+i*3] = hex[valueb>>4];
			out[9+i*3] = hex[valueb&0xf];
			out[10+i*3] = ' ';
			if(!iscntrl(valueb))
			    out[58+i] = (char)valueb;
			else
			    out[58+i] = '_';
		    }
		    out[56] = ' ';
		    out[57] = ' ';
		    out[74] = '\n';
		    out[75] = '\r';
		    out[76] = '\0';
		}
		else
		{
		    for(i=0;i<8;i++)
		    {
			valuel = *(UWORD *)(curaddr+2*i);
			out[8+i*5] = hex[valuel>>12];
			out[9+i*5] = hex[(valuel>>8)&0xf];
			out[10+i*5] = hex[(valuel>>4)&0xf];
			out[11+i*5] = hex[valuel&0xf];
			out[12+i*5] = ' ';
			if(!iscntrl(valuel>>8))
			    out[50+2*i] = (char)(valuel>>8);
			else
			    out[50+2*i] = '_';
			if(!iscntrl(valuel&0xff))
			    out[51+2*i] = (char)(valuel&0xff);
			else
			    out[51+2*i] = '_';
		    }
		    out[48] = ' ';
		    out[49] = ' ';
		    out[66] = '\n';
		    out[67] = '\r';
		    out[68] = '\0';
		}
		PutStr(out);
		if(WaitForChar(Input(), 0))
		    break;
	    }
	    addr1 = curaddr;
	    break;
	case 'i':
	    addr2 = addr1 + 511; /* default */
	    a = skipws(a+1);
	    if(*a)
	    {
		addr1 = xtol(a);
		addr2 = addr1 + 511; /* default */
		a = skipws(skipnws(a));
		if(*a)
		{
		    addr2 = xtol(a);
		}
	    }
	    out[6] = ' ';
	    out[7] = ' ';
	    out[72] = '\n';
	    out[73] = '\r';
	    out[74] = '\0';
	    
	    for(curaddr=addr1;curaddr<=addr2;curaddr+=64)
	    {
		out[0] = hex[(curaddr>>20)&0xf];
		out[1] = hex[(curaddr>>16)&0xf];
		out[2] = hex[(curaddr>>12)&0xf];
		out[3] = hex[(curaddr>>8)&0xf];
		out[4] = hex[(curaddr>>4)&0xf];
		out[5] = hex[(curaddr)&0xf];
		
		for(i=0;i<64;i++)
		{
		    valueb = *(UBYTE *)(curaddr+i);
		    if(!iscntrl(valueb))
			out[8+i] = (char)valueb;
		    else
			out[8+i] = '_';
		}
		PutStr(out);
		if(WaitForChar(Input(), 0))
		    break;
	    }
	    addr1 = curaddr;
	    break;
	case '*':
	    a = skipws(a+1);
	    if(*a)
	    {
		curaddr = addr1 = xtol(a);
		a = skipws(skipnws(a));
		while(*a)
		{
		    valueb = xtol(a);
		    *(UBYTE *)curaddr++ = valueb;
		    a = skipws(skipnws(a));
		}
	    }
	    break;
	case 'f':
	    addr2 = addr1 + 0; /* default */
	    a = skipws(a+1);
	    if(*a)
	    {
		addr1 = xtol(a);
		addr2 = addr1 + 0; /* default */
		a = skipws(skipnws(a));
		if(*a)
		{
		    addr2 = xtol(a);
		    a = skipws(skipnws(a));
		    if(*a)
		    {
			valuel = xtol(a);
			if(isxdigit(*(a+2)))
			{
			    if(addr1&1)
			    {
				for(curaddr=addr1;curaddr<=addr2;curaddr+=2)
				{
				    *(UBYTE *)curaddr = (valuel>>8);
				    *(UBYTE *)(curaddr+1) = valuel&0xff;
				}
			    }
			    else
			    {
				for(curaddr=addr1;curaddr<=addr2;curaddr+=2)
				{
				    *(UWORD *)curaddr = valuel;
				}
			    }
			}
			else
			{
			    /* One byte */
			    memset((void *)addr1, valuel, addr2-addr1+1);
			}
			break;
		    }
		}
	    }
	    PutStr("Usage: f <start> <end> <fill-byte/word>\n\r");
	    break;
	case '\0':
	    break;
#if 0
	case 'S':
	    if(*(a+2))
	    {
		valueb = count = xtol(a+2,2);
		for(i=0;i<count;i++)
		{
		    valueb += xtol(a+4+2*i,2);
		}
		valueb ^= 0xff;
		if(!valueb)	/* Checksum correct */
		{
		    switch(*(a+1))
		    {
		    case '1': /* Data record, 2-byte address */
			curaddr = xtol(a+4,4);
			a += 6;
			count -= 3;
			break;
		    case '2': /* Data record, 3-byte address */
			curaddr = xtol(a+4,6);
			a += 8;
			count -= 4;
			break;
		    case '3': /* Data record, 4-byte address */
			curaddr = xtol(a+4,8);
			a += 10;
			count -= 5;
			break;
		    case '0': /* Info-record */
			Write(ser, "S-Record Info:  ", 16);
			i = 0;
			while(count--)
			{
			    out[i++] = xtol((a+=2),2);
			}
			out[i++] = '\n';
			out[i++] = '\r';
			Write(ser, out, i);
			a = NULL;
			break;
		    case '5': /* Record counts */
		    case '7': /* Execution start address, 4-byte */
		    case '8': /* Execution start address, 3-byte */
		    case '9': /* Execution start address, 2-byte */
			a = NULL;
			break;
		    case '6':
		    case '4':
		    default:
			a = NULL;
			Write(ser, "Invalid S-record!\n\r", 19);
		    }
		    if(a)
		    {
			while(count--)
			{
			    *(UBYTE *)curaddr++ = xtol((a+=2),2);
			}
		    }
		}
		else
		{
		    Write(ser, "Checksum error!\n\r", 17);
		}
	    }
	    break;
#endif
	default:
	    PutStr("Unknown command!\n\r");
	}
    }
    return 0;
}

static LONG CheckLock(struct List *path, APTR lock)
{
    struct AssignList *al = (struct AssignList *)HeadNode(path);

    while(al)
    {
	if(SameLock(al->al_Lock, lock))
	{
	    return -1;
	}
	al = (struct AssignList *)NextNode((struct Node *)al);
    }
    return 0;
}

/* Usage: path [dir [..]] */
/*        path reset */
LONG DoPath(int argc, char *argv[])
{
    struct List *path = &((struct Process *)FindTask(NULL))->pr_Path;
    struct AssignList *al;
    int i;
    LONG error = RETURN_OK;

    /* Remove path */
    if(argc==2 && !stricmp(argv[1], "reset"))
    {
	while((al = (struct AssignList *)RemHead(path)))
	{
	    UnLock(al->al_Lock);
	    FreeMem(al, sizeof(struct AssignList));
	}
	return RETURN_OK;
    }

    /* Show path */
    if(argc<2)
    {
	char *buffer = (char *)AllocMem(1024, 0);

	if(buffer)
	{
	    PutStr("Current Working Directory\n");
	    al = (struct AssignList *)HeadNode(path);
	    while(al)
	    {
		if(getpath(buffer, 1024, al->al_Lock))
		{
		    PutStr(buffer);
		    PutStr("\n");
		}
		al = (struct AssignList *)NextNode((struct Node *)al);
	    }
	    FreeMem(buffer, 1024);
	    return RETURN_OK;
	}
	else
	{
	    PutStr("Memory Allocation error for 1024 bytes\n");
	    return RETURN_FAIL;
	}
    }

    /* Remove entries */
    if(!stricmp(argv[argc-1], "remove"))
    {
	for(i=1;i<argc-1;i++)
	{
	    APTR lock;

	    if((lock = Lock(argv[i], ACCESS_READ)))
	    {
		al = (struct AssignList *)HeadNode(path);
		while(al)
		{
		    if(SameLock(al->al_Lock, lock))
		    {
			Remove((struct Node *)al);
			UnLock(al->al_Lock);
			FreeMem(al, sizeof(struct AssignList));
			/* al not valid!! "break;" is needed!! */
			break;
		    }
		    al = (struct AssignList *)NextNode((struct Node *)al);
		}
		UnLock(lock);
	    }
	    else
	    {
		PrintFault(IoErr(), argv[i]);
		error = RETURN_ERROR;
	    }
	}
	return error;
    }

    /* Add entries */
    for(i=1;i<argc;i++)
    {
	APTR lock;

	if((lock = Lock(argv[i], ACCESS_READ)))
	{
	    /* Check if it is already in path */
	    if(CheckLock(path, lock))
	    {
		UnLock(lock);
	    }
	    else
	    {
		if((al = (struct AssignList *)
		    AllocMem(sizeof(struct AssignList), 0)))
		{
		    al->al_Lock = lock;
		    AddTail(path, (struct Node *)al);
		}
		else
		{
		    UnLock(lock);
		    PrintFault(ERROR_NO_FREE_STORE, argv[i]);
		    error = RETURN_ERROR;
		}
	    }
	}
	else
	{
	    PrintFault(IoErr(), argv[i]);
	    error = RETURN_ERROR;
	}
    }
    return error;
}


LONG DoPrgtofile(int argc, char *argv[])
{
    struct Program *prg;
    APTR file;

    if(argc!=3)
    {
	PutStr("Usage: prgtofile <program> <file>\n");
	return RETURN_ERROR;
    }
    if((prg = (struct Program *)FindProgram(argv[1])))
    {
	if((file = Open(argv[2], MODE_NEWFILE)))
	{
	    Write(file, (char *)prg->pm_Exec, prg->pm_Size);
	    Close(file);
	    return RETURN_OK;
	}
	else
	{
	    PrintFault(IoErr(), argv[2]);
	    return RETURN_ERROR;
	}
    }
    else
    {
	PutStr(argv[1]);
	PutStr(": Program not found\n");
	return RETURN_ERROR;
    }
}

LONG DoRelabel(int argc, char *argv[])
{
    if(argc != 3)
    {
	PutStr("Usage: relabel <disk> <newname>\n");
	return RETURN_FAIL;
    }
    if(!Relabel(argv[1], argv[2]))
    {
	PrintFault(IoErr(), argv[1]);
	return RETURN_FAIL;
    }
    return RETURN_OK;
}

LONG DoRename(int argc, char *argv[])
{
    APTR lock;

    if(argc != 3)
    {
	PutStr("Usage: rename <oldname> <newname>\n");
	return RETURN_FAIL;
    }
    if(!(lock = Lock(argv[1], ACCESS_READ)))
    {
	PrintFault(IoErr(), argv[1]);
	return RETURN_FAIL;
    }
    UnLock(lock);
    if(!Rename(argv[1], argv[2]))
    {
	PrintFault(IoErr(), argv[2]);
	return RETURN_FAIL;
    }
    return RETURN_OK;
}

LONG DoReset(int argc, char *argv[])
{
    PutStr("Resetting by disallowing interrupts for too long.\n");
    Disable();
    while(1)
	;
    Enable();    /* never reached, cosmetic */
    return RETURN_OK;
}

LONG DoRm(int argc, char *argv[])
{
    int i;
    LONG error = RETURN_OK;

    for(i=1;i<argc;i++)
    {
	if(!DeleteFile(argv[i]))
	{
	    PrintFault(IoErr(), argv[i]);
	    error = RETURN_WARN;
	}
	if((SetSignal(0, SIGF_CTRL_C) & SIGF_CTRL_C))
	{
	    Write(Error(), "^C\n***aborted\n", 14);
	    error = RETURN_FAIL;
	    break;
	}
    }
    return error;
}

static int ccsum(UBYTE *buffer, int size)
{
    long csum = 0;

    while(size)
    {
	csum += buffer[--size];
    }
    return 255-(csum&255);
}

LONG DoSet(int argc, char *argv[])
{
    struct LocalVar *var;

    if(argc==1)
    {
	/* Display all variables */
	var = (struct LocalVar *)
	    ((struct Process *)FindTask(NULL))->pr_LocalVars.lh_Head;
	while(var->lv_Node.ln_Next)
	{
	    if(var->lv_Flags == LV_VAR)
	    {
		PutStr(var->lv_Node.ln_Name);
		PutStr(" \t\"");
		PutStr((char *)var->lv_Value);
		PutStr("\"\n");
	    }
	    var = (struct LocalVar *)var->lv_Node.ln_Next;
	}
	return RETURN_OK;
    }
    if(argc==2)
    {
	if((var = FindVar(argv[1], LV_VAR)))
	{
	    PutStr(var->lv_Node.ln_Name);
	    PutStr(" \t\"");
	    PutStr((char *)var->lv_Value);
	    PutStr("\"\n");
	    return RETURN_OK;
	}
	return RETURN_ERROR;
    }
    if(argc==3)
    {
	if(!SetVar(argv[1], argv[2], strlen(argv[2])+1, LV_VAR))
	{
	    PutStr("Not set!\n");
	    return RETURN_ERROR;
	}
	return RETURN_OK;
    }
    PutStr("Usage: set [<name> [<value>]]\n");
    return RETURN_ERROR;
}

LONG DoUnset(int argc, char *argv[])
{
    int i;
    LONG error = RETURN_OK;

    for(i=1;i<argc;i++)
    {
	if(!DeleteVar(argv[i], LV_VAR))
	{
	    PutStr(argv[i]);
	    PutStr(": not found!\n");
	    error = RETURN_WARN;
	}
    }
    return error;
}

static int addtaskheader(char *a)
{
    return sprintf(a, "Task Name     Address   State  Pri    SigWait CPU/ms  Stack   Used\n");
}

static int addtaskinfo(char *a, struct Task *task)
{
    static const char * const states[8]= {"INVALID","ADDED  ","RUNNING","READY  ","WAITING","EXCEPT ","REMOVED","zombie?"};
    char *name = task->tc_Node.ln_Name;

    if(task->tc_Node.ln_Type==NT_PROCESS &&
       ((struct Process *)task)->pr_ProgramName)
    {
	name = ((struct Process *)task)->pr_ProgramName;
    }
    return sprintf(a, "%-12s 0x%06x %s %4ld 0x%08x %6d %6d %6d\n",
		   name,
		   (ULONG)task,
		   states[task->tc_State],
		   task->tc_Node.ln_Pri,
		   task->tc_SigWait,
		   task->tc_Elapsed*10,
		   task->tc_StackSize,
		   (ULONG)task->tc_Stack+task->tc_StackSize-(ULONG)task->tc_SPReg);
}

LONG DoTop(int argc, char *argv[])
{
    struct ExecBase *SysBase = *((struct ExecBase **)4);
    char *a, *b, c;
    ULONG alloc_size = 2048;
    struct Task *task;

    if((a = (char *)AllocMem(alloc_size, 0)))
    {
	SetMode(Input(), MODE_RAW);
	while(1)
	{
	    b = a;
	    b += addtaskheader(b);
	    task = FindTask(NULL);
	    b += addtaskinfo(b, task);

	    Disable();
	    task = (struct Task *)SysBase->TaskReady.lh_Head;
	    while((task->tc_Node.ln_Next) && (b-a<alloc_size-80))
	    {
		b += addtaskinfo(b, task);
		task = (struct Task *)task->tc_Node.ln_Next;
	    }
	    task = (struct Task *)SysBase->TaskWait.lh_Head;
	    while((task->tc_Node.ln_Next) && (b-a<alloc_size-80))
	    {
		b += addtaskinfo(b, task);
		task = (struct Task *)task->tc_Node.ln_Next;
	    }
	    Enable();
	    PutStr("\033[1;1H\033[2J");
	    Write(Output(), a, b-a);
	    if(WaitForChar(Input(), 300))
	    {
		Read(Input(), &c, 1);
		break;
	    }
	}
	FreeMem(a, alloc_size);
	SetMode(Input(), MODE_CONSOLE);
	return RETURN_OK;
    }
    PutStr("Memory allocation failed!\n");
    return RETURN_FAIL;
}

#define TYPE_HEX  1

LONG DoType(int argc, char *argv[])
{
    int i;
    LONG error = RETURN_OK, len, mode = 0, offset;
    APTR handle;
    char buffer[256];

    if(argc<2)
    {
	PutStr("Usage: type <filename> [<file2> [..]] [hex]\n");
	return RETURN_WARN;
    }

    for(i=1;i<argc;i++)
    {
	if(!stricmp(argv[i], "hex"))
	{
	    mode |= TYPE_HEX;
	}
    }

    for(i=1;i<argc;i++)
    {
	if(stricmp(argv[i], "hex"))
	{
	    if((handle = Open(argv[i], MODE_OLDFILE)))
	    {
		char buf2[80]="", *a = &buf2[0];

		offset = 0;
		while((len = Read(handle, buffer, 256)))
		{
		    if((mode & TYPE_HEX))
		    {
			int k;

			for(k=0;k<len;k++)
			{
			    if((offset&15)==0)
			    {
				a = &buf2[0];
				Write(Output(), a, strlen(a));
				a += sprintf(a, "\n%08x :", offset);
			    }
			    if((offset&1)==0)
			    {
				*a++ = ' ';
			    }
			    a += sprintf(a, "%02x", buffer[k]);
			    offset++;
			}
		    }
		    else
		    {
			Write(Output(), buffer, len);
			/* offset += len; */
		    }
		    if((SetSignal(0, SIGF_CTRL_C) & SIGF_CTRL_C))
		    {
			Write(Error(), "^C\n***aborted\n", 14);
			error = RETURN_FAIL;
			break;
		    }
		}
		if((mode & TYPE_HEX))
		{
		    Write(Output(), buf2, strlen(buf2));
		    Write(Output(), "\n", 1);
		}
		Close(handle);
	    }
	    else
	    {
		PrintFault(IoErr(), argv[i]);
		error = RETURN_WARN;
	    }
	}
    }
    return error;
}

LONG DoWhich(int argc, char *argv[])
{
    int n, l, error=RETURN_OK;
    struct Program *prg;	/* for commands on the resident list */
    struct FileInfoBlock fib;
    struct AssignList *al;
    long i;
    APTR lock; 				/* on source directory */
    char temp[256];
    struct SegList *seglist;

    if(argc < 2)
    {
	PutStr("Usage: which <commands>\n");
	return RETURN_WARN;
    }
    for(n=1; n<argc; n++)
    {
	if(FindVar(argv[n], LV_ALIAS))
	{
	    PutStr("Shell Alias\n");
	}
	/* Search my internal list for the command. */
	for(l = 0; Internals[l].name; l++)
	{
	    if(!stricmp(argv[n], Internals[l].name))
	    {
		PutStr("Shell Internal\n");
		break;
	    }
	}

	/* then current directory */
	if(lock = Lock(argv[n], ACCESS_READ))
	{
	    if(Examine(lock, &fib))
	    {
		if(fib.fib_DirEntryType < 0)
		{
		    i = fib.fib_Protection;
		    if((i & FIBF_SCRIPT))
		    {
			PutStr(argv[n]);
			PutStr(" [script]\n");
		    }
		    else if((i & FIBF_EXECUTE) ||
			    (i & FIBF_GRP_EXECUTE) ||
			    (i & FIBF_OTR_EXECUTE))
		    {
			if(seglist = LoadSeg(argv[n]))
			{
			    UnLoadSeg(seglist);
			    PutStr(argv[n]);
			    PutStr("\n");
			}
		    }
		}
	    }
	    UnLock(lock);
	}
	/* Then search the path */

	al = (struct AssignList *)
	    HeadNode(&((struct Process *)FindTask(NULL))->pr_Path);
	while(al)
	{
	    APTR oldcd = CurrentDir(al->al_Lock);
	    
	    if(oldcd)
	    {
		/* Currentdir changed */
		if(lock = Lock(argv[n], ACCESS_READ))
		{
		    if(Examine(lock, &fib))
		    {
			if(fib.fib_DirEntryType < 0)
			{
			    i = fib.fib_Protection;
			    if((i & FIBF_SCRIPT))
			    {
				if(!getpath(temp, 256, lock))
				    temp[0] = '\0';
				PutStr(temp);
				PutStr(" [script]\n");
			    }
			    else if((i & FIBF_EXECUTE) ||
				    (i & FIBF_GRP_EXECUTE) ||
				    (i & FIBF_OTR_EXECUTE))
			    {
				if(!getpath(temp, 256, lock))
				    temp[0] = '\0';
				if(seglist = LoadSeg(argv[n]))
				{
				    UnLoadSeg(seglist);
				    PutStr(temp);
				    PutStr("\n");
				}
			    }
			}
		    }
		    UnLock(lock);
		}
		CurrentDir(oldcd); /* Back to the old currentdir */
	    }
	    al = (struct AssignList *)NextNode((struct Node *)al);
	}

	/*
	   Search the Resident list for the command.
	 */
	if((prg = (struct Program *)FindProgram(argv[n])))
	{
	    PutStr("Resident\n");
	}
	if(strlen(argv[n]) < 255-6)
	{
	    sprintf(temp, "%s.b68k", argv[n]);
	    if((prg = (struct Program *)FindProgram(temp)))
	    {
		PutStr("Resident .b68k\n");
	    }
	}
    }
    return error;
}

void FreeCmdCache()
{
    if(CommandCache.buffer)
    {
	FreeMem(CommandCache.buffer, CommandCache.buffersize);
	CommandCache.buffer = 0;
	CommandCache.buffersize = 0;
	CommandCache.bufferindex = 0;
    }
    if(CommandCache.indices)
    {
	FreeMem(CommandCache.indices, CommandCache.maxcommands*sizeof(short));
	CommandCache.indices = 0;
	CommandCache.maxcommands = 0;
	CommandCache.commands = 0;
    }
}

int FindCmdCache(char *cmd)
{
    int j;

    for(j=0;j<CommandCache.commands;j++)
    {
	if(!stricmp(cmd, CommandCache.buffer+CommandCache.indices[j]))
	{
	    return -1;
	}
    }
    return 0;
}

static void ReAllocCmdCache(int sizedelta,int numdelta)
{
    char *temp;
    short *temp2;

    if(sizedelta)
    {
	if(temp = (char *)AllocMem(CommandCache.buffersize+sizedelta, 0))
	{
	    if(CommandCache.buffer)
	    {
		memcpy(temp, CommandCache.buffer, CommandCache.buffersize);
		FreeMem(CommandCache.buffer, CommandCache.buffersize);
	    }
	    CommandCache.buffer = temp;
	    CommandCache.buffersize += sizedelta;
	}
    }
    if(numdelta)
    {
	if(temp2 = (short *)AllocMem((CommandCache.maxcommands+numdelta)*sizeof(short), 0))
	{
	    if(CommandCache.indices)
	    {
		memcpy(temp2, CommandCache.indices, CommandCache.maxcommands*sizeof(short));
		FreeMem(CommandCache.indices, CommandCache.maxcommands*sizeof(short));
	    }
	    CommandCache.indices = temp2;
	    CommandCache.maxcommands += numdelta;
	}
    }
}


LONG DoRehash(int argc, char *argv[])
{
    struct AssignList *al;
    struct FileInfoBlock fib;
    struct Program *prg;
    long i,n;
    char temp[80];
    struct ExecBase *SysBase = *((struct ExecBase **)4);

    if(CommandCache.buffersize==0 && CommandCache.maxcommands==0)
    {
	ReAllocCmdCache(HASH_BUFINITIAL, HASH_NUMINITIAL);
	CommandCache.bufferindex = 0;
    }

    CommandCache.commands = 0;
    CommandCache.bufferindex = 0;

    for(i=0;Internals[i].name;i++)
    {
	n = strlen(Internals[i].name) + 1;
	if(CommandCache.commands == CommandCache.maxcommands)
	{
	    ReAllocCmdCache(0, HASH_NUMDELTA);
	}
	if(CommandCache.bufferindex + n >= CommandCache.buffersize)
	{
	    ReAllocCmdCache(HASH_BUFDELTA, 0);
	}

	if(!FindCmdCache(Internals[i].name))
	{
	    strcpy(CommandCache.buffer+CommandCache.bufferindex,
		   Internals[i].name);
	    CommandCache.indices[CommandCache.commands++] =
		CommandCache.bufferindex;
	    CommandCache.bufferindex += n;
	}
    }

    /* Forbid(); We assume that no-one else is messing with OUR path */
    al = (struct AssignList *)
	HeadNode(&((struct Process *)FindTask(NULL))->pr_Path);
    while(al)
    {
	if(Examine(al->al_Lock, &fib))
	{
	    while(ExNext(al->al_Lock, &fib))
	    {
		if(fib.fib_DirEntryType < 0)
		{
		    i = fib.fib_Protection;

		    n = strlen(fib.fib_FileName) + 1;
		    if(CommandCache.commands == CommandCache.maxcommands)
		    {
			ReAllocCmdCache(0, HASH_NUMDELTA);
		    }
		    if(CommandCache.bufferindex + n >=
		       CommandCache.buffersize)
		    {
			ReAllocCmdCache(HASH_BUFDELTA, 0);
		    }

		    if((i & FIBF_SCRIPT) ||
		       (i & FIBF_EXECUTE) ||
		       (i & FIBF_GRP_EXECUTE) ||
		       (i & FIBF_OTR_EXECUTE))
		    {
			if(!FindCmdCache(fib.fib_FileName))
			{
			    strcpy(CommandCache.buffer +
				   CommandCache.bufferindex,
				   fib.fib_FileName);
			    CommandCache.indices[CommandCache.commands++] =
				CommandCache.bufferindex;
			    CommandCache.bufferindex += n;
			}
		    }
		}
	    }
	    ExamineEnd(al->al_Lock, &fib);
	}
	al = (struct AssignList *)NextNode((struct Node *)al);
    }
    /* Permit(); */

    /*
       Search the Resident list for the command.
       */

    Forbid();
    prg = (struct Program *)HeadNode(&SysBase->ProgramList);
    
    while(prg)
    {
	n = strlen(prg->pm_Node.ln_Name) + 1;
	    
	strncpy(temp, prg->pm_Node.ln_Name, n-1);
	temp[n-1] = '\0';
	if(CommandCache.commands == CommandCache.maxcommands)
	{
	    ReAllocCmdCache(0, HASH_NUMDELTA);
	}
	if(CommandCache.bufferindex + n >= CommandCache.buffersize)
	{
	    ReAllocCmdCache(HASH_BUFDELTA, 0);
	}

	if(!FindCmdCache(temp))
	{
	    strcpy(CommandCache.buffer+CommandCache.bufferindex, temp);
	    CommandCache.indices[CommandCache.commands++] =
		CommandCache.bufferindex;
	    CommandCache.bufferindex += n;
	}
	prg = (struct Program *)NextNode((struct Node *)prg);
    }
    Permit();

    /* Current directory is handled separately in ccomplist() */

    if(argc)
    {
	sprintf(temp, "%d commands found\n", CommandCache.commands);
	PutStr(temp);
	sprintf(temp, "(buffer at 0x%x, %d/%d bytes used)\n",
		CommandCache.buffer,
		CommandCache.bufferindex,
		CommandCache.buffersize);
	PutStr(temp);
    }
    return RETURN_OK;
}

int ccomp(char *name)
{
    char	*found, *name2;
    int		len=strlen(name), i=0, j;
    BOOL	done=FALSE;
    const	int size=256;

    if(len>=256)	/* wont process longer commandnames than maximum.. */
    {
	PutStr("\a");
	return 0;
    }

    j=0;
    while(name[j] && name[j]!=':' && name[j]!='/')
	j++;

    if(name[j])
    {
	/* Absolute or relative path, normal filename completion */
	return fcomp(name);
    }

    if(found = (char *)AllocMem(size, MEMF_CLEAR))
    {
	if(CommandCache.buffersize==0 || CommandCache.maxcommands==0)
	    DoRehash(0, NULL);

	for(j=0;j<CommandCache.commands;j++)
	{
	    name2 = CommandCache.buffer+CommandCache.indices[j];
	    if(!strncmp(name, name2, len))
	    {
		if(name2[len]=='\0')
		{
		    /* exact command name */
		    strcpy(found, name);
		    done=TRUE;
		    break;
		}
		if(found[0])
		{
		    i=len;
		    while(found[i]==name2[i])
			i++;
		    if(found[i])	/* the commands are not identical */
		    {
			found[i] = '\0';
			done = FALSE;
		    }
		}
		else
		{
		    strcpy(found, name2);
		    done = TRUE;
		    if(!name[0])
			break;
		}
	    }
	}
	
	if(!done)
	{
	    int success = fcomp(name);

	    i=0;
	    while(tolower(name[i])==tolower(found[i]))
		i++;
	    if(found[i])
		name[i]='\0';
	    else
		done = success;
	}
	else
	{
	    strcat(name+len, " ");
	}
	    
	if(found[0])
	    strcpy(name+len, found+len);
	
	FreeMem(found,size);
    }
    return done;
}


void ccomplist(char *name, int Columns)
{
    int	len = strlen(name), j, num=0;
    struct FileInfoBlock *fib;
    APTR mylock;
    char *name2;
    
    if(len>=256)	/* wont process longer commandnames than maximum.. */
    {
	PutStr("\a");
	return;
    }

    j=0;
    while(name[j] && name[j]!=':' && name[j]!='/')
	j++;

    if(name[j])
    {
	/* Absolute or relative path, normal filename completion */
	fcomplist(name, Columns);
	return;
    }

    if(CommandCache.buffersize==0 || CommandCache.maxcommands==0)
	DoRehash(0,NULL);

    for(j=0;j<CommandCache.commands;j++)
    {
	name2 = CommandCache.buffer+CommandCache.indices[j];
	if(!strncmp(name, name2, len))
	{
	    num += strlen(name2)+2;
	    if(num > Columns)
	    {
		PutStr("\n");
		PutStr(name2);
		PutStr("  ");
		num = strlen(name2)+2;
	    }
	    else
	    {
		PutStr(name2);
		PutStr("  ");
	    }
	}
    }

    if(fib = (struct FileInfoBlock *)AllocMem(sizeof(struct FileInfoBlock), 0))
    {
	if(mylock=Lock("",ACCESS_READ))
	{
	    Examine(mylock, fib);
	    while(ExNext(mylock, fib))
	    {
		if(strncmp(name, fib->fib_FileName, len))
		    continue;

		j = fib->fib_Protection;
		if(fib->fib_DirEntryType > 0)
		{
		    num += strlen(fib->fib_FileName)+2;
		    if(num > Columns)
		    {
			PutStr("\n");
			PutStr(fib->fib_FileName);
			PutStr("/ ");
			num = strlen(fib->fib_FileName)+2;
		    }
		    else
		    {
			PutStr(fib->fib_FileName);
			PutStr("/ ");
		    }
		}
		else
		{
		    if((j & FIBF_SCRIPT) ||
		       (j & FIBF_EXECUTE) ||
		       (j & FIBF_GRP_EXECUTE) ||
		       (j & FIBF_OTR_EXECUTE))
		    {
			num += strlen(fib->fib_FileName)+2;
			if(num > Columns)
			{
			    PutStr("\n");
			    PutStr(fib->fib_FileName);
			    PutStr("  ");
			    num = strlen(fib->fib_FileName)+2;
			}
			else
			{
			    PutStr(fib->fib_FileName);
			    PutStr("  ");
			}
		    }
		}
	    }
	    ExamineEnd(mylock, fib);
	    UnLock(mylock);
	    if(num)
		PutStr("\n");
	}
	FreeMem(fib, sizeof(struct FileInfoBlock));
    }
}
