#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include <exec/execbase.h>
#include <exec/exec_protos.h>
#include <exec/tasks.h>
#include <exec/io.h>
#include <dos/dos.h>

#include "internal.h"

int shell(struct Task *parent);
int ExecuteFileHandle(APTR handle);
int ParseLine(char *buffer, int bufsize, /* Space for arguments */
	      char **argv, int maxargs,  /* Space for argument pointers */
	      char **diri, char **diro,  /* Redirection strings */
	      const char *argstr);       /* Unformatted argument string */
int RebuildLine(char *buffer, int bufsize, int argc, char *argv[]);
int LocateCommand(char *cmd, struct SegList **segl);
/* 0=not found, 1=command, 2=script, 3=directory */
LONG DoCd(int argc, char *argv[]);

const UWORD kludgejump = 0x4ef9;
const ULONG kludgeaddr = (ULONG)shell;

void __chkabort(void){}

char *skipws(char *a)
{
    while(isspace(*a))
	a++;
    return a;
}

char *skipnws(char *a)
{
    while(*a && !isspace(*a))
	a++;
    return a;
}

int getpath(char *buffer, int bufmax, APTR baselock)
{
    char *loc = buffer+bufmax-1; /* Last usable location */
    APTR lock, newlock;
    struct FileInfoBlock fib;
    int len;

    *loc = '\0'; /* trailing NUL */
    lock = baselock;
    while(lock)
    {
	Examine(lock, &fib);
	newlock = ParentDir(lock);
	if(lock != baselock)
	    UnLock(lock);
	lock = newlock;
	len = strlen(fib.fib_FileName);
	if(bufmax < len+2) /* Take NUL into account also! */
	{
	    strcpy(buffer, loc);
	    if(lock)
		UnLock(lock);
	    return DOS_FALSE;
	}
	if(fib.fib_DirEntryType >= 0) /* Is it a dir ? */
	{
	    --bufmax;
	    --loc;
	    *loc = (lock)?'/':':';
	}
	loc -= len;
	strncpy(loc, fib.fib_FileName, (size_t)len);
	bufmax -= len;
    }
    strcpy(buffer, loc);
    return DOS_TRUE;
}

char *process_prompt(char *p, int size)
{
    static char buf[256], tmp[256];
    char *q = &buf[0];
    long t;

    if(!p)
	return NULL;

    *q = '\0';
    while(*p)
    {
	if(size < 3)
	    return NULL;

	if(*p == '\\')
	{
	    p++;
	    switch(*p)
	    {
	    case 'e':
		*q++ = '\033';
		break;
	    case 'n':
		*q++ = '\n';
		break;
	    case 'r':
		*q++ = '\r';
		break;
	    case 't':
		*q++ = '\t';
		break;
	    case 'a':
		*q++ = '\a';
		break;
	    default:
		*q++ = *p;
	    }
	    p++;
	    *q = '\0';
	    size--;
	}
	else if(*p == '^')
	{
	    p++;
	    *q++ = *p++ & 0x1f;
	    size--;
	}
	else if(*p == '%')
	{
	    p++;
	    switch(*p)
	    {
	    case 'S':
		if(getpath(tmp, 256,
			   ((struct Process *)FindTask(NULL))->pr_CurrentDir))
		{
		    if(size < strlen(tmp))
			return NULL;

		    strcpy(q, tmp);
		    while(*q)
		    {
			size--;
			q++;
		    }
		}
		else
		{
		    *q++ = '%';
		    *q++ = *p;
		    *q = '\0';
		    size -= 2;
		}
		break;
	    case 'D':
		if(getpath(tmp, 256,
			   ((struct Process *)FindTask(NULL))->pr_CurrentDir))
		{
		    t = strlen(tmp)-2;
		    while(t>=0 && tmp[t]!='/' && tmp[t]!=':')
			t--;
		    
		    if(size < strlen(tmp+t+1))
			return NULL;

		    strcpy(q, tmp+t+1);
		    while(*q)
		    {
			size--;
			q++;
		    }
		}
		else
		{
		    *q++ = '%';
		    *q++ = *p;
		    *q = '\0';
		    size -= 2;
		}
		break;
	    case 'd':
		if(getpath(tmp, 256,
			   ((struct Process *)FindTask(NULL))->pr_CurrentDir))
		{
		    t = strlen(tmp)-2;
		    while(t>=0 && tmp[t]!='/' && tmp[t]!=':')
			t--;
		    
		    if(t>0)
			t--;
		    while(t>=0 && tmp[t]!='/' && tmp[t]!=':')
			t--;
		    
		    if(size < strlen(tmp+t+1))
			return NULL;

		    strcpy(q, tmp+t+1);
		    while(*q)
		    {
			size--;
			q++;
		    }
		}
		else
		{
		    *q++ = '%';
		    *q++ = *p;
		    *q = '\0';
		    size -= 2;
		}
		break;
	    default:
		*q++ = '%';
		*q++ = *p;
		*q = '\0';
		size -= 2;
	    }
	    p++;
	}
	else
	{
	    *q++ = *p++;
	    *q = '\0';
	    size--;
	}
    }
    return buf;
}

void PutStr(char *str)
{
    Write(Output(), str, strlen(str));
}



static const char welcome[]="New Shell Process\n";

#define INLINESIZE 256

struct DOSLibrary *DOSBase;

int shell(struct Task *parent)
{
    struct Process *myself = (struct Process *)FindTask(NULL);
    struct SegList *segl;
    char *oldprgname = myself->pr_ProgramName;
    APTR lock;

    if(!(DOSBase = (struct DOSLibrary *)OpenLibrary(DOSL_NAME)))
	return 50;

    if(!myself->pr_CurrentDir)
    {
	if(!(lock = Lock("ROM:", ACCESS_READ)))
	{
	    if((segl = ROMLoadSeg("rom.b68k")))
	    {
		CreateProc("rom-launcher", 5, segl, 4096, NULL);
		Delay(10);
		lock = Lock("ROM:", ACCESS_READ);
	    }
	}
	if(!lock)
	{
	    if(!(lock = Lock("RAM:", ACCESS_READ)))
	    {
		if((segl = ROMLoadSeg("ram.b68k")))
		{
		    CreateProc("ram-launcher", 5, segl, 4096, NULL);
		    Delay(10);
		    lock = Lock("RAM:", ACCESS_READ);
		}
	    }
	}
	if(!lock)
	{
	    if(!(lock = Lock("NIL:", ACCESS_READ)))
	    {
		if((segl = ROMLoadSeg("nil.b68k")))
		{
		    CreateProc("nil-launcher", 5, segl, 4096, NULL);
		    Delay(10);
		    lock = Lock("NIL:", ACCESS_READ);
		}
	    }
	}
	if(lock)
	{
	    myself->pr_CurrentDir = lock;
	    myself->pr_FileSystemTask = ((struct FileLock *)lock)->fl_Task;
	}
	else
	{
	    PutStr("No current directory: mount NIL:, RAM: or ROM: !\n");
	    return RETURN_ERROR;
	}
    }

    PutStr((char *)welcome);

    if((lock = Open("ROM:s/shell-startup", MODE_OLDFILE)))
    {
	ExecuteFileHandle(lock);
	Close(lock);
    }
    while(ExecuteFileHandle(Input()));

    myself->pr_ProgramName = oldprgname;
    FreeCmdCache();
    CloseLibrary((struct Library *)DOSBase);
    return 0;
}

#define ARGSTR_MAX 1024 /* Max size of the command line */
#define ARGN_MAX   64   /* Max number of arguments */
#define PROMPT_MAX 100

int ExecuteFileHandle(APTR handle)
{
    ULONG alloc_size =
	INLINESIZE + ARGSTR_MAX + PROMPT_MAX + ARGN_MAX*sizeof(char *) + 80;
    char *alloc_buf = (char *)AllocMem(alloc_size, MEMF_CLEAR);
    char **argv=(char **)alloc_buf /*[ARGN_MAX]*/, *diri, *diro;
    struct SegList *segl,*oldsegl;
    APTR redirout, redirin, oldoutput, oldinput;
    struct Remember *mem;
    ULONG sigalloc;
    int i, retval;
    char *line/*[INLINESIZE]*/= alloc_buf+ARGN_MAX*sizeof(char *), *a;
    char *argbuffer/*[ARGSTR_MAX]*/ = line+INLINESIZE,
         *prompt/*[PROMPT_MAX]*/ = argbuffer+ARGSTR_MAX,
         *out/*[80]*/ = prompt+80;
    int argc;
    struct Process *myself = (struct Process *)FindTask(NULL);

    if(!alloc_buf)
    {
	PutStr("Memory allocation error!\n");
	return -1;
    }

    while(1)
    {
	if(handle == Input())
	{
	    if(GetVar("prompt", prompt, PROMPT_MAX, LV_VAR) == -1)
		strcpy(prompt, "%S > ");
	    if(!(a = process_prompt(prompt, PROMPT_MAX)))
		a = "shell > ";
	    PutStr(a);
	    if(GetLine(handle, a, line, INLINESIZE, 1))
	    {
		/* EOF */
		FreeMem(alloc_buf, alloc_size);
		return -1;
	    }
	}
	else
	{
	    i = 0;
	    while(1)
	    {
		if(Read(handle, line+i, 1) != 1)
		{
		    /* EOF */
		    FreeMem(alloc_buf, alloc_size);
		    return -1;
		}
		if(!line[i])
		    break;
		if(line[i]=='\r' || line[i]=='\n')
		{
		    line[i+1] = '\0';
		    break;
		}
		i++;
	    }
	}

	a = skipws(line);
	i = 0;
	while(*a && !isspace(*a))
	{
	    prompt[i++] = *a++;
	}
	if(i)
	{
	    prompt[i] = '\0';
	    if(GetVar(prompt, prompt, PROMPT_MAX, LV_ALIAS)!=-1)
	    {
		if(strlen(prompt) + strlen(a) < INLINESIZE)
		{
		    char temp[INLINESIZE];

		    strcpy(temp, a);
		    strcpy(line, prompt);
		    strcat(line, temp);
		}
		else
		{
		    PutStr("Alias discarded for result being to long!\n");
		}
	    }
	}
	argc = ParseLine(argbuffer, ARGSTR_MAX,
			 argv, ARGN_MAX,
			 &diri, &diro,
			 line);
	if(!argc)
	    continue; /* empty command */
	if(!strcmp(argv[0], "quit"))
	    break;

	sigalloc = myself->pr_Task.tc_SigAlloc;

	redirout = NULL;
	redirin = NULL;
	oldoutput = myself->pr_COS;
	oldinput = myself->pr_CIS;
	if(diri)
	{
	    if(redirin = Open(diri+1, MODE_OLDFILE))
	    {
		myself->pr_CIS = redirout;
	    }
	    else
	    {
		PrintFault(IoErr(), diri+1);
		retval = 20;
		goto cleanup;
	    }
	}
	if(diro)
	{
	    i = 0;
	    while(diro[i]=='>')
		i++;
	    if(i>1)
	    {
		if(redirout = Open(diro+i, MODE_READWRITE))
		{
		    if(!Seek(redirout, 0, OFFSET_END))
		    {
			Close(redirout);
			redirout = NULL;
			PutStr("Seek Failed!\n");
			retval = 20;
			goto cleanup;
		    }
		}
		else
		{
		    PrintFault(IoErr(), diro+i);
		    retval = 20;
		    goto cleanup;
		}
	    }
	    else
	    {
		if(!(redirout = Open(diro+i, MODE_NEWFILE)))
		{
		    PrintFault(IoErr(), diro+i);
		    retval = 20;
		    goto cleanup;
		}
	    }
	    if(redirout)
	    {
		myself->pr_COS = redirout;
	    }
	}

	retval = 0;
	i = 0;
	while(Internals[i].name)
	{
	    if(!strcmp(Internals[i].name, argv[0]))
	    {
		myself->pr_ProgramName = argv[0];
		retval = (*Internals[i].func)(argc, argv);
		myself->pr_ProgramName = "shell";
		break;
	    }
	    i++;
	}
	if(!Internals[i].name)
	{
	    if(!RebuildLine(line, INLINESIZE, argc, argv))
	    {
		/* <name> <args> */
		segl = NULL;

		i = LocateCommand(argv[0], &segl);
		if(!i)
		{
		    PrintFault(IoErr(), argv[0]);
		    retval = 20;
		}
		else if(i==3)
		{
		    char *newargv[2];

		    newargv[0] = "cd";
		    newargv[1] = argv[0];
		    retval = DoCd(2, newargv);
		}
		else if(i==2)
		{
		    ExecuteFileHandle((APTR)segl);
		    retval = 0;
		    Close((APTR)segl);
		    segl = NULL; /* To prevent it to be executed */
		}
		else if(i==1)
		{
		    if(segl)
		    {
			oldsegl = (struct SegList *)myself->pr_SegList;
			myself->pr_SegList = (APTR)segl;
			myself->pr_ProgramName = argv[0];
			retval =  RunCommand(segl, 4096, line, strlen(line));
			myself->pr_ProgramName = "shell";
			if(myself->pr_SegList)
			    UnLoadSeg((struct SegList *)myself->pr_SegList);
			myself->pr_SegList = (APTR)oldsegl;
		    }
		}
	    }
	    else
	    {
		PutStr("Could not rebuild argument line!\n");
		retval = 10;
	    }
	}
cleanup:
	if(redirin)
	{
	    myself->pr_CIS = oldinput;
	    Close(redirin);
	}
	if(redirout)
	{
	    myself->pr_COS = oldoutput;
	    Close(redirout);
	}
	if(retval != 0)
	{
	    sprintf(out, "%s failed: returncode %ld\n", argv[0], retval);
	    PutStr(out);
	}
	if(myself->pr_Task.tc_SigAlloc != sigalloc)
	{
	    sprintf(out, "Signals not freed by the program! (before: 0x%08x, after: 0x%08x)\n\rRestoring old allocation state\n\r", sigalloc, myself->pr_Task.tc_SigAlloc);
	    myself->pr_Task.tc_SigAlloc = sigalloc;
	    PutStr(out);
	}
	while((mem = (struct Remember *)RemHead(&FindTask(NULL)->tc_MemList)))
	{
	    sprintf(out, "Freeing @ 0x%08x, %ld bytes\n\r",
		    mem+1, mem->rm_Size - sizeof(struct Remember));
	    PutStr(out);
	    FreeMem(mem, mem->rm_Size);
	}
    }
    FreeMem(alloc_buf, alloc_size);
    return 0;
}

int LocateCommandPath(char *cmd, struct SegList **segl);
int LocateCommandPath(char *cmd, struct SegList **segl)
{
    APTR lock;

    if((lock = Lock(cmd, ACCESS_READ)))
    {
	struct FileInfoBlock fib;
		
	Examine(lock, &fib);
	UnLock(lock);
	if(fib.fib_DirEntryType > 0)
	{
	    return 3;
	}
	else if(fib.fib_Protection & FIBF_SCRIPT)
	{
	    if((*segl = (struct SegList *)Open(cmd, MODE_OLDFILE)))
	    {
		return 2;
	    }
	    return 0;
	}
	else if(fib.fib_Protection & FIBF_EXECUTE)
	{
	    if((*segl = LoadSeg(cmd)))
	    {
		return 1;
	    }
	    return 0;
	}
	else
	{
	    SetIoErr(ERROR_EXECUTE_PROTECTED);
	    return 0;
	}
    }
    SetIoErr(ERROR_OBJECT_NOT_FOUND);
    return 0;
}

int LocateCommand(char *cmd, struct SegList **segl)
{
    char out[80];
    struct AssignList *al;
    int cmdmode = 0, retval, i;
    LONG lasterr = ERROR_OBJECT_NOT_FOUND;

    i = 0;
    while(cmd[i])
    {
	if(cmd[i]==':')
	    cmdmode |= 1; /* Absolute path */
	if(cmd[i]=='/')
	    cmdmode |= 2; /* Relative path */
	i++;
    }

    if(!cmdmode) /* No path specified */
    {
	/* Search from currentdir, path and then the resident list */

	/* Try the current directory */
	retval = LocateCommandPath(cmd, segl);
	if(retval)
	    return retval;
	if(IoErr() != ERROR_OBJECT_NOT_FOUND)
	    lasterr = IoErr();

	/* search path */
	al = (struct AssignList *)
	    HeadNode(&((struct Process *)FindTask(NULL))->pr_Path);
	while(al)
	{
	    APTR oldcd = CurrentDir(al->al_Lock);
	    
	    if(oldcd)
	    {
		/* Currentdir changed */
		retval = LocateCommandPath(cmd, segl);
		if(IoErr() != ERROR_OBJECT_NOT_FOUND)
		    lasterr = IoErr();
		CurrentDir(oldcd); /* Back to the old currentdir */
		if(retval)
		{
		    return retval;
		}
	    }
	    al = (struct AssignList *)NextNode((struct Node *)al);
	}
	SetIoErr(lasterr);

	strcpy(out, cmd);
	strcat(out, ".b68k");

	if((*segl = ROMLoadSeg(cmd)) ||
	   (*segl = ROMLoadSeg(out)))
	{
	    return 1; /* Resident Command found */
	}
	return 0;
    }

    /* Try relative or absolute path */
    retval = LocateCommandPath(cmd, segl);
    return retval;
}

int ParseLine(char *buffer, int bufsize, /* Space for arguments */
	      char **argv, int maxargs,  /* Space for argument pointers */
	      char **diri, char **diro,
	      const char *argstr)
{
    int argn = 0, mode = 0;
    const char *a = argstr;
    char *d = buffer;

    *diri = NULL;
    *diro = NULL;
    argv[0] = d;
    while(*a)
    {
	while(isspace(*a))
	    a++;
	if(!*a)
	    break;
	while(*a)
	{
	    if(*a == '\"')
	    {
		mode = 1 - mode;
		a++;
	    }
	    else
	    {
		if(!mode && isspace(*a))
		    break;

		if(!mode && *a=='\\' && *(a+1))
		    a++;

		*d++ = *a++;
		if(--bufsize<=0)
		    return argn;
	    }
	}
	*d++ = '\0';

	if(argv[argn][0]=='<' && !*diri)
	{
	    *diri = argv[argn];
	}
	else if(argv[argn][0]=='>' && !*diro)
	{
	    *diro = argv[argn];
	}
	else
	{
	    if(++argn == maxargs)
		return argn;
	}
	argv[argn] = d;
    }
    return argn;
}

int RebuildLine(char *buffer, int bufsize, int argc, char *argv[])
{
    char *a = buffer, *b;
    int i, escaped, len;

    for(i=0;i<argc;i++)
    {
	escaped = 0;
	b = argv[i];
	if(!(len = strlen(b)))
	    escaped = 1;
	while(*b)
	{
	    if(*b == '\\' || isspace(*b))
	    {
		escaped = 1;
		bufsize -= 2;
	    }
	    else if(*b == '\"')
	    {
		escaped = 2;
		break;
	    }
	    b++;
	}
	if(escaped == 2)
	{
	    b = argv[i];
	    while(*b)
	    {
		if(bufsize <= 2)
		    return -1;
		if(*b == '\"' || *b == '\\' || isspace(*b))
		{
		    *a++ = '\\';
		    bufsize--;
		}
		*a++ = *b++;
		bufsize--;
	    }
	}
	else
	{
	    bufsize -= len;
	    if(bufsize < 4)
		return -1;

	    if(escaped)
		*a++ = '\"';
	    strcpy(a, argv[i]);
	    a += len;
	    if(escaped)
		*a++ = '\"';
	}
	if(i != argc-1)
	{
	    *a++ = ' ';
	    bufsize--;
	}
    }
    *a++ = '\n';
    *a++ = '\0';
    return 0;
}

