#include <stdio.h>
#include <stdlib.h>

#if 0
int printf(char *, ...);
int fprintf(FILE *, char *, ...);
int fread(char *, int, int, FILE *);
int fseek(FILE *, int, int);
void fclose(FILE *);
#endif

struct exec
{
    unsigned char   a_toolversion;  /* version of toolset used to create */
    unsigned char   a_machtype;	/* machine type */
    unsigned short  a_magic;	/* magic number */
    
    unsigned long   a_text;	/* size of text segment */
    unsigned long   a_data;	/* size of initialized data */
    unsigned long   a_bss;	/* size of uninitialized data */
    unsigned long   a_syms;	/* size of symbol table */
    unsigned long   a_entry;	/* entry point */
    unsigned long   a_trsize;	/* size of text relocation */
    unsigned long   a_drsize;	/* size of data relocation */
};

#define	OMAGIC	0407		/* old impure format */
#define	NMAGIC	0410		/* read-only text */
#define	ZMAGIC	0413		/* demand load format */

/* machine types */
#define M_OLDSUN2	0	/* old sun-2 executable files */
#define M_68010		1	/* runs on either 68010 or 68020 */
#define M_68020		2	/* runs only on 68020 */
#define M_SPARC		3	/* runs only on SPARC */

struct reloc_info_68k
{
    long  r_address;	/* address which is relocated */
    unsigned int
    r_symbolnum:24,	/* local symbol ordinal */
    r_pcrel:1, 	        /* was relocated pc relative already */
    r_length:2,	        /* 0=byte, 1=word, 2=long */
    r_extern:1,	        /* does not include value of sym referenced */
    r_baserel:1,	/* linkage table relative */
    r_jmptable:1,	/* pc-relative to jump table */
    r_relative:1,	/* relative relocation */
	:1;
};

struct	nlist
{
    union
    {
	char	*n_name;	/* for use when in-core */
	long	n_strx;		/* index into file string table */
    } n_un;
    unsigned char	n_type;		/* type flag (N_TEXT,..)  */
    char	n_other;		/* unused */
    short	n_desc;			/* see <stab.h> */
    unsigned long	n_value;	/* value of symbol (or sdb offset) */
};

/*
 * Simple values for n_type.
 */
#define	N_UNDF	0x0		/* undefined */
#define	N_ABS	0x2		/* absolute */
#define	N_TEXT	0x4		/* text */
#define	N_DATA	0x6		/* data */
#define	N_BSS	0x8		/* bss */
#define	N_COMM	0x12		/* common (internal to ld) */
#define	N_FN	0x1e		/* file name symbol */

#define	N_EXT	01		/* external bit, or'ed in */
#define	N_TYPE	0x1e		/* mask for all the type bits */

#define	N_STAB	0xe0		/* if any of these bits set, a dbx symbol */


int showreloc(struct reloc_info_68k *reloc,struct nlist *symboltable,int symbolnum)
{
    static const char *size[4]={"byte","word","long","unknown"};

    if(reloc->r_extern)
    {
	if(reloc->r_symbolnum < symbolnum)
	{
	    printf("Addr: 0x%08lx (%s) External %s",reloc->r_address, size[reloc->r_length], symboltable[reloc->r_symbolnum].n_un.n_name);
	    if((symboltable[reloc->r_symbolnum].n_type == N_EXT) &&
	       symboltable[reloc->r_symbolnum].n_value)

	    {
		printf(" (global BSS %ld bytes)\n",
		       (long)symboltable[reloc->r_symbolnum].n_value);
		return 01;
	    }
	    else
	    {
		if(reloc->r_pcrel)
		{
		    printf(" (PC-Relative)");
		}
		printf("\n");
	    }
	}
	else
	{
	    printf("Addr: 0x%08lx (%s) External (unavailable)\n",reloc->r_address, size[reloc->r_length]);
	}
	return 1;
    }
    else
    {
	if(reloc->r_pcrel)
	{
	    printf("Addr: 0x%08lx (%s) PC-relative\n",reloc->r_address, size[reloc->r_length]);
	}
	if(reloc->r_symbolnum & N_STAB)
	{
	}
	else
	{
	    printf("Addr: 0x%08lx (%s) ",reloc->r_address, size[reloc->r_length]);
	    switch(reloc->r_symbolnum & ~N_EXT)
	    {
	    case N_ABS:
		printf("Absolute\n");
		break;
	    case N_TEXT:
		printf("Text-relative\n");
		break;
	    case N_DATA:
		printf("Data-relative\n");
		break;
	    case N_BSS:
		printf("BSS-relative\n");
		break;
	    case N_COMM:
		printf("Common\n");
		break;
	    case N_FN:
		printf("Filename\n");
		break;
	    default:
		printf("unknown\n");
		break;
	    }
	}
    }
    return 0;
}

void main(int argc, char *argv[])
{
    FILE *fp;
    struct exec header;
    struct reloc_info_68k *relocs;
    struct nlist *symboltable;
    char *stringtable,temp[80];
    int l,i,relocnum,symbolnum,n,possible;
    long stringsize,extcode,extdata;

    if(argc < 2)
    {
	fprintf(stderr,"Usage: %s a.out-filenames\n",argv[0]);
	exit(10);
    }
    for(n=1; n<argc; n++)
    {
	if((fp=fopen(argv[n],"rb")))
	{
	    possible = 1;
	    if(fread((char *)&header, sizeof(struct exec), 1, fp))
	    {
		if(header.a_magic!=OMAGIC && header.a_magic!=NMAGIC && header.a_magic!=ZMAGIC)
		{
		    printf("\"%s\": Not a.out format!\n", argv[n]);
		    fclose(fp);
		    continue;
		}
		if(argc>2)
		{
		    strcpy(temp, "::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::");
		    if(strlen(argv[n])<72)
		       temp[strlen(argv[n])+6]= '\0';
		    printf("%s\n:: %s ::\n%s\n", temp, argv[n], temp);
		}
		if(header.a_toolversion & 0x80)
		    printf("Dynamic\n");
		else
		    printf("Static\n");
		printf("Toolversion : %ld (", (long)(header.a_toolversion & 0x7f));
		switch(header.a_toolversion & 0x7f)
		{
		case 0:
		    printf("Assembler");
		    break;
		case 1:
		    printf("Linker");
		    break;
		case 68:
		    printf("Xlink");
		    break;
		default:
		    printf("unknown");
		    break;
		}
		printf(")\nMachine type: %ld (", (long)header.a_machtype);
		switch(header.a_machtype)
		{
		case M_68010:
		    printf("68010");
		    break;
		case M_68020:
		    printf("68020");
		    break;
		case M_SPARC:
		    printf("SPARC");
		    break;
		case 0:
		    printf("68010");
		    break;
		default:
		    printf("unknown");
		    break;
		}
		printf(")\nMagic number: 0x%04lx (", (long)header.a_magic);
		switch(header.a_magic)
		{
		case OMAGIC:
		    printf("OMAGIC");
		    break;
		case NMAGIC:
		    printf("NMAGIC");
		    break;
		case ZMAGIC:
		    printf("ZMAGIC");
		    break;
		default:
		    printf("unknown");
		    break;
		}
		printf(")\nCode size   : 0x%lx\n", header.a_text);
		printf("Data size   : 0x%lx\n", header.a_data);
		printf("BSS size    : 0x%lx\n", header.a_bss);
		printf("Symbol size : 0x%lx\n", header.a_syms);
		printf("Entry point : 0x%lx\n", header.a_entry);
		printf("Code reloc size: 0x%lx\n", header.a_trsize);
		printf("Data reloc size: 0x%lx\n", header.a_drsize);
		
		if(header.a_trsize || header.a_drsize || header.a_syms)
		{
		    if( (header.a_magic == OMAGIC || header.a_magic == NMAGIC)
		       && header.a_machtype!=M_SPARC)
		    {
			relocnum = (header.a_trsize+header.a_drsize)/
			    sizeof(struct reloc_info_68k);
			extcode = 0;
			extdata = 0;
			symbolnum = 0;
			if( ( relocs = (struct reloc_info_68k *)
			     calloc(sizeof(struct reloc_info_68k), relocnum)))
			{
			    /* Skip code and data in this program */
			    fseek(fp, header.a_text+header.a_data, 1/*current*/);
			    fread((char *)relocs, sizeof(struct reloc_info_68k), relocnum, fp);
			    if( ( symboltable =
				 (struct nlist *)calloc(header.a_syms, 1)))
			    {
				symbolnum = header.a_syms/sizeof(struct nlist);
				fread((char *)symboltable, header.a_syms, 1, fp);
				if(fread((char *)&stringsize, 4, 1, fp))
				{
				    stringsize -= 4;
				}
				else
				{
				    stringsize = 0;
				}
				if( ( stringtable = (char *)calloc(stringsize,1)))
				{
				    fread(stringtable, stringsize, 1, fp);

				    for(i=0;i<symbolnum;i++)
				    {
					symboltable[i].n_un.n_name = stringtable+symboltable[i].n_un.n_strx-4;
					if(!strcmp("_end",symboltable[i].n_un.n_name))
					{
					    possible = 0;
					}
				    }
				    l = header.a_trsize/sizeof(struct reloc_info_68k);
				    if(l)
				    {
					printf("Code relocations:\n");
					for(i=0; i<l; i++)
					{
					    extcode += showreloc(&relocs[i], symboltable,symbolnum);
					}
					if(extcode)
					    printf("%ld unresolved Code references!\n",extcode);
				    }
				    l = header.a_drsize/sizeof(struct reloc_info_68k);
				    if(l)
				    {
					printf("Data relocations:\n");
					for(;l--;i++)
					{
					    extdata += showreloc(&relocs[i], symboltable,symbolnum);
					}
					if(extdata)
					    printf("%ld unresolved Data references!\n",extdata);
				    }
				    free(stringtable);
				}
				free(symboltable);
			    }
			    free(relocs);
			}
			if(possible)
			{
			    if(!extcode && !extdata)
			    {
				if(!relocnum)
				{
				    printf("Position-Independent Code\n");
				}
				else
				{
				    printf("Relocation possible\n");
				}
			    }
			    else
			    {
				if(!symbolnum)
				{
				    printf("Unusable file (Unresolved symbols and no symbol information)\n");
				}
				else
				{
				    printf("Relocation possible after external references are resolved\n");
				}
			    }
			}
			else
			{
			    printf("Relocation NOT possible! (unless this is PIC)\nLinked to address 0x%lx\n",header.a_entry);
			}
		    }
		    else
		    {
			printf("Executable format not handled!\n");
		    }
		}
		else
		{
		    printf("Stripped!\n");
		}
	    }
	    else
	    {
		printf("\"%s\": does not include a header\n",argv[n]);
	    }
	    fclose(fp);
	}
	else
	{
	    printf("Can't open \"%s\"\n",argv[n]);
	}
	printf("\n");
    }
    exit(0);
}
