#include <signal.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/uio.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <termios.h>
#include <signal.h>

#include "crctab.c"

/* For BSD */
#ifndef TCGETATTR
#define TCGETATTR TCGETS
#endif

/* For BSD */
#ifndef TCSETATTR
#define TCSETATTR TCSETS
#endif

static char Base64[]=
"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";
/* The Base64 alphabet (big and small letters, numbers, plus and slash) */

long calcsum(int size, long *buf, int clear);
long bintobase(FILE *in, int out, int mode);

#define  TIMEOUT   10

static int ChildExited = 0;
static int ChildExit = 0;

static void SigHandler()
{
    ChildExited=-1;
    return;
}


struct termios org, new;

void rememberterminal()
{
  ioctl(0, TCGETATTR, &org);   /* nykyiset arvot talteen */
}

void initterminal(int timeout)
{
  ioctl(0, TCGETATTR, &new); 
  new.c_lflag &= ~ECHO;
  new.c_lflag &= ~ICANON; /* ei-kanoninen input-moodi */

  /* seuraavat arvot maarittelevat luettavien merkkien minimimaaran ja */
  /* niiden lukemiseen maksimissaan kaytettavan ajan kymmenesosasekunteina */
  /* tarkempaa selvitysta loytyy komennolla 'man termio' */
  new.c_cc[VMIN] = 0;   /* ei odoteta valttamatta yhtaan merkkia */
  new.c_cc[VTIME] = timeout; /* TIMEOUT*0.1s */
  ioctl(0, TCSETATTR, &new);
}
 
void backterminal()
{
  ioctl(0, TCSETATTR, &org);  
}

static void SigCont()
{
    initterminal(TIMEOUT);
    signal(SIGCONT, SigCont);
    return;
}

static int sock;

static void SigInt()
{
    write(sock, "\003", 1);
    signal(SIGINT, SigInt);
}

static void SigQuit()
{
    write(sock, "\031", 1);
    signal(SIGQUIT, SigQuit);
    signal(SIGTSTP, SigQuit);
}

long calcsum(int size, long *buf, int clear)
{
    int i,j,a,index=0;
    static int cs0=0,cs1=0,cs2=0;       /* Initialize to 0 only once */
    long value;

    static char a0[]=
{0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1};
    static char a1[]=
{1,1,1,1,1,1,1,1,3,3,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,2,2,
 2,2,2,2,2,2,2,3,3,3,3,3,3,3,3,3,4,4,4,4,4,4,4,4,4,5,5,5,5,
 5,5,5,5,5,6,6,6,6,6,6,6,6,6,7,7,7,7,7,7,7,7,7,8,8,8,8,8,8,8};
    static char a2[]=
{1,2,3,4,5,6,7,8,1,2,1,2,3,4,5,6,7,8,0,1,2,3,4,5,6,7,8,0,1,
 2,3,4,5,6,7,8,0,1,2,3,4,5,6,7,8,0,1,2,3,4,5,6,7,8,0,1,2,3,
 4,5,6,7,8,0,1,2,3,4,5,6,7,8,0,1,2,3,4,5,6,7,8,0,1,2,3,4,5,6};

    if(clear)
    {
	cs0=cs1=cs2=0;
    }
    for(i=0;i<size;i++)		/* Handle all elements */
    {
	value = buf[i];		/* Get the next 24 bits */
	for(j=0;j<8;j++)	/* Make it 8*3 bits */
	{
	    if((a=(value & 0x07)))	/* Get 3 bits */
	    {
		cs0 += a0[index] * a;	/* Matrix multiplication */
		cs1 += a1[index] * a;
		cs2 += a2[index] * a;
	    }
	    value >>= 3;	/* Forget the processed 3 bits */
	    index++;		/* Go to the next row in the matrix */
	}
    }
    cs0 %= 9;   /* Convert to a remainder class */
    cs1 %= 9;   /* These can be outside of the loop, because */
    cs2 %= 9;   /* 88*9*8 = 6336, which fits into integer nicely */
    return ((cs0<<8) | (cs1<<4) | cs2); /* return a 12-bit checksum */
}

#define MODE_CSUM64     (0)
#define MODE_BASE64	(1<<1)
long bintobase(FILE *in,int out,int mode)
{
    unsigned char buf[80],outbuf[80];
    long quads[20];	/* We have 9 extra for Base64-case (longer lines) */
    long csum,value;
    int size,i,pos,outp,numbytes=33;

    if((mode & MODE_BASE64))
	numbytes=57;

    while((size=fread(buf,sizeof(char),numbytes,in)))
    {
	/* Handle all triplets */
	for(outp=pos=i=0;i+2<size;i+=3)
	{
	    quads[pos++]= (value=(buf[i]<<16)|(buf[i+1]<<8)|buf[i+2]);
	    /* We assign to the quads-table in base64-mode also, but it */
	    /* should not slow down things too much. I just didn't want */
	    /* to duplicate the code. */
	    outbuf[outp++]=Base64[(value>>18)&0x3f];
	    outbuf[outp++]=Base64[(value>>12)&0x3f];
	    outbuf[outp++]=Base64[(value>>6)&0x3f];
	    outbuf[outp++]=Base64[value&0x3f];
	}
	/* Handle leftover doubles */
	if(i+1<size)
	{
	    quads[pos++]= (value = (buf[i]<<16) | (buf[i+1]<<8));
	    outbuf[outp++]=Base64[(value>>18)&0x3f];
	    outbuf[outp++]=Base64[(value>>12)&0x3f];
	    outbuf[outp++]=Base64[(value>>6)&0x3f];
	    outbuf[outp++]='=';
	}
	else	/* Handle leftover singles */
	if(i<size)
	{
	    quads[pos++]= (value = (buf[i]<<16));
	    outbuf[outp++]=Base64[(value>>18)&0x3f];
	    outbuf[outp++]=Base64[(value>>12)&0x3f];
	    outbuf[outp++]='=';
	    outbuf[outp++]='=';
	}
	if(mode & MODE_BASE64)
	{
	    outbuf[outp++]='\n';
	    write(out,outbuf,outp);
	}
	else
	{
	    csum=calcsum(pos,quads,1); /* Totally unrelated checksums */
	    outbuf[outp++]=Base64[(csum>>6)&0x3f];
	    outbuf[outp++]=Base64[(csum&0x3f)];
	    outbuf[outp++]='\n';
	    write(out,outbuf,outp);
	}
    }
    return 0;
}

static unsigned char *receivebuf(void)
{
    unsigned char header[4], *buffer;
    int  rec = 0, len, csum = 0, i;

    while(rec < 4)
    {
	rec += read(sock, header+rec, 4-rec);
    }
    len = header[0]*256 + header[1];
    if(!(buffer = (unsigned char *)malloc(len)))
    {
	return NULL;
    }
    memcpy(buffer, header, 4);
    while(rec < len)
    {
	rec += read(sock, buffer+rec, len-rec);
    }
    for(rec=0;rec<len;rec++)
	csum += buffer[rec];
    if((csum & 255) != 255)
    {
	printf("Csum: %d (%d != 255)\n", csum, csum & 255);
	fflush(stdout);
	write(1, "csum error\n", 11);

	free(buffer);
	return NULL;
    }
    return buffer;
}


#define NUL 0x00
#define SOH 0x01
#define STX 0x02
#define EOT 0x04
#define ACK 0x06
#define NAK 0x15
#define CAN 0x18
#define MAXRETRIES	10
#define MAXBLOCKSIZE	1024

static int receiveblock(int fid, unsigned char *block, int expblocknum, int *size, int crcmode)
{
    unsigned char keys[2];
    unsigned int sum, crc;
    int blokki, i=0, j;
    fd_set readfds;
    struct timeval timeout;

    while(1)
    {
	FD_ZERO(&readfds);
	FD_SET(fid, &readfds);
	timeout.tv_sec = 10;
	timeout.tv_usec = 0;
	/* wait for SOH/STX/EOT (10 secs) */

	if(select(FD_SETSIZE, &readfds, NULL, NULL, &timeout)>0)
	    j = read(fid, keys, 1);
	else
	    j = 0;
	if(j<1)
	{
	    return -4;
	}
	if(keys[0]==SOH)
	{
	    *size = 128; /* normal block size */
	    break;
	}
	if(keys[0]==STX)
	{
	    *size = 1024; /* 1k block size */
	    break;
	}
	if(keys[0]==EOT)
	    return -3;
	if(keys[0]==CAN)
	    return -2;
    }

    j = read(fid, keys, 2); /* Get block # and its complement */
    if(j==1)
    {
	j += read(fid, keys+1, 1); /* Get block # complement */
    }
    if(j!=2)
    {
	return -2;
    }
    blokki = keys[0];
    if(blokki != 255-keys[1])
    {
	blokki = expblocknum;	/* we can't be sure which one is wrong */
    }

    /* Handle the wrapping in block numbers (only 8 bits are transferred) */
    blokki |= (expblocknum&~255);

    i = 0;
    while(i<*size)
    {
	j = read(fid, block+i, *size-i);
	if(j>0)
	    i += j;
	else
	    break;					/* carrier lost */
    }
    crc = sum = 0;
    for(i=0;i<*size;i++)
    {
	crc = updcrc(block[i], crc);
	sum += block[i];
    }
    sum &= 255;

    if(crcmode)
    {
	j = read(fid, keys, 1);
	if(j!=1)
	{
	    fprintf(stderr,"CRC16 error: %x\n", crc);
	    return -2;
	}
	crc = updcrc(keys[0], crc);
	j = read(fid, keys+1, 1);
	if(j!=1)
	{
	    fprintf(stderr,"CRC16 error: %x\n", crc);
	    return -2;
	}
	crc = updcrc(keys[1], crc);
	if(crc & 0xffff)
	{
	    fprintf(stderr,"CRC16 error: %x\n", crc);
	    return -2;
	}
    }
    else
    {
	j = read(fid, keys, 1);
	
	if(j!=1 || keys[0]!=sum)
	{
	    fprintf(stderr,"Checksum error: %x %x\n", sum, keys[0]);
	    return -2;
	}
    }
    return blokki;
}


int rx(int fid, char *filename)
{
    unsigned char block[MAXBLOCKSIZE], a;
    int blocknum=1, recblock, insize, crcmode = 0, blksize;
    int files = 0, retries;
    FILE *fp;
    long filesize = -1;
    unsigned long timeval;

    fprintf(stderr, "Receiving...\n");
    while(1)
    {
	timeval = time(NULL);
	insize = 0;
	retries = 0;
	blocknum = 1;
	if(0)
	{
	    write(fid, "\025", 1); /* initial NAK */
	}
	else
	{
	    write(fid, "C", 1); /* initial NAK for crc16 */
	    fprintf(stderr, "Using CRC16\n");
	    crcmode = 1;
	}
	while(1)
	{
	    recblock = receiveblock(fid, block, blocknum, &blksize, crcmode);
	    if(read(0, &a, 1))
	    {
		if(a==0x03 || a==24) /* ^C ^X */
		{
		    write(fid, "\030\030", 2); /* CAN CAN */
		    printf("\nTransfer aborted.\n\r");
		    return files;
		}
	    }
	    if(recblock==-4)
	    {
		fprintf(stderr, "Timeout in receiveblock\n");
		if(++retries>MAXRETRIES)
		{
		    recblock = -2;
		    break;
		}
		if(!fp)
		    write(fid, "C", 1); /* Initial NAK for crc16 */
		else
		    write(fid, "\025", 1); /* Some other NAK */
		continue;
	    }
	    if(recblock==-2 || recblock==-3)
		break;

	    if(recblock == 0)
	    {
		if(!block[0])
		{
		    write(fid, "\006", 1); /* ACK the block */
		    return files;
		}

		if((fp = fopen(block, "wb")))
		{
		    filesize = atol(block+strlen(block)+1);
		    fprintf(stderr, "Receiving \"%s\" (%ld bytes) with ymodem(-batch)\n",
			    block, filesize);
		    write(fid, "\006", 1); /* ACK the block */
		    write(fid, "\025", 1); /* NAK to get the first block */
		    /* TODO: for CRC16 send 'C' instead */
		}
		else
		{
		    fprintf(stderr, "Could not open '%s' for writing\n",
			    block);
		    write(fid, "\030\030", 2); /* CAN CAN */
		    printf("\nTransfer aborted.\n\r");
		    return files;
		}
		timeval = time(NULL);
		blocknum = 1;
	    }
	    else if(recblock >= 0)
	    {
		if(recblock == blocknum)
		{
		    if(!fp)
		    {
			/* This is then Xmodem..
			   Use the filename given to us. */
			fp = fopen(filename, "wb");
			if(!fp)
			{
			    fprintf(stderr,
				    "Could not open \"%s\" for writing\n",
				    filename);
			    write(fid, "\030\030", 2); /* CAN CAN */
			    printf("\nTransfer aborted.\n\r");
			    return files;
			}
			fprintf(stderr,
				"Receiving \"%s\" with xmodem\n",
				block);
		    }
		    /* if right block, save it */
		    if((filesize<0) || (filesize-insize >= blksize))
		    {
			if(fp)
			    fwrite(block, blksize, 1, fp);
			insize += blksize;
		    }
		    else
		    {	
			if(fp)
			    fwrite(block, filesize-insize, 1, fp);
			insize = filesize;
		    }
		    blocknum++;
		    write(fid, "\006", 1); /* ACK the block */
		}
		else if(recblock < blocknum)
		{
		    write(fid, "\006", 1); /* ACK if we are ahead */
		}
		else
		{
		    if(fp)
		    {
			fclose(fp);
		    }
		    write(fid, "\030\030", 2); /* CAN CAN */
		    printf("\nTransfer aborted.\n\r");
		    return files;
		}
	    }
	    else
	    {
		/* if block is bad, but we are ahead, ACK it */
		if(recblock<blocknum)
		{
		    write(fid, "\006", 1);
		}
		else
		{
		    /* if we need this block or can't make it clear, NAK it */
		    fd_set readfds;
		    struct timeval timeout;

		    FD_ZERO(&readfds);
		    FD_SET(fid, &readfds);
		    timeout.tv_sec = 0;
		    timeout.tv_usec = 1000;

		    if(select(FD_SETSIZE, &readfds, NULL, NULL, &timeout)>0)
		    {
			/* Clear the input buffer to get in sync better */
			read(fid, block, MAXBLOCKSIZE);
		    }
		    write(fid, "\025", 1); /* NAK */
		    retries++;
		}
	    }
	    fprintf(stderr, "Block: %d Bytes: %d\r", recblock, insize);
	}
	if(recblock==-2)
	{
	    if(fp)
		fclose(fp);
	    write(fid, "\030\030", 2); /* CAN CAN */
	    printf("\nTransfer aborted.\n\r");
	    return files;
	}
	write(fid, "\006", 1);	/* acknowledge the EOT */
	fclose(fp);
	fp = NULL;
	files++;

	timeval = time(NULL)-timeval;
	if(timeval)
	    printf("\nTransfer successful. (%ld cps)\n", insize/timeval);
	else
	    printf("\nTransfer successful.\n");
    }
    return files;
}


int getkey(int fid, unsigned long);
int getblock(FILE *fid, unsigned char *block, int size);
void sendblock(int fid, unsigned char *block, unsigned long blocknum, int size, int crcmode);

#define MAXBLOCKSIZE 1024

int sx(int fid, char **file)
{
    int c;
    int retries = 0, firstblock, blocksize, crcmode = 0;
    unsigned char block[MAXBLOCKSIZE], a;
    int blocknum, lastblock;
    FILE *fp;
    unsigned long size = 0;
    char *fn;
    unsigned long timeval;

    printf("Sending using x/ymodem-chksum/crc16. Use ctrl-X's to cancel\n\r");

    while(1)
    {
	fp = NULL;
	blocksize = 128;
	if(*file)
	{
	    if((fp = fopen(*file, "rb")))
	    {
		fseek(fp, 0, 2); /* offset-end */
		size = ftell(fp);
		fseek(fp, 0, 0);
	    }
	    else
	    {
		printf("'%s' not found.\n\r", *file++);
		continue;
	    }
	}

	for(c=0;c<blocksize;c++)
	{
	    block[c] = 0;
	}
	if(*file)
	{
	    fn = *file;
	    c = 0;
	    while(file[0][c])
	    {
		if(file[0][c++]=='/')
		    fn = &file[0][c];
	    }
	    sprintf(block, "%s%c%d", fn, 0, size);
	    printf("Sending \"%s\"\n", *file);
	}
	crcmode = 0;
	lastblock = 0;
	blocknum = 0;
	firstblock = 1;
	retries = 0;

	while(1)
	{
	    while(retries<MAXRETRIES)
	    {
		c = getkey(fid, 1000);
		if(read(0, &a, 1))
		{
		    if(a==0x03 || a==24) /* ^C ^X */
		    {
			fclose(fp);
			write(fid, "\030\030", 2); /* CAN CAN */
			printf("\nTransfer aborted.\n\r");
			return 0;
		    }
		}
		if(firstblock)
		{
		    timeval = time(NULL);
		    /* block 0 ACK'ed, wait for start of transfer */
		    if(c==ACK)
		    {
			if(!fp)
			{
			    return 0;
			}
			firstblock = 0;
			blocknum = 1;
			blocksize = 1024;
			if(getblock(fp, block, blocksize))
			{
			    lastblock = 1;
			}
		    }
		    if(c=='C')
		    {
fprintf(stderr, "CRC16-mode initialized\n");
			crcmode = 1;
			c = NAK;
		    }
		    if(c==NAK)
		    {
			sendblock(fid, block, blocknum, blocksize, crcmode);
			++retries;
		    }
		}
		else
		{
		    if(c==ACK)
		    {
			if(retries)
			    retries--;
			break;
		    }
		    if(c==CAN)
		    {
			retries += 2;
		    }
		    if(c=='C')
		    {
			c = NAK;
		    }
		    if(c==NAK)
		    {
			/* TODO: if retries >0, blocknum==1 */
			/* fall back to 128-byte blocks */
			sendblock(fid, block, blocknum, blocksize, crcmode);
			firstblock = 0;
		    }
		    ++retries;
		}
	    }
	    if(retries>=MAXRETRIES)
	    {
		retries = MAXRETRIES;
		break;
	    }
	    if(lastblock)
	    {
		write(fid, "\004", 1); /* EOT */
		firstblock = 0;
		break;
	    }
	    blocksize = 1024;
	    if(getblock(fp, block, blocksize))
	    {
		lastblock = 1;
		write(fid, "\004", 1); /* EOT */
		firstblock = 0;
		break;
	    }
	    blocknum++;
	    sendblock(fid, block, blocknum, blocksize, crcmode);
	    firstblock = 0;
	}
	if(retries>=MAXRETRIES)
	{
	    fclose(fp);
	    write(fid, "\030\030", 2); /* CAN CAN */
	    fprintf(stderr, "Transfer aborted due to too many retries!\n");
	    return 0;
	}
	retries=0;
	while(retries<MAXRETRIES)
	{
	    c = getkey(fid, 1000);
	    if(c==ACK)
		break;
	    if(c==CAN)
	    {
		retries += 2;
	    }
	    if(c==NAK)
	    {
		write(fid, "\004", 1); /* EOT */
	    }
	    ++retries;
	}
	timeval = time(NULL)-timeval;
	if(retries>=MAXRETRIES)
	{
	    fprintf(stderr, "\nFinal ACK missing.\n");
	}
	else
	{
	    if(timeval)
		printf("\nTransfer successful. (%ld cps)\n", size/timeval);
	    else
		printf("\nTransfer successful.\n");
	}
	fclose(fp);
	file++;
    }
    return 0;
}

int getkey(int fid, unsigned long timev)
{
    fd_set readfds;
    struct timeval timeout;
    char q;
    int j;

    FD_ZERO(&readfds);
    FD_SET(fid, &readfds);
    timeout.tv_sec = timev/100;
    timeout.tv_usec = (timev%100) * 1000;

    if(select(FD_SETSIZE, &readfds, NULL, NULL, &timeout)>0)
	j = read(fid, &q, 1);
    else
	j = 0;

    if(j==1)
    {
	return q;
    }
    return -1;
}


int getblock(FILE *fp, unsigned char *block, int blocksize)
{
    int i;

    i = fread(block, 1, blocksize, fp);
    if(!i)
    {
	return -1; /* No more data */
    }
    if(i<blocksize)
    {
	while(i<blocksize)
	    block[i++] = '\0';
    }
    return 0;
}


void sendblock(int fid, unsigned char *block, unsigned long blocknum, int size, int crcmode)
{
    unsigned char q[3];
    unsigned long crc = 0, sum = 0;
    int i;

fprintf(stderr, "\rSending block %ld", blocknum);
    for(i=0;i<size;i++)
    {
	crc = updcrc(block[i], crc);
	sum += block[i];
    }

    if(size==128)
	q[0] = SOH;
    else
	q[0] = STX;
    q[1] = (blocknum&255);
    q[2] = 255-(blocknum&255);
    write(fid, q, 3);
    write(fid, block, size);

    if(crcmode)
    {
	crc = updcrc(0, updcrc(0, crc));
	q[0] = (unsigned char)(crc>>8);
	q[1] = (unsigned char)(crc&255);
	write(fid, q, 2);
    }
    else
    {
	q[0] = sum;
	write(fid, q, 1);
    }
}



#define READBUF 512

void main(int argc, char *argv[])
{
    int len, z;
    pid_t child;
    char line[READBUF], out[100];
    struct sockaddr_in server;
    struct hostent *hp;
    char *host, *port;
    FILE *fp;
 
    host="sigx8.cs.tut.fi";
    port="9100";

    if(argc>1)
	host = argv[1];
    if(argc>2)
	port = argv[2];

    sock=socket(AF_INET, SOCK_STREAM, 0);
    if(sock<0)
    {
	fprintf(stderr,"Could not get socket\n");
	exit(10);
    }
  
    server.sin_family=AF_INET;
    hp=gethostbyname(host);
    server.sin_port=htons(atoi(port));
    memcpy(&server.sin_addr, hp->h_addr, hp->h_length);

    if(connect(sock, (struct sockaddr *)&server, sizeof(server)) < 0)
    {
	fprintf(stderr,"Could not connect\n");
	close(sock);
	exit(11);
    }

    rememberterminal();
    printf("^P to quit, ^U for BASE64 upload, ^T/^R for ymodem up/download\n");
    printf("Connection established to %s, port %s\n",host,port);
    signal(SIGCONT, SigCont);
    signal(SIGCHLD, SIG_IGN);
    signal(SIGQUIT, SIG_IGN);
    signal(SIGINT, SIG_IGN);
    signal(SIGTSTP, SIG_IGN);
    if((child = fork()))
    {
	if(child == -1)
	{
	    fprintf(stderr, "fork() error\n");
	    close(sock);
	    exit(12);
	}
	initterminal(TIMEOUT);
	signal(SIGUSR1, SigHandler);
	signal(SIGINT, SigInt);
	signal(SIGQUIT, SigQuit);
	signal(SIGTSTP, SigQuit);
	while(!ChildExited)
	{
	    /* read palauttaa lukemiensa merkkien lukumaaran */
	    if((len = read(0, line, READBUF)))
	    {
		if(line[0] == 18) /* ^R */
		{
		    kill(child, SIGSTOP);
		    signal(SIGINT, SIG_IGN);
		    signal(SIGQUIT, SIG_IGN);

		    initterminal(0);
		    rx(sock, "/tmp/xmodem-tmp");
		    write(sock, "\n", 1);
		    initterminal(TIMEOUT);

		    signal(SIGINT, SigInt);
		    signal(SIGQUIT, SigQuit);
		    kill(child, SIGCONT);
		}
		else if(line[0] == 21) /* ^U */
		{
		    write(1, "File to upload: ",16);
		    backterminal();
		    len = read(0, line, READBUF);
		    if(len)
		    {
			line[len--] = '\0';
			while(line[len]=='\n' || line[len]=='\r')
			    line[len--] = '\0';
			    
			if((fp = fopen(line, "rb")))
			{
			    fseek(fp, 0, 2); /* Wind to the end of file */
			    len = ftell( fp );
			    fseek(fp, 0, 0); /* Rewind to the start  */
			    z = strlen(line)-1;
			    while(z && line[z]!='/')
				z--;
			    if(line[z]=='/')
				z++;
			    sprintf(out, "L %s %d\n", line+z, len);
			    write(sock, out, strlen(out));

			    sleep(1);
			    bintobase(fp, sock /* 1/sock */, MODE_BASE64);
			    fclose(fp);
			}
			else
			{
			    sprintf(out, "Could not open \"%s\" for reading\n", line);
			    write(1, out, strlen(out));
			}
		    }
		    initterminal(TIMEOUT);
		}
		else if(line[0] == 20) /* ^T */
		{
		    write(1, "File to upload: ",16);
		    backterminal();
		    kill(child, SIGSTOP);
		    signal(SIGINT, SIG_IGN);
		    signal(SIGQUIT, SIG_IGN);
		    len = read(0, line, READBUF);
		    initterminal(TIMEOUT);
		    if(len)
		    {
			fd_set readfds;
			struct timeval timeout;
			char *file[10];
			int files = 0, l;

			line[len--] = '\0';
			while(line[len]=='\n' || line[len]=='\r')
			    line[len--] = '\0';
			    
			file[files++] = line;
			l = 0;
			while(l<len)
			{
			    while(line[l] && line[l]!=' ')
				l++;
			    if(line[l])
				line[l++] = '\0';
			    while(line[l]==' ')
				l++;
			    if(line[l])
			    {
				file[files++] = line+l;
			    }
			}
			file[files] = NULL;

			write(sock, "rx\r", 3);

			sleep(1);
			/*usleep(100000);*/ /* 1/10 sec */
			FD_ZERO(&readfds);
			FD_SET(sock, &readfds);
			timeout.tv_sec = 0;
			timeout.tv_usec = 1000;

			if(select(FD_SETSIZE, &readfds, NULL, NULL, &timeout)>0)
			{
			    char tmpbuf[64];

			    read(sock, tmpbuf, 64);
			}
			initterminal(0);
			sx(sock, file);
			initterminal(TIMEOUT);
		    }
		    signal(SIGINT, SigInt);
		    signal(SIGQUIT, SigQuit);
		    kill(child, SIGCONT);
		}
		else if(line[0] == 16) /* ^P */
		{
		    write(1, "Connection Closed.\n", 19);
		    break;
		}
		else
		{
		    write(sock, line, len);
		}
	    }
	}
	kill(child, SIGKILL);
    }
    else
    {
	signal(SIGINT, SIG_IGN);
	while((len = read(sock, line, READBUF)))
	{
	    write(1, line, len);

	    if(strstr(line, "***S"))
	    {
		write(1, "Incoming data! ^R to start transfer, several ^X to abort\n", 57);
		line[0] = '\0';
	    }

	}
	if(!ChildExit)
	    write(1, "Connection Closed By Foreign Host.\n", 35);
	kill(0, SIGUSR1);
	exit(0);
    }
    backterminal();
    close(sock);
    exit(0);
}

