#include <stdio.h>
#include <stdlib.h>

#if 0
/* Our system does not have proper <stdio.h> ! */
int printf(char *, ...);
int fprintf(FILE *, char *, ...);
int fread(char *, int, int, FILE *);
int fwrite(char *,int,int,FILE *);
int fseek(FILE *, int, int);
void fclose(FILE *);
#endif

/* Some shorthands */
#define UWORD unsigned short
#define ULONG unsigned long
#define UBYTE unsigned char

#define MAXCONFS 32 /* Number of configurations to merge */

const UBYTE defheader[13] = {0,9,15,240,15,240,15,240,15,240,0,0,1};

int main(int argc, char *argv[])
{
    UBYTE header[13], *temp, *configs[MAXCONFS];
    int n,i, config = 0, error = 0;
    FILE *fp;
    UBYTE chunktype;
    UWORD chunklen;
    ULONG size, count, sizes[MAXCONFS], counts[MAXCONFS];

    if(argc<3)
    {
	fprintf(stderr, "Usage: %s <outfile> <infile1> [<infile2> .. ]\n",
		argv[0]);
	return 20;
    }
    if(argc+2>MAXCONFS)
    {
	fprintf(stderr, "Only %d configurations can be merged in one pass\n");
	return 20;
    }

    for(n=2;n<argc;n++)
    {
	if((fp = fopen(argv[n], "rb")))
	{
	    if(fread((char *)header, 1, 13, fp) != 13)
	    {
		fclose(fp);
		fprintf(stderr, "Short file: \"%s\".\n", argv[n]);
		error = 20;
		goto cleanup;
	    }
	    /* TODO: check header */

	    if(memcmp(header, defheader, 13))
	    {
		fclose(fp);
		fprintf(stderr, "Invalid header in file: \"%s\".\n", argv[n]);
		error = 20;
		goto cleanup;
	    }

	    while(fread((char *)&chunktype, 1, 1, fp))
	    {
		switch(chunktype)
		{
		case 'a':
		    fread((char *)&chunklen, sizeof(UWORD), 1, fp);
		    size = chunklen;
		    break;
		case 'b':
		    fread((char *)&chunklen, sizeof(UWORD), 1, fp);
		    size = chunklen;
		    break;
		case 'c':
		    fread((char *)&chunklen, sizeof(UWORD), 1, fp);
		    size = chunklen;
		    break;
		case 'd':
		    fread((char *)&chunklen, sizeof(UWORD), 1, fp);
		    size = chunklen;
		    break;
		case 'e':
		    fread((char *)&size, sizeof(ULONG), 1, fp);
		    break;
		default:
		    fclose(fp);
		    fprintf(stderr, "Unknown chunk in file \"%s\"!\n",
			    argv[n]);
		    error = 20;
		    goto cleanup;
		}
		temp = NULL;
		if(size)
		{
		    if(!(temp = calloc(size, sizeof(char))))
		    {
			fclose(fp);
			fprintf(stderr, "Memory allocation error!\n");
			error = 20;
			goto cleanup;
		    }
		    if(fread((char *)temp, 1, size, fp) != size)
		    {
			fclose(fp);
			fprintf(stderr, "Short chunk in file \"%s\"!\n",
				argv[n]);
			free(temp);
			error = 20;
			goto cleanup;
		    }
		}
		switch(chunktype)
		{
		case 'a':
		    printf("Name of the design: %s\n", temp);
		    break;
		case 'b':
		    printf("Type of chip (?):   %s\n", temp);
		    break;
		case 'c':
		    printf("Date of creation:   %s\n", temp);
		    break;
		case 'd':
		    printf("Time of creation:   %s\n", temp);
		    break;
		case 'e':
		    count = ((temp[1]&0x0f)<<20) |
			(temp[2]<<12) | (temp[3]<<4) | (temp[4]>>4);
		    printf("Configuration data: %d bytes, %d bits, %d extra one-bits\n",
			   size, count, size*8-count);
		    sizes[config] = size;
		    counts[config] = count;
		    configs[config++] = temp;
		    temp = NULL;
		    break;
		default:
		    fclose(fp);
		    free(temp);
		    fprintf(stderr, "Unknown chunk in file \"%s\"!\n",
			    argv[n]);
		    error = 20;
		    goto cleanup;
		}
		if(temp)
		    free(temp);
	    }
	    fclose(fp);
	}
	else
	{
	    fprintf(stderr, "Could not open \"%s\" for reading!\n", argv[n]);
	    error = 20;
	    goto cleanup;
	}
    }
    /* Now we have "config" number of configurations */
    if((fp = fopen(argv[1], "wb")))
    {
	printf("Writing to \"%s\"\n", argv[1]);
	size = 5;
	for(i=0;i<config;i++)
	{
	    size += sizes[i]-5;
	}
	count = size*8;
	fwrite((char *)defheader, 1, 13, fp);
	fwrite("e", 1, 1, fp);
	fwrite((char *)&size, sizeof(ULONG), 1, fp);
	header[0] = 0xff;
	header[1] = 0x20 | (count>>20);
	header[2] = (count>>12) & 0xff;
	header[3] = (count>>4) & 0xff;
	header[4] = ((count<<4) & 0xf0) | 0x0f;
	fwrite((char *)header, 1, 5, fp);

	for(i=0;i<config;i++)
	{
	    fwrite((char *)configs[i]+5, 1, sizes[i]-5, fp);
	}

	fclose(fp);
    }
    else
    {
	fprintf(stderr, "Could not open \"%s\" for writing!\n", argv[1]);
	error = 20;
    }

cleanup:
    while(config)
    {
	free(configs[--config]);
    }
    return error;
}
